/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	ARP */
/*****************************************************************************/
#include "arp.h"
/*****************************************************************************/
/*	ARP-Setup */
/*****************************************************************************/
void	arp_init(void){
	/*	The Beginning of the arp-query-queue:
		This is a list of arp-queries */
	EnterCritical();	//CRITICAL SECTION
	gspArpQueryQueue		= 0;
	gsppArpQueryQueueLast	= &gspArpQueryQueue;
	/*	No Arp Request from outside */
	gu8ArpRequest			= 0;
	/*	No Arp Transmit */
	gu8ArpTransmit			= 0;
	LeaveCritical();	//CRITICAL S3ECTION
	/*	Setup the Timer */
	TIMERSETCALLBACK(&gsArpTimer,arp_timer_callback);
	StartTimer(&gsArpTimer,ARP_QUERYTIMEOUT);
}
/*****************************************************************************/
/*	ARP:	Timer */
/*****************************************************************************/
void	arp_timer_callback(TIMER* pTimer){
ARPQUERYNODE**		sppNode;
void (*pCallback)(ARPQUERYNODE* pNode,u8 u8Code);
	sppNode	= &gspArpQueryQueue;		//Start at the beginning
	while(*sppNode != 0){
		/*	Send one more time a request */
		(*sppNode)->u8Retry++;
		if((*sppNode)->u8Retry > ARP_RETRY){
			/*	We tried to often -> stop */
			pCallback	= (*sppNode)->pCallback;
			*sppNode	= (*sppNode)->pNext;
			/*	Call the Callback-Function: Lookup failed */
			(pCallback)((*sppNode),ARP_QUERYERROR);
		}else{
			/*	Ensure that we transmit the request */
			gu8ArpTransmit	= 1;
			/*	Get next entry of the list */
			sppNode	= &(*sppNode)->pNext;
		}
	}
	if(gu8ArpTransmit	== 1){
	/*	If we should send a request, wake up the Network-Task */
		gu8NetworkSend = 1;
		SignalEvent(&NetworkEVENT);
	}
	/*	Restart the timer again. NOTE: We dont need exact timing, thus
		we simply start the timer, instead of restarting it. */
	StartTimer(pTimer,ARP_QUERYTIMEOUT);
}
/*****************************************************************************/
/*	Sends an ARP-query	*/
/*****************************************************************************/
/*	IMPRORTANT: This uses the <gcTxBuffer> to assemble the message. Therefor
	the content isn't secure and the function can only be used within the
	TX-Task!!! */
void	arp_sendquery(ARPQUERYNODE* pQueryNode){
/*	start the arp-timeout timer */
/*	add the node to the list */
	EnterCritical();		//CRITICAL SECTION
	pQueryNode->pNext	= gspArpQueryQueue;
	gspArpQueryQueue	= pQueryNode;
/*	Reset the Retransmit-Counter */
	pQueryNode->u8Retry	= 0;
/*	Ensure we transmit the stuff */
	gu8ArpTransmit	= 1;
	gu8NetworkSend	= 1;
	LeaveCritical();		//CRITICAL SECTION
	SignalEvent(&NetworkEVENT);
}
/*****************************************************************************/
void	arp_transmit(void){
/*	ARP-Transmit Handler */
ETHERNETHEADER*			spEthernetHeader		= (ETHERNETHEADER*)&gu8NetworkBuffer;
ARPHEADER*				spARPHeader;
ARPQUERYNODE*			spNode;
	if(gu8ArpTransmit == 1){
		/*	We should transmit a request */
		gu8ArpTransmit = 0;
		EnterCritical();	//CRITICAL SECTION
		spNode	= gspArpQueryQueue;		//Start at the beginning
		LeaveCritical();	//CRITICAL SECTION
		while(spNode != 0){

			/*	Send an ARP-Request-Message */
			/*	Send an Ethernet-Packet:
					I'm the sender (mymac)
					Everyone is the receiver (broadmac)
					Protocol: ARP (0x0806)	*/
				spARPHeader	= ethernet_writeheader(spEthernetHeader,&mymac,&broadmac,0x0806);
			/*	Setup the ARP-Header */
				/*	Hardware: ETHERNET (1) */
				spARPHeader->u8HardwareTypeHigh = HI(1);
				spARPHeader->u8HardwareTypeLow  = LO(1);
				/*	Protocol-Type: IP (0x0800) */
				spARPHeader->u8ProtocolTypeHigh = HI(0x0800);
				spARPHeader->u8ProtocolTypeLow  = LO(0x0800);
				/*	HLEN: 6 */
				spARPHeader->u8Hlen = 6;
				/*	PLEN: 4 */
				spARPHeader->u8Plen = 4;
				/*	Operation: ARP-Request */
				spARPHeader->u8OperationLow		= LO(1);
				spARPHeader->u8OperationHigh	= HI(1);
				spARPHeader->sSenderIP			= myip;
				spARPHeader->sSenderMAC			= mymac;
				spARPHeader->sTargetIP			= spNode->sIP;
				spARPHeader->sTargetMAC			= zeromac;
			/*	Send the Request */
				ethernet_sendframe(spEthernetHeader,sizeof(ETHERNETHEADER)+sizeof(ARPHEADER));
			/*	Get next entry of the list */
			EnterCritical();	//CRITICAL SECTION
			spNode	= spNode->pNext;
			LeaveCritical();	//CRITICAL SECTION
		}//END Loop
	}//END Transmit
	
	if(gu8ArpRequest == 1){
	/*	An ARP-Response-Message should be send */
		EnterCritical();	//CRITICAL SECTION
	/*	Write the ethernet-Header, Protocol-Type ARP (0x0806) */
		spARPHeader = ethernet_writeheader(spEthernetHeader,&mymac,&gsArpMac,0x0806);
	/*	Write the ARP-Header */
		/*	Hardware: ETHERNET (1) */
		spARPHeader->u8HardwareTypeHigh = HI(1);
		spARPHeader->u8HardwareTypeLow  = LO(1);
		/*	Protocol-Type: IP (0x0800) */
		spARPHeader->u8ProtocolTypeHigh = HI(0x0800);
		spARPHeader->u8ProtocolTypeLow  = LO(0x0800);
		/*	HLEN: 6 */
		spARPHeader->u8Hlen = 6;
		/*	PLEN: 4 */
		spARPHeader->u8Plen = 4;
		/*	Type: ARP-Response (2)*/
		spARPHeader->u8OperationLow		= LO(2);
		spARPHeader->u8OperationHigh	= HI(2);
		/*	Return to Sender */
		spARPHeader->sTargetIP		= gsArpIp;
		spARPHeader->sTargetMAC		= gsArpMac;
		spARPHeader->sSenderIP		= myip;
		spARPHeader->sSenderMAC		= mymac;
	/*	ARP-Request handled */
		gu8ArpRequest	=	 0;
		LeaveCritical();	//CRITICAL S3ECTION
		/*	Send the frame over the ether */
		ethernet_sendframe(spEthernetHeader,(sizeof(ETHERNETHEADER)+sizeof(ARPHEADER)));
	}
}
/*****************************************************************************/
/* ARP-Packet received */
/*****************************************************************************/
void	arp_receivepackage(void){
/*	We got an ARP-Packet, analyse it */
void (*pCallback)(ARPQUERYNODE* pNode,char cCode);
ARPHEADER*		spARPHeader		= (void*)(((uchar*)&gu8NetworkBuffer)+sizeof(ETHERNETHEADER));
ARPQUERYNODE**	sppARPQueryNode;
ARPQUERYNODE*	spARPQueryNode;
u16				u16Size;
u16				u16Operation;
/*	Read the ARP-Header */
	u16Size	=	ethernet_readframe(spARPHeader,sizeof(ARPHEADER));
	if(u16Size != sizeof(ARPHEADER)){
		/*	The Packet doesn't contain a full header */
		return;
	}
/*	Check if everything's ok */
	/*	Hardware: ETHERNET (1) */
	if(spARPHeader->u8HardwareTypeHigh != HI(1))		return;
	if(spARPHeader->u8HardwareTypeLow  != LO(1))		return;
	/*	Protocol-Type: IP (0x0800) */
	if(spARPHeader->u8ProtocolTypeHigh != HI(0x0800))	return;
	if(spARPHeader->u8ProtocolTypeLow  != LO(0x0800))	return;
	/*	HLEN: 6 */
	if(spARPHeader->u8Hlen != 6)	return;
	/*	PLEN: 4 */
	if(spARPHeader->u8Plen != 4)	return;
/*	We got a valid Packet */
u16Operation	= spARPHeader->u8OperationLow + (((u16)spARPHeader->u8OperationHigh)<<8);
	switch(u16Operation){
		case	1:
/*---------------------------------------------------------------------------*/
/*	This is an ARP-Request */
			if(IPCOMPARE(spARPHeader->sTargetIP,myip)){
				if(gu8ArpRequest == 1){
					/*	There is an arp-request pending */
					break;
				}
				/*	This is for us: prepare an ARP-Response:
					If there are severall ARP-Requests until the Tx-Task
					handels this, only the last one will be responded! */
				EnterCritical();	//CRITICAL SECTION
				gsArpMac		=	spARPHeader->sSenderMAC;
				gsArpIp			=	spARPHeader->sSenderIP;
				gu8ArpRequest	=	1;
				LeaveCritical();	//CRITICAL SECTION
				/*	Activate the TX-Task */
				gu8NetworkSend = 1;
				break;
			}else{
				/*	The IP doesn't match my ip! */
				break;
			}
/*---------------------------------------------------------------------------*/
/*	This is an ARP-Response */
		case	2:

/*	We search in the list of pending arp-requests for a match */
	EnterCritical();		//CRITICAL SECTION
/*	Start at the Beginning */
	sppARPQueryNode		= &gspArpQueryQueue;
	spARPQueryNode		= 0;
	while(*sppARPQueryNode != 0){
		if(IPCOMPARE((*sppARPQueryNode)->sIP,spARPHeader->sSenderIP)){
			/*	The IPs match */
			spARPQueryNode	= *sppARPQueryNode;
			/* copy the MAC */
			spARPQueryNode->sMAC	= spARPHeader->sSenderMAC;
			/*	Remove the Item from the list */
				if((*sppARPQueryNode)->pNext != 0){
					(*sppARPQueryNode)	= ((*sppARPQueryNode)->pNext)->pNext;
				}else{
					(*sppARPQueryNode) = 0;
				}
			break;
		}
		/*	Get the next item */
		sppARPQueryNode	=	&((*sppARPQueryNode)->pNext);
	}
	//DANGER!!!! (Timer synchronisation....)
	LeaveCritical();		//CRITICAL SECTION
	/*	If an entry was found, call the callback */
	if(spARPQueryNode != 0){
		/*	Call the callback */
		pCallback	= spARPQueryNode->pCallback;
		(pCallback)(spARPQueryNode,ARP_QUERYSUCCESS);
	}
		break;
/*---------------------------------------------------------------------------*/
/*	Unsupported Request */
		default:
			break;
	}
		return;
}
/*****************************************************************************/
/* Simple User Interface: DO NOT USE INSIDE THE NETWORK TASK (uses Events!!!)*/
/*****************************************************************************/
char	arp_simplequery(ARPSIMPLEQUERY* pQueryNode){
//Init the event
		EVENTINIT(pQueryNode->sEvent);
//Use internal callback-function
		pQueryNode->pCallback		= arp_simplequerycallback;
//Make a normal arp-request
		arp_sendquery((ARPQUERYNODE*)pQueryNode);
//Wait for the Event to get signaled
		WaitEvent(&pQueryNode->sEvent);
//Return the status-code
		return	pQueryNode->u8Status;
}
/*---------------------------------------------------------------------------*/
void	arp_simplequerycallback(ARPQUERYNODE* pNode,u8 u8Code){
//Save the return code
((ARPSIMPLEQUERY*)pNode)->u8Status = u8Code;
//Signal the Event
		SignalEvent(&((ARPSIMPLEQUERY*)pNode)->sEvent);
}
/*****************************************************************************/

