/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	ARP */
/*****************************************************************************/
#ifndef __ARP
#define	__ARP
#include <stdlib.h>
#include "global.h"
#include "system.h"
#include "ethernet.h"
#include "ip-udp.h"
#include "network.h"
/*---------------------------------------------------------------------------*/
/* constants */
/*---------------------------------------------------------------------------*/
#define	ARP_QUERYTIMEOUT	500
#define	ARP_RETRY			5
#define	ARP_QUERYSUCCESS	1
#define	ARP_QUERYERROR		0
/*---------------------------------------------------------------------------*/
/* structs */
/*---------------------------------------------------------------------------*/
typedef struct _ARPQUERYNODE ARPQUERYNODE;
typedef struct	_ARPHEADER ARPHEADER;
typedef struct _ARPSIMPLEQUERY ARPSIMPLEQUERY;

struct	_ARPHEADER{
u8		u8HardwareTypeHigh;
u8		u8HardwareTypeLow;
u8		u8ProtocolTypeHigh;
u8		u8ProtocolTypeLow;
u8		u8Hlen;
u8		u8Plen;
u8		u8OperationHigh;
u8		u8OperationLow;
MACADDRESS	sSenderMAC;
IPADDRESS	sSenderIP;
MACADDRESS	sTargetMAC;
IPADDRESS	sTargetIP;
};

struct	_ARPQUERYNODE{
ARPQUERYNODE*	pNext;		//Next node to look up
IPADDRESS		sIP;		//IP which should be looked up
MACADDRESS		sMAC;		//Looked-up MAC
u8				u8Retry;	//Retry-Counter
void*			pCallback;	//Function to call when looked up or timeout
							//Function is of the following type:
							//void	(*pCallback)(ARPQUERYNODE* pNode,char cCode);
};


struct 	_ARPSIMPLEQUERY{
//first the normal query-structure
ARPQUERYNODE*	pNext;		//Next node to look up
IPADDRESS		sIP;		//IP which should be looked up
MACADDRESS		sMAC;		//Looked-up MAC
u8				u8Retry;	//Retry-Counter
void*			pCallback;	//Function to call when looked up or timeout
							//Function is of the following type:
							//void	(*pCallback)(ARPQUERYNODE* pNode,char cCode);
//now the extended fields
u8				u8Status;	//Status
EVENT			sEvent;		//Event used to signal the completness of the querry	
};

char	arp_getmacaddress(IPADDRESS* spIP,MACADDRESS* spMAC);
/*---------------------------------------------------------------------------*/
/* API */
/*---------------------------------------------------------------------------*/
void	arp_init(void);
void	arp_timer_callback(TIMER* pTimer);
void	arp_transmit(void);
void	arp_timercallback(TIMER* pTimer);
void	arp_sendquery(ARPQUERYNODE* pQueryNode);
void	arp_receivepackage(void);
//simple user interface, not to be used inside the Network-Task (uses events
//to block the task until the arp-request is finished!
char	arp_simplequery(ARPSIMPLEQUERY* pQueryNode);
//internal callback-function
void	arp_simplequerycallback(ARPQUERYNODE* pNode,u8 u8Code);
/*---------------------------------------------------------------------------*/
/* globals */
/*---------------------------------------------------------------------------*/
/*	ARP-Request from outside: They store the information for
	the response */
MACADDRESS		gsArpMac;
IPADDRESS		gsArpIp;
u8				gu8ArpRequest;
u8				gu8ArpTransmit;
/*	The Beginning of the arp-query-queue:
	This is a list of arp-queries */
ARPQUERYNODE*	gspArpQueryQueue;
ARPQUERYNODE**	gsppArpQueryQueueLast;
/*	Arp Timer */
TIMER			gsArpTimer;
/*****************************************************************************/
#endif
