/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	DHCP Functions */
/*****************************************************************************/
#include "dhcp.h"
/*---------------------------------------------------------------------------*/
void	dhcp_init(void){
/*	Setup the DHCP Services */
	gucDHCPState	= DHCP_INITIALIZE;
	EVENTINIT(gsDHCPReadyEvent);
}
/*---------------------------------------------------------------------------*/
/*	Initialize the dhcp-header:
	* All IPs are set to zero
	* Hardware-MAC is set to the own MAC
	* Message-Type: REPLY
	* Seconds / ID are both set to zero (not realy used)
	*/
void	dhcp_initreplyheader(void* pBuffer){
char		ci;
DHCPHEADER*	pDHCPHeader	= (DHCPHEADER*)pBuffer;
/*	Write zero into every field */
	for(ci=0;ci < sizeof(DHCPHEADER);ci++){
		((char*)pDHCPHeader)[(int)ci] = 0;
	}
/*	Type of Message: REQUEST=1 / REPLY=2 */
pDHCPHeader->ucOP					=	1;
pDHCPHeader->ucHtype				=	1;	//ETHERNET
pDHCPHeader->ucHlen					=	6;	//6 octet hw address
//pDHCPHeader->ucHops				=	0;	//Client has to write 0
//pDHCPHeader->ucTransactionID0-3	=	0;	//unique Transaction ID (set to zero)
//pDHCPHeader->ucSecondsLow/High	=	0;	//Seconds elapsed since the Client began
pDHCPHeader->ucFlagsHigh		=	HI(0x8000);	//use Broadcast
pDHCPHeader->ucFlagsLow			=	LO(0x8000);
//pDHCPHeader->sClientIP		=	zeroip;
//pDHCPHeader->sYourIP			=	zeroip;
//pDHCPHeader->sServerIP		=	zeroip;
//pDHCPHeader->sRouterIP		=	zeroip;
//pDHCPHeader->ucLowerHardwareAddress[0-9]	= 0;
pDHCPHeader->sMac				=	mymac;
}
/*---------------------------------------------------------------------------*/
/*	DHCP: Write the option-header */
void	dhcp_writeoptionheader(DHCPOPTIONHEADER* psHeader,uchar ucMessageID){
/*	Writes the Beginning of an DHCP-Header:
	*	The MagicCookie 
	*	The Message-TypeOption
	*/
/*	Write the so called "MagicCookie" */
	psHeader->ucMagicCookie[0]	=	0x63;
	psHeader->ucMagicCookie[1]	=	0x82;
	psHeader->ucMagicCookie[2]	=	0x53;
	psHeader->ucMagicCookie[3]	=	0x63;
/*	Write the MessageType Option */
	psHeader->ucDHCPMessageTypeOptionCode		= 53;
	psHeader->ucDHCPMessageTypeOptionLength		= 1;
	psHeader->ucDHCPMessageTypeOptionMessageID	= ucMessageID;
}
/*---------------------------------------------------------------------------*/
char	dhcp_checkmagiccookie(void){
/*	Check the Magic Cookie */
uchar	ucBuffer[4];
	if(ethernet_readframe(ucBuffer,4)!= 4)	goto error;
	if(ucBuffer[0] != 0x63)	goto	error;
	if(ucBuffer[1] != 0x82)	goto	error;
	if(ucBuffer[2] != 0x53)	goto	error;
	if(ucBuffer[3] != 0x63)	goto	error;
	return	0;
error:
	return	-1;
}
/*---------------------------------------------------------------------------*/
/*	DHCP: Send Message */
/*---------------------------------------------------------------------------*/
void	dhcp_sendmessage(MACADDRESS* spDestMAC, IPADDRESS* spDestIP, void* pDHCPMessage,uint uiLength){
	return;	//DANGER!!!!
/*	Assembles the header of the message and sends everything over the ether
	ATTENTION:	myip has to be zero whenever no ip is assigned by the
				dhcp-server!! This IP is used as the Client-IP in the
				DHCP-Header and as the Source-IP in the IP-Header */
uint			uiChecksum;
/*------- HEADER -------*/
ETHERNETHEADER*	spEthernetHeader	= (ETHERNETHEADER*)&gu8NetworkBuffer;
/*	Send the Message via Ethernet-Boradcast, Protocol: IP (0x0800) */
IPHEADER*		spIPHeader			=\
	ethernet_writeheader(spEthernetHeader,&mymac,spDestMAC,0x0800);
/*	Write the IP-Header: 
	Protokoll:		17 (UDP)
	Messagesize:	Header (IP,UDP,DHCP),two strings(64+128 byte), Body */	
UDPHEADER*		spUDPHeader		=\
	ip_writeheader(spIPHeader,17,\
		sizeof(UDPHEADER)+sizeof(DHCPHEADER)+64+128+uiLength,\
		&dhcpip,spDestIP,0);
/*	Write the UDP-Header:
	SourcePort:	68
	Dest.Port:	67
	Messagesize:	Header (UDP),two strings(64+128 byte), Body */	
DHCPHEADER*		spDHCPHeader		=\
	udp_writeheader(spUDPHeader,68,67,\
		sizeof(DHCPHEADER)+64+128+uiLength);
/*	Initialize the dhcp-header:
	* All IPs are set to zero
	* Hardware-MAC is set to the own MAC
	* Message-Type: REPLY
	* Seconds / ID are both set to zero (not realy used)
	*/
	dhcp_initreplyheader(spDHCPHeader);
	/*	The Client-IP is the IP of the lcoal machine:
		Must only be filled in if Client is in BOUND,RENEW or REBINDING state */
		//spDHCPHeader->sClientIP	= dhcpip !!!!;
/*------- CHECKSUM -------*/
/*	Calculate the Checksum for the UDP-Pseudo-Header */
//	uiChecksum	=	udp_pseudoheaderchecksum(&dhcpip,spDestIP,uiLength+128+64+sizeof(DHCPHEADER));
/*	Calculate the Checksum for the UDP/DHCP-Header */
//	uiChecksum	=	ip_checksum(uiChecksum,spUDPHeader,sizeof(UDPHEADER)+sizeof(DHCPHEADER));
/*	Calculate the Checksum for the zeros between the DHCP-Header and Message */
//	uiChecksum	=	ip_constchecksum(uiChecksum,0,(128+64)/2);
/*	Calculate the Checksum for the DHCP-Message */
//	uiChecksum	=	ip_checksum(uiChecksum,pDHCPMessage,uiLength);
/*	Write the Checksum into the Header
	ATTENTION: The final checksum is the 1-Complement of the Checksum!!! */
//	udp_writechecksum(spUDPHeader,~uiChecksum);
	udp_writechecksum(spUDPHeader,0);	//do not use the checksum!
/*------- SEND DATA -------*/
/*	Wait for the CS8900 to send the Data */
	ethernet_waittxready(sizeof(ETHERNETHEADER)+sizeof(IPHEADER)+sizeof(UDPHEADER)+sizeof(DHCPHEADER)+64+128+uiLength);
/*	Send the Header over the ether */
	ethernet_senddata(spEthernetHeader,sizeof(ETHERNETHEADER)+sizeof(IPHEADER)+sizeof(UDPHEADER)+sizeof(DHCPHEADER));
/*	Send the two strings (zeros only) */
	ethernet_sendbyte(0,128+64);
/*	Send the DHCP-Message */
	ethernet_senddata(pDHCPMessage,uiLength);
	ethernet_waittxdone();
}
/*****************************************************************************/
/* DHCP-Send function */
/*****************************************************************************/
void	dhcp_transmit(void){
/* DHCP-Send procedure */
void*	vpDHCPMessage		=	(void*)((char*)&gu8NetworkBuffer + sizeof(ETHERNETHEADER)+sizeof(IPHEADER)+sizeof(UDPHEADER)+ sizeof(DHCPHEADER));
uchar	ucDHCPState;
/* Get the current status code */
	EnterCritical();
	ucDHCPState		= gucDHCPState;
	LeaveCritical();
/*---------------------------------------------------------------------------*/	
/* DHCP Finit-State Machine */
	switch(ucDHCPState){
/*---------------------------------------------------------------------------*/	
/* DHCP-Initialize */
		case DHCP_INITIALIZE:
		/*	Send a Discover-Message */
				dhcp_writeoptionheader((DHCPOPTIONHEADER*)vpDHCPMessage,DHCP_DISCOVER_MESSAGE);
		//specify an ip we want to get
				((DHCPDISCOVERMESSAGE*)vpDHCPMessage)->sRequestIpOption.ucDHCPMessageTypeOptionCode		= 50;
				((DHCPDISCOVERMESSAGE*)vpDHCPMessage)->sRequestIpOption.ucDHCPMessageTypeOptionLength	= 4;
				((DHCPDISCOVERMESSAGE*)vpDHCPMessage)->sRequestIpOption.sIP								= myip;
		//Write Termination
				((DHCPDISCOVERMESSAGE*)vpDHCPMessage)->ucTermination = 0xFF;
		//Send the Message
				dhcp_sendmessage(&broadmac,&broadip,vpDHCPMessage,sizeof(DHCPDISCOVERMESSAGE));
		//Change the state:
				EnterCritical();
				gucDHCPState = DHCP_SELECT;
				LeaveCritical();
		break;
/*---------------------------------------------------------------------------*/	
/* DHCP-OFFERRECEIVED */
		case DHCP_OFFERRECEIVED:
		/* The server has overed us an ip, we accept it: Send a Request-Message */
		/*	It's an REQUEST (ID 3) */
			dhcp_writeoptionheader((DHCPOPTIONHEADER*)vpDHCPMessage,3);
		/*	Server Identifier */
			((DHCPREQUESTMESSAGE*)vpDHCPMessage)->ucDHCPOptionCode0		=	54;
			((DHCPREQUESTMESSAGE*)vpDHCPMessage)->ucDHCPOptionLength0	=	sizeof(IPADDRESS);
			((DHCPREQUESTMESSAGE*)vpDHCPMessage)->sIP0					=	dhcpserverip;
		/*	Requested IP address */
			((DHCPREQUESTMESSAGE*)vpDHCPMessage)->ucDHCPOptionCode1		=	50;
			((DHCPREQUESTMESSAGE*)vpDHCPMessage)->ucDHCPOptionLength1	=	sizeof(IPADDRESS);
			((DHCPREQUESTMESSAGE*)vpDHCPMessage)->sIP1					=	dhcpip;
		/*	Termination */
			((DHCPREQUESTMESSAGE*)vpDHCPMessage)->ucTermination			=	255;
		/*	Send the Message */
			dhcp_sendmessage(&broadmac,&broadip,vpDHCPMessage,sizeof(DHCPREQUESTMESSAGE));
		/*	Change State to REQUEST */
			EnterCritical();
			gucDHCPState	=	DHCP_REQUEST;
			LeaveCritical();
				break;
/*---------------------------------------------------------------------------*/	
/* Some other status, dont send anything */
		default:
		break;
	}
}
/*****************************************************************************/
/* DHCP Receive function */
/*****************************************************************************/
void	dhcp_receivepackage(void){
/*	The currently proccessed package is a DHCP-Package. Check it out.*/
uchar		ucOption[2];
uint		uiLength;
uint		uiReadLength;
uchar		ucDHCPMessageType	= 0;
IPADDRESS	sSubNet;
uchar		ucSubNetValid	= 0;
IPADDRESS	sRouter;
uchar		ucRouterValid	= 0;
IPADDRESS	sNameserver;
uchar		ucNameserverValid	= 0;
uchar		ucDHCPState;
DHCPHEADER*	spDHCPHeader		= (void*)(&gu8NetworkBuffer);

		/*	Read the DHCP-Header in */
		uiReadLength	=	ethernet_readframe(spDHCPHeader,sizeof(DHCPHEADER));
		if(uiReadLength != sizeof(DHCPHEADER)){
			/*	The received packet doesn't contail a full DHCP-Header - drop */
			return;
		}
		/*	We have an DHCP-Header, process it */
		/*	Verify that it's an reply (code 2) */
		if(spDHCPHeader->ucOP != 2){
			return;
		}
		/*	Verify that the Hardware is an Ethernet */
		if(spDHCPHeader->ucHtype != 1){
			return;
		}
		/*	Verify that the Address is 6 octets long */
		if(spDHCPHeader->ucHlen != 6){
			return;
		}
/*---------------------------------------------------------------------------*/	
/*	Process the DHCP-Package */
/*	Drop the long field in the middle */
	ethernet_dropframecontents(64+128);
//Check for the magic Cookie
	if(dhcp_checkmagiccookie()!=0)	return;
//Read the options in:
		while(1){
			uiLength	= ethernet_readframe(ucOption,2);	
			//test for the end option
			if((uiLength == 1) && (ucOption[0] == 255)) break;
			if(uiLength != 2)	return;
				switch(ucOption[0]){
					case 53:		//DHCP-Message-Type option
						if(ucOption[1]!=1)											return;
						if(ethernet_readframe(&ucDHCPMessageType,1) != 1)			return;
						ucOption[1]	= 0;	//option completly read in
						break;						
					case 1:			//SUBNET-Option (l=4)
						if(ucOption[1]!=4)								return;
						if(ethernet_readframe(&sSubNet,4) != 4)			return;
						ucSubNetValid	= 1;	//option found
						ucOption[1]		= 0;	//option completly read in
						break;						
					case 3:			//Router-Option (l=4*n, routers listed in order of preference)
						if(ucOption[1]<4)								return;
						if(ethernet_readframe(&sRouter,4) != 4)			return;
						ucRouterValid	= 1;	//option found
						ucOption[1]		-= 4;	//four bytes read in
						break;
					case 5:			//Nameserver-Option (l=4*n, servers listed in order of preference)
						if(ucOption[1]<4)								return;
						if(ethernet_readframe(&sNameserver,4) != 4)		return;
						ucNameserverValid	= 1;	//option found
						ucOption[1]			-= 4;	//four bytes read in
						break;
				}
			//Advance to the next option
			if(ethernet_dropframecontents(ucOption[1]) != ucOption[1]) 		return;
		}
/*---------------------------------------------------------------------------*/	
/* DHCP Finit-State Machine */
/* Get the current status code */
	EnterCritical();
	ucDHCPState		= gucDHCPState;
	LeaveCritical();

	switch(gucDHCPState){
		case	DHCP_INITIALIZE:
		/*	This is the initialize-state, no request has been send */
		/*	In this State we don't expect a DHCP-Message, drop */	
			return;
			break;
		case	DHCP_SELECT:
/*---------------------------------------------------------------------------*/	
/*	SELECT-State: Wait for an DHCPOFFER-Message */
			if(ucDHCPMessageType == DHCP_OFFER_MESSAGE){
				EnterCritical();
				/* We save all aquired data */
				if(ucSubNetValid != 0){	subnetmask = sSubNet;}
				if(ucRouterValid != 0){	routerip = sRouter;}
				if(ucNameserverValid != 0){	nameserverip = sSubNet;}
				/* The IP-Address the server is willing to assign
					is written in the header */
				dhcpip			= spDHCPHeader->sYourIP;
				/* We have to remember which dhcp-server has offered us an ip */
				dhcpserverip	= spDHCPHeader->sServerIP;
				/*	Change the state: send REQUEST-Message now */
				gucDHCPState = DHCP_OFFERRECEIVED;
				/*  Ensure that we send something back */
				gu8NetworkSend = 1;
				LeaveCritical();
			}
			break;
		case	DHCP_REQUEST:
/*---------------------------------------------------------------------------*/	
/*	DHCP_REQUEST-State: Wait for an DHCPACK */
		if(ucDHCPMessageType == DHCP_ACK_MESSAGE){
			/* Received an Acknowledge */
			EnterCritical();
			myip	= dhcpip;
			/* Change state to bound */
			gucDHCPState	= DHCP_BOUND;		
			LeaveCritical();
			/*	Signal that we have finished the configuration */
			SignalEvent(&gsDHCPReadyEvent);
		}
			break;
		case	DHCP_BOUND:
		/*	We are bound to an IP address */
			break;
		default:
			break;
	}
		return;
}
/*****************************************************************************/

