/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	DHCP */
/*****************************************************************************/
#ifndef __DHCP
#define	__DHCP
#include <stdlib.h>
#include "global.h"
#include "system.h"
#include "network.h"
#include "ethernet.h"
#include "ip-udp.h"
#include "arp.h"
/*---------------------------------------------------------------------------*/
/* DHCP-FinitState-States: DHCP-Client */
/*	INITIALIZE-State: Client sends DISCOVER-Messages */
#define	DHCP_INITIALIZE		0
/* Client has send a Discover-Message and Server has responded */
#define	DHCP_OFFERRECEIVED	10
/*	TX-Task should send an Discover-Message */
#define	DHCP_SENDDISCOVER	100
/*	SELECT-State: Wait for an OFFER-Message */
#define	DHCP_SELECT			1
/*	REQUEST-State: Client sends REQUEST-Messages and
					waits for Acknowledgments */
#define	DHCP_REQUEST		2
/*	TX-Task should send an Request-Message */
#define	DHCP_SENDREQUEST	102
/*	BOUND-State: Client got it's ip */
#define	DHCP_BOUND			3
/*---------------------------------------------------------------------------*/
#define	DHCP_DISCOVER_MESSAGE	1
#define	DHCP_OFFER_MESSAGE		2
#define	DHCP_REQUEST_MESSAGE	3
#define	DHCP_DECLINE_MESSAGE	4
#define	DHCP_ACK_MESSAGE		5
#define	DHCP_NACK_MESSAGE		6
#define	DHCP_RELEASE_MESSAGE	7


typedef	struct	_DHCPMESSAGE DHCPMESSAGE;
typedef struct	_DHCPHEADER DHCPHEADER;
typedef	struct	_DHCPOPTIONHEADER DHCPOPTIONHEADER;
typedef	struct	_DHCPMESSAGETYPEOPTION DHCPMESSAGETYPEOPTION;
typedef	struct	_DHCPREQUESTMESSAGE DHCPREQUESTMESSAGE;
typedef	struct	_DHCPDISCOVERMESSAGE DHCPDISCOVERMESSAGE;
typedef struct	_DHCPREQUESTIPOPTION DHCPREQUESTIPOPTION;

struct	_DHCPHEADER{
uchar		ucOP;
uchar		ucHtype;
uchar		ucHlen;
uchar		ucHops;
uchar		ucTransactionID3;
uchar		ucTransactionID2;
uchar		ucTransactionID1;
uchar		ucTransactionID0;
uchar		ucSecondsHigh;
uchar		ucSecondsLow;
uchar		ucFlagsHigh;
uchar		ucFlagsLow;
IPADDRESS	sClientIP;
IPADDRESS	sYourIP;
IPADDRESS	sServerIP;
IPADDRESS	sRouterIP;
MACADDRESS	sMac;
uchar		ucLowerHardwareAddress[10];
};

struct	_DHCPMESSAGE{
DHCPHEADER	sHeader;
uchar		ucServerHostName[64];
uchar		ucBootFileName[128];
};

struct		_DHCPOPTIONHEADER{
uchar		ucMagicCookie[4];
uchar		ucDHCPMessageTypeOptionCode;
uchar		ucDHCPMessageTypeOptionLength;
uchar		ucDHCPMessageTypeOptionMessageID;
};

struct	_DHCPMESSAGETYPEOPTION{
uchar		ucDHCPMessageTypeOptionCode;
uchar		ucDHCPMessageTypeOptionLength;
uchar		ucDHCPMessageTypeOptionMessageID;
};

struct	_DHCPREQUESTIPOPTION{
uchar		ucDHCPMessageTypeOptionCode;
uchar		ucDHCPMessageTypeOptionLength;
IPADDRESS	sIP;
};

struct _DHCPREQUESTMESSAGE{
/*	HEADER	*/
DHCPOPTIONHEADER	sHeader;
/*	Server Identifier */
uchar				ucDHCPOptionCode0;
uchar				ucDHCPOptionLength0;
IPADDRESS			sIP0;
/*	Requested IP Address */
uchar				ucDHCPOptionCode1;
uchar				ucDHCPOptionLength1;
IPADDRESS			sIP1;
/*	Termination */
uchar				ucTermination;
};

struct _DHCPDISCOVERMESSAGE{
/*	HEADER	*/
DHCPOPTIONHEADER	sHeader;
DHCPREQUESTIPOPTION	sRequestIpOption;
/*	Termination */
uchar				ucTermination;
};
/*---------------------------------------------------------------------------*/
/*	API */
/*---------------------------------------------------------------------------*/
void	dhcp_init(void);
/* send/receive function for network-task */
void	dhcp_receivepackage(void);
void	dhcp_transmit(void);
/* internal stuff */
void	dhcp_sendmessage(MACADDRESS* spDestMAC, IPADDRESS* spDestIP, void* pDHCPMessage,uint uiLength);
void	dhcp_writeoptionheader(DHCPOPTIONHEADER* psHeader,uchar ucMessageID);
void	dhcp_initreplyheader(void* pBuffer);
/*---------------------------------------------------------------------------*/
/*	Globals */
/*---------------------------------------------------------------------------*/
uchar		gucDHCPState;
EVENT		gsDHCPReadyEvent;
//ulong		gulDHCPTransaction;
/*---------------------------------------------------------------------------*/
#endif
