/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
/*	If you get a warning from delay.x, ensure that the
	CPU frequency is set in the AVR-Studio Project settings */
#include <util/delay.h>
#include "global.h"
#include "inttypes.h"
#include "spi.h"
#include "enc28j60.h"


void	enc_reset(void){
	/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
	/*	Setup the SPI-Speed */
	enc_spiconfig();
	/*	Make the CS-Pin low */
	ENC_DESELECT_CS
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	/*	System Reset */
		spi_rw(0xFF);
	/*	Make CS-Pin high again */
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	ENC_SELECT_CS
	/*	Release the bus */
	SignalEvent(&gsSpiLockEvent);
}


void	enc_bitfieldset_lowlevel(u8 u8Address,u8 u8Data){
/* 	WE HAVE TO OWN THE SPI-BUS!!! */
	/*	Make the CS-Pin low */
	ENC_DESELECT_CS
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	/*	Write Instruction : Consists of a 3-Bit Opcode (100: BitFieldSet) and 
	 * 	the 5-Bit address */
		spi_rw(0x80 | (0x1F & u8Address));
	/*	Write the Data */
		spi_rw(u8Data);
	/*	Make CS-Pin high again */
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	ENC_SELECT_CS
}

void	enc_bitfieldclear_lowlevel(u8 u8Address,u8 u8Data){
/* 	WE HAVE TO OWN THE SPI-BUS!!! */
	/*	Make the CS-Pin low */
	ENC_DESELECT_CS
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	/*	Write Instruction : Consists of a 3-Bit Opcode (101: BitFieldClear) and 
	 * 	the 5-Bit address */
		spi_rw(0xA0 | (0x1F & u8Address));
	/*	Write the Data */
		spi_rw(u8Data);
	/*	Make CS-Pin high again */
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	ENC_SELECT_CS
}

u8	enc_getbank(void){
/*	We have to own the SPI-Bus */
u8	u8Tmp;
/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
	/*	Setup the SPI-Speed */
	enc_spiconfig();
	u8Tmp	= enc_readcontrolregister_lowlevel(ENC_ECON1);
	/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
	return		u8Tmp;
}


void	enc_selbank(u8 u8Address){
/*	We have to own the SPI-Bus */
u8	u8Tmp;
/*	Adjust bits BSEL1:BSEL0 from the ECON1-Register */
	u8Tmp	= (0x03 & u8Address);
	enc_bitfieldset_lowlevel(ENC_ECON1,u8Tmp);
	u8Tmp	= (~u8Tmp) & 0x03;
	enc_bitfieldclear_lowlevel(ENC_ECON1,u8Tmp);
}


void	enc_bitfieldset(u8 u8Address,u8 u8Data){
/*	Makes a logical OR between the Data and the given Register
 *  REMARK: the lower 5 bits specify the register address,
 * 	the higher 2 the bank address. bit.7 is unused */
/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
	/*	Setup the SPI-Speed */
	enc_spiconfig();
/*	Adjust bits BSEL1:BSEL0 from the ECON1-Register */
	/*	Adjust bits BSEL1:BSEL0 from the ECON1-Register */
	enc_selbank((0x60 & u8Address) >> 5);
	/*	Now write to the register */
	enc_bitfieldset_lowlevel(u8Address,u8Data);
/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
}

void	enc_bitfieldclear(u8 u8Address,u8 u8Data){
/*	Makes a logical Notand between the Data and the given Register
 *  REMARK: the lower 5 bits specify the register address,
 * 	the higher 2 the bank address. bit.7 is unused */
/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
	/*	Setup the SPI-Speed */
	enc_spiconfig();
/*	Adjust bits BSEL1:BSEL0 from the ECON1-Register */
	enc_selbank((0x60 & u8Address) >> 5);
	/*	Now write to the register */
	enc_bitfieldclear_lowlevel(u8Address,u8Data);
/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
}


void	enc_spiconfig(void){
/*	Internal Function, used to setup the SPI speed, etc */
	/*	The enc28j60 needs the following configuration:
			SPIE = 0: No Interrupts used so far
			SPE	= 1: SPI enabled
			DORD = 0: MSB first 
			MSTR = 1: Master Operation
			CPOL = 0: SCK low when idle
			CPHA = 0: Sample on the leading edge
			(SPI2X = 1): Choose fosc/2 as the SCK-frequency
			SPR1 = 0
			SPR0 = 0 */
//			SPCR0	= 	_BV(SPE0) | _BV(MSTR0) | _BV(SPR0);
			SPCR0	= 	_BV(SPE0) | _BV(MSTR0);
	/*	Set SPI2X = 1!! (diffrent register!)  */
			SPSR0	= _BV(SPI2X);
}
/*
Read Buffer Memory     0 0 1 1 1 0 1 0
(RBM)
Write Control Register 0 1 0 a a a a a
(WCR)
Write Buffer Memory    0 1 1 1 1 0 1 0
(WBM)*/
/*
The RBM command is started by pulling the CS pin low.
The RBM opcode is then sent to the ENC28J60,
followed by the 5-bit constant, 1Ah. After the RBM com-
mand and constant are sent, the data stored in the
memory pointed to by ERDPT will be shifted out MSb
first on the SO pin. If the host controller continues to
provide clocks on the SCK pin, without raising CS, the
byte pointed to by ERDPT will again be shifted out MSb
first on the SO pin. In this manner, with AUTOINC
enabled, it is possible to continuously read sequential
bytes from the buffer memory without any extra SPI
command overhead. The RBM command is terminated
by raising the CS pin.
*/


void	enc_writecontrolregister_lowlevel(u8 u8Address,u8 u8Data){
/*	Writes a control-register, no bank-select is done,
 * 	WE HAVE TO OWN THE SPI-BUS!!! */
	/*	Make the CS-Pin low */
	ENC_DESELECT_CS
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	/*	Write Instruction : Consists of a 3-Bit Opcode (010: Write control register) and 
	 * 	the 5-Bit address */
		spi_rw(0x40 | (0x1F & u8Address));
	/*	Write the Data */
		spi_rw(u8Data);
	/*	Make CS-Pin high again */
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
	ENC_SELECT_CS
}
u8	enc_readcontrolregister_lowlevel(u8 u8Address){
/*	Reads a control-register, no bank-select is done,
 * 	WE HAVE TO OWN THE SPI-BUS!!! */
	/*	DANGER:	If the address specifies one of the MAC or MII registers,
		a dummy byte will first be shifted out on the SO pin.
		After the dummy byte, the data will be shifted out MSb
		first on the SO pin. The RCR operation is terminated by
		raising the CS pin.
	- Bit.7 of u8Address 1: Read Dummy-Byte, 0: Do not read Dummy-Byte*/ 
u8	u8Data;
	/*	Make the CS-Pin low */
	ENC_DESELECT_CS
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	/*	Write Instruction : Consists of a 3-Bit Opcode (000: Read control register) and 
	 * 	the 5-Bit address */
	spi_rw(0x00 | (0x1F & u8Address));
	/*	Read the Data */
	u8Data	= 	spi_rw(0);
	if((u8Address & 0x80)!=0){
		/*	We should read one dummy-byte */
		u8Data	= 	spi_rw(0);
	}
	/*	Make CS-Pin high again */
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	ENC_SELECT_CS
	return	u8Data;
}

u8	enc_readcontrolregister(u8 u8Address){
/*	Returns the contents of the control-register given by u8Address */
/*	DANGER:	If the address specifies one of the MAC or MII registers,
	a dummy byte will first be shifted out on the SO pin.
	After the dummy byte, the data will be shifted out MSb
	first on the SO pin. The RCR operation is terminated by
	raising the CS pin. 
	- Bit.7 of u8Address 1: Read Dummy-Byte, 0: Do not read Dummy-Byte*/ 
u8	u8Data;
/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
/*	Setup the SPI-Speed */
	enc_spiconfig();
/*	Adjust bits BSEL1:BSEL0 from the ECON1-Register */
	enc_selbank((0x60 & u8Address) >> 5);
/*	Now read the register */
u8Data	=	enc_readcontrolregister_lowlevel(u8Address);	
/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
	return	u8Data;
}

void	enc_writecontrolregister(u8 u8Address,u8 u8Data){
/*	Writes the contents to the control-register given by u8Address
 *  REMARK: the lower 5 bits specify the register address,
 * 	the higher 2 the bank address. bit.7 is unused */
/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
	/*	Setup the SPI-Speed */
	enc_spiconfig();
/*	Adjust bits BSEL1:BSEL0 from the ECON1-Register */
	enc_selbank((0x60 & u8Address) >> 5);
/*	Now write to the register */
	enc_writecontrolregister_lowlevel(u8Address,u8Data);	
/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
}


void	enc_writephyregister(u8 u8Address,u16 u16Data){
/*	Writes data to one of the enc28j60 phy-registers */
	/*	Wait for the BUSY-Bit in the MISTAT-Register to get cleared */
	while((enc_readcontrolregister(ENC_MISTAT) & 0x01) != 0);
	/*	First setup the address of the PHY-Register */
	enc_writecontrolregister(ENC_MIREGADR,u8Address);
	/*	Write FIRST the low-Byte, than the high-Byte (Data are
	 * 	transfered to the register after the high-byte is written */
	enc_writecontrolregister(ENC_MIWRL,(u8)u16Data);
	enc_writecontrolregister(ENC_MIWRH,(u8)(u16Data>>8));
}

u16		enc_readphyregister(u8 u8Address){
/*	Writes data to one of the enc28j60 phy-registers */
u8	u8tmp;
u16	u16data;
	/*	Wait for the BUSY-Bit in the MISTAT-Register to get cleared */
	while((enc_readcontrolregister(ENC_MISTAT) & 0x01) != 0);
	/*	First setup the address of the PHY-Register */
	enc_writecontrolregister(ENC_MIREGADR,u8Address);
	/*	Set the MIIRD-Bit (bit.0) in the MICMD Register to start the read operation */
	u8tmp	= enc_readcontrolregister(ENC_MICMD);
	u8tmp	|=	0x01;	
	enc_writecontrolregister(ENC_MICMD,u8tmp);
	/*	We have to wait around 10.24us - Poll Busy-Bit */
	while((enc_readcontrolregister(ENC_MISTAT) & 0x01) != 0);
	/*	clear the MIIRD-Bit */
	u8tmp	= enc_readcontrolregister(ENC_MICMD);
	u8tmp	&=	0xFE;	
	enc_writecontrolregister(ENC_MICMD,u8tmp);
	/*	Read the data */
	u16data		= enc_readcontrolregister(ENC_MIRDH) << 8;
	u16data		|= enc_readcontrolregister(ENC_MIRDL);
	return	u16data;
}



