/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	Ethernet-Stuff */
/*****************************************************************************/
#include "ethernet.h"
#include <util/delay.h>
#include <avr/io.h>
/*****************************************************************************/
/*	Initialize */
/*****************************************************************************/
void	ethernet_init(void){
u16	u16tmp;
u8	u8tmp;
u16	u16erg0;
u16	u16erg1;
/*	Setup all MAC-Addressed */
	SETMACADDRESS(mymac,MYMAC0,MYMAC1,MYMAC2,MYMAC3,MYMAC4,MYMAC5);
	SETMACADDRESS(debugmac,DEBUGMAC0,DEBUGMAC1,DEBUGMAC2,DEBUGMAC3,DEBUGMAC4,DEBUGMAC5)
	SETMACADDRESS(broadmac,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF);
	SETMACADDRESS(zeromac,0,0,0,0,0,0);
/*	Setup the hardware */
	//-----------------------------------------------------------------------------
	/*	Setting up the ENC28J60:
	 * 	This is done according to the steps written in the datasheet chapter 6 */
	/*	Configure the receive-Buffer */

	DDRD	|= _BV(4);
	PORTD	|= _BV(4);
	_delay_ms(2);
	PORTD	&= ~_BV(4);
	_delay_ms(2);
	PORTD	|= _BV(4);
	_delay_ms(2);

	
	enc_reset();
	_delay_ms(2);
	
	
	
	
	
	
	/*	When the BIST is used, the ECON1
		register’s DMAST, RXEN and TXRTS bits should all be
		clear.	*/
		u8tmp	= enc_readcontrolregister(ENC_ECON1);
		enc_writecontrolregister(ENC_ECON1,0x00);
		u8tmp	= enc_readcontrolregister(ENC_ECON1);
	/*	To use the BIST:
		1.  Program the EDMAST register pair to 0000h. */
		enc_writecontrolregister(ENC_EDMASTL,0x00);
		enc_writecontrolregister(ENC_EDMASTH,0x00);
	/*	2.  Program EDMAND and ERXND register pairs to
		    1FFFh. */
		enc_writecontrolregister(ENC_EDMANDL,0xFF);
		enc_writecontrolregister(ENC_EDMANDH,0x1F);
		enc_writecontrolregister(ENC_ERXNDL,0xFF);
		enc_writecontrolregister(ENC_ERXNDH,0x1F);

		enc_writecontrolregister(ENC_ERXSTL,0x00);
		enc_writecontrolregister(ENC_ERXSTH,0x00);
		/*	It is recommended to programm the ERXRDPT register as well. The low
		 * 	byte has to be written first (6.1)  (The RXWRPT gets updated automatically 
		 * 	See Errata!!! */
			enc_writecontrolregister(ENC_ERXRDPTL,0xFF);
			enc_writecontrolregister(ENC_ERXRDPTH,0x1F);
	/*	3.  Configure the DMA for checksum generation by
		    setting CSUMEN in ECON1. */
		enc_bitfieldset(ENC_ECON1,0x10);
		/*	bit.4: CSUMEN */
	/*	4.  Write the seed/initial shift value byte to the
		    EBSTSD register (this is not necessary if
		    Address Fill mode is used). */
		enc_writecontrolregister(ENC_EBSTSD,0x55);
	/*	5.  Enable Test mode, select the desired test, select
		    the desired port configuration for the test. */
		/* bit 7-5 PSV2:PSV0: Pattern Shift Value bits
		        When TMSEL<1:0> = 10:
		        The bits in EBSTSD will shift left by this amount after writing to each memory location.
		        When TMSEL<1:0> = 00, 01 or 11:
		        This value is ignored.
		bit 4   PSEL: Port Select bit
		        1 = DMA and BIST modules will swap ports when accessing the memory
		        0 = Normal configuration
		bit 3-2 TMSEL1:TMSEL0: Test Mode Select bits
		        11 = Reserved
		        10 = Pattern shift fill
		        01 = Address fill
		        00 = Random data fill
		bit 1   TME: Test Mode Enable bit
		        1 = Enable Test mode
		        0 = Disable Test mode
		bit 0   BISTST: Built-in Self-Test Start/Busy bit
		        1 = Test in progress; cleared automatically when test is done
		        0 = No test running
		 */
		enc_writecontrolregister(ENC_EBSTCON,0x1A);
		/*	All bits zero, except bit.1 (TestModeEnable) */
	/*	6.  Start the BIST by setting EBSTCON.BISTST. */
		enc_bitfieldset(ENC_EBSTCON,0x01);
		/*	Set bit.0: BISTST */
		do{
			u8tmp	= enc_readcontrolregister(ENC_EBSTCON);
		}while((u8tmp & 0x01) != 0);

		u16erg1	= enc_readcontrolregister(ENC_EBSTCSL);
		u16erg1	|= ((u16)enc_readcontrolregister(ENC_EBSTCSH))<<8;
		enc_writecontrolregister(ENC_EBSTCON,0x00);
		
	/*	7.  Start the DMA checksum by setting DMAST in
		    ECON1. The DMA controller will read the
		    memory at the same rate the BIST controller will
		    write to it, so the DMA can be started any time
		    after the BIST is started.*/

		u8tmp	= enc_readcontrolregister(ENC_EIR);

		
		enc_writecontrolregister(ENC_ECON1,0x30);
		/*	Set bit.5 DMAST */
	/*	8.  Wait for the DMA to complete by polling the
		    DMAST bit or receiving the DMA interrupt (if
		    enabled). */
		do{
			u8tmp	= enc_readcontrolregister(ENC_EIR);
			u8tmp	= enc_readcontrolregister(ENC_ECON1);
		}while((u8tmp & 0x20) != 0);
	/*	9.  Compare the EDMACS registers with the
		    EBSTCS registers. */
		u16erg0	= enc_readcontrolregister(ENC_EDMACSL);
		u16erg0	|= ((u16)enc_readcontrolregister(ENC_EDMACSL))<<8;

				
		
	
	
	
	
		enc_reset();
		_delay_ms(2);
		
	
		enc_writecontrolregister(ENC_ECON1,0x00);

		/*	Configure the Packet Filter */
			enc_writecontrolregister(ENC_ERXFCON,0xA3);
		/*	Set	bit.7:UCEN: Unicast Filter Enable bit
		 * 	Set bit.5:CRCEN Invalid CRC: drop
			Set bit.1:MCEN Multicast: ok
			Set bit.0:BCEN disabled: ok */
		
		enc_writecontrolregister(ENC_ERXSTL,(u8)ENC_RXST);
		enc_writecontrolregister(ENC_ERXSTH,(u8)(ENC_RXST>>8));
		enc_writecontrolregister(ENC_ERXNDL,(u8)ENC_RXND);
		enc_writecontrolregister(ENC_ERXNDH,(u8)(ENC_RXND>>8));


	/*	It is recommended to programm the ERXRDPT register as well. The low
	 * 	byte has to be written first (6.1)  (The RXWRPT gets updated automatically 
	 * 	See Errata!!! */
		enc_writecontrolregister(ENC_ERXRDPTL,(u8)ENC_RXND);
		enc_writecontrolregister(ENC_ERXRDPTH,(u8)(ENC_RXND>>8));
	/*	Setup the Read-Pointer */
		gu16EthernetNextRxFrame	= ENC_RXST;
	/*	Every unused memory can be used as Transmit-Buffer */
	/*	Danger: The chip writes a seven byte status code AFTER the last byte
	 *  in the packet!!!! There has to be enough space */	
	//-----------------------------------------------------------------------------
	/*	Wait for OST (Oscillator Start Up): Poll the CLKRDY-Bit in the ESTAT-Register:
	 * 	This has to be done before any MAC and PHY Registers are accessed (see section 2.2) */
	/*	The ENC28J60 contains an Oscillator Start-up Timer
	 	OST) to ensure that the oscillator and integrated PHY
		have stabilized before use. The OST does not expire
		until 7500 OSC1 clock cycles (300 μs) pass after
		Power-on Reset or wake-up from Power-Down mode
		occurs. During the delay, all Ethernet registers and
		buffer memory may still be read and written to through
	 	the SPI bus. However, software should not attempt to
	 	transmit any packets (set ECON1.TXRTS), enable
	 	reception of packets (set ECON1.RXEN) or access any
		MAC, MII or PHY registers during this period. */
	/*	Poll the Bit: wait until it gets set */
	//	while((enc_readcontrolregister(ENC_ESTAT) & 0x01) == 0);
	//-----------------------------------------------------------------------------
	/*	Setup the MAC */
		/*	Enable MAC to receive FRAMEs:
		 * 	Set the MARXEN Bit in the MACON1 register
		 * Set TXPAUS and RXPAUS to allow IEEE-Flow-control (we want full duplex mode ( */
		enc_writecontrolregister(ENC_MACON1,0x01 | 0x04 | 0x08);
		/*	Set bit.0: MARXEN
		 * 	Set bit.2: RXPAUS
		 *	Set bit.3: TXPAUS */
		/*	Configure MACON3 */
		enc_writecontrolregister(ENC_MACON3,0xE0 | 0x02 |0x01);
		/*	Set bit.7-5: Framepadding to 64 Bytes, CRC will be appended
		 * 	Set bit.1: FRMLNEN: Frame type/length field will be checked
		 * 	in the case it represents a length, this will be compared and reported
		 * 	Set bit.0: FULDPX: Enable Full Duplex-Mode (DANGER:  PHCON1.PDPXMD  must be set as well)*/ 
		/*	Configure MACON4 */
		enc_writecontrolregister(ENC_MACON4,0x40);
		/*	Set bit.6: DEFER (MAC will wait until the Medium is free indefinitely (needed for IEEE802.3 compliance)) */
		/*	Setup the MAMXFL-Register (Maximum frame length) */
		enc_writecontrolregister(ENC_MAMXFLL,(u8)ENC_MAXFRAMELENGTH);
		enc_writecontrolregister(ENC_MAMXFLH,(u8)(ENC_MAXFRAMELENGTH>>8));
		/*	Setup the MABBIPG-Register: FullDuplex (0x15) Bit.0-6: InterPacketGap (Fullduplex
		 * 	enabled, 0x15: 9.6us, minimum IEEE specifed time */
		enc_writecontrolregister(ENC_MABBIPG,0x15);
		/*	Setup the MAIPGL-Register: Should be set to 0x12, did not found what it means */
		enc_writecontrolregister(ENC_MAIPGL,0x12);
		/*	Steps 7 and 8 are only required if Half-Duplex Mode is used */
		/*	Setup the MAIPGH-Register: Should be set to 0x0C, did not found what it means */
	//	enc_writecontrolregister(ENC_MAIPGH,0x0C);
		/*	Values of MACLCON1/2 doesnt have to be changed if distances are normal, see datasheet */
		/*	Setup the MAC-Address */
		enc_writecontrolregister(ENC_MAADR1,mymac.u8Address[0]);
		enc_writecontrolregister(ENC_MAADR2,mymac.u8Address[1]);
		enc_writecontrolregister(ENC_MAADR3,mymac.u8Address[2]);
		enc_writecontrolregister(ENC_MAADR4,mymac.u8Address[3]);
		enc_writecontrolregister(ENC_MAADR5,mymac.u8Address[4]);
		enc_writecontrolregister(ENC_MAADR6,mymac.u8Address[5]);
	//-----------------------------------------------------------------------------
	/*	Setup the PHY-Stuff */
		u16tmp	= enc_readphyregister(ENC_PHCON1);
		u16tmp	|= 0x0100;	//set PDPXMD: FullDuplex (Take care to also adjust FULDPX-Bit in the MACON3-Register
		u16tmp	&= 0xBFFF;	//clear bit.14:LOOPBK
		enc_writephyregister(ENC_PHCON1,u16tmp);

		
	//-----------------------------------------------------------------------------
	/*	Ensure that the AUTOINC-Bit in ECON2 is set:
		If the AUTOINC bit in the ECON2 register is set, after
		the last bit of each byte is written, the EWRPT Pointer
		will automatically be incremented to point to the next
		sequential address (current address + 1). If address,
		1FFFh, is written with AUTOINC set, the Write Pointer
		will increment to 0000h. */
		enc_bitfieldset(ENC_ECON2,0x80);
		/*	Set bit.7: AUTOINC	*/
//-----------------------------------------------------------------------------
/*	Setup the Interrupts */
	/*	Enable AVR-INT1-Pin (ENC28J60-Interrupt):
		A falling Flag should trigger (ATmega644) */
		EICRA	|=	(_BV(ISC11));

//-----------------------------------------------------------------------------
/*	Setup the Receiver */
/*		7.2.1         ENABLING RECEPTION
		Assuming that the receive buffer has been initialized,
		the MAC has been properly configured and the receive
		filters have been configured to receive Ethernet
		packets, the host controller should:
		1.    If an interrupt is desired whenever a packet is
		      received, set EIE.PKTIE and EIE.INTIE. */
		enc_bitfieldset(ENC_EIE,0x80 | 0x40);
		/*	Set bit.7: INTIE
		 * 	Set bit.6: PLTIE	*/		
/*		2.    If an interrupt is desired whenever a packet is
		      dropped due to insufficient buffer space, clear
		      EIR.RXERIF and set both EIE.RXERIE and
		      EIE.INTIE */
/*		3.    Enable reception by setting ECON1.RXEN.
		After setting RXEN, the Duplex mode and the Receive
		Buffer Start and End Pointers should not be modified.
		Additionally, to prevent unexpected packets from arriv-
		ing, it is recommended that RXEN be cleared before
		altering the receive filter configuration (ERXFCON) and
		MAC address.	*/
		enc_bitfieldset(ENC_ECON1, 0x04);
		/*	Set bit.2: RXEN */
		u8tmp	= enc_readcontrolregister(ENC_EREVID);


}
/*****************************************************************************/
/*	Write an Ethernet-Header to the given Address */
/*****************************************************************************/
void*	ethernet_writeheader(void* pHeader,MACADDRESS* psSource,MACADDRESS* psDestination,u16 u16Protocol){
ETHERNETHEADER*	psEthernetHeader	=	pHeader;
psEthernetHeader->sDestination		=	*psDestination;
psEthernetHeader->sSource			=	*psSource;
psEthernetHeader->u8TypeHigh		=	HI(u16Protocol);
psEthernetHeader->u8TypeLow			=	LO(u16Protocol);
	return	(s8*)pHeader+sizeof(ETHERNETHEADER);
}
/*****************************************************************************/
/*	SEND STUFF */
/*****************************************************************************/
/*	Wait until we can transmit this amount of data */
void	ethernet_waittxready(u16 u16Length){
/*	Save the amount of data in the frame */
	gu16EthernetTxFrameData	= u16Length;
/*	To achieve the example layout shown in Figure 7-2 an
	to transmit a packet, the host controller should:
	1.    Appropriately program the ETXST Pointer to
	      point to an unused location in memory. It will
	      point to the per packet control byte. In the
	      example, it would be programmed to 0120h. It is
	      recommended that an even address be used for
	      ETXST. */
	enc_writecontrolregister(ENC_ETXSTL,(u8)ENC_TXST);
	enc_writecontrolregister(ENC_ETXSTH,(u8)(ENC_TXST>>8));
	/*	2.    Use the WBM SPI command to write the per
		      packet control byte, the destination address, the
		      source MAC address, the type/length and the
		      data payload.:
		The WBM command is started by lowering the CS pin.
		The WBM opcode should then be sent to the ENC28J60, followed by the 5-bit constant,
		1Ah. After the WBM command and constant are sent, the data to be stored in the memory
		pointed to by EWRPT should be shifted out MSb first to the ENC28J60. After 8 data
		bits are received, the Write Pointer will automatically increment if AUTOINC is set.
		The host controller can continue to provide clocks on the SCK pin and send data on 
		the SI pin, without raising CS, to keep writing to the memory.
		The WBM command is terminated by bringing up the CS pin. */
	enc_writecontrolregister(ENC_EWRPTL,(u8)ENC_TXST);
	enc_writecontrolregister(ENC_EWRPTH,(u8)(ENC_TXST>>8));

	/*	Wait until the SPI-Bus is ready */
		WaitEvent(&gsSpiLockEvent);
		/*	Setup the SPI-Speed */
		enc_spiconfig();
		/*	Make the CS-Pin low */
		ENC_DESELECT_CS
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		/*	Write Instruction : Consists of a 3-Bit Opcode (011: Write control register) and 
		 * 	the 5-Bit code 11010 */
		spi_rw(0x7A);
		/*	Write the Packet Control-Byte: see datasheet*/
		spi_rw(0x0E);
}
/*****************************************************************************/
/*	Wait until the Hardware has send the package */
/*****************************************************************************/
void	ethernet_waittxdone(void){
u16	u16tmp;
u8	u8tmp;
	/*	Make the CS-Pin high */
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	asm volatile("nop"::);
	ENC_SELECT_CS
	/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
/*	3.    Appropriately program the ETXND Pointer. It
      should point to the last byte in the data payload (mind: the first byte is
      the packet-control-byte). */
	u16tmp	= gu16EthernetTxFrameData + ENC_TXST;
	enc_writecontrolregister(ENC_ETXNDL,(u8)u16tmp);
	enc_writecontrolregister(ENC_ETXNDH,(u8)(u16tmp>>8));
/*	4.    Clear EIR.TXIF, set EIE.TXIE and set EIE.INTIE
      to enable an interrupt when done (if desired). */
	enc_bitfieldclear(ENC_EIR,0x0A);
	/*	Clear bit.3: TXIF	
	 * 	Clear bit.1: TXERIF */
	enc_bitfieldset(ENC_EIE,0x80 | 0x08 | 0x02);
	/*	Set bit.7: INTIE
	 * 	Set bit.2: TXERIE
	 * 	Set bit.3: TXIE */
/*	Disable the receiver Interrupt */
	enc_bitfieldclear(ENC_EIE,0x40);
	/*	Clear bit.6: PKTIE */	

/*	5.    Start the transmission process by setting
      ECON1.TXRTS. */
	enc_bitfieldset(ENC_ECON1,0x08);
	/*	Set bit.3: TXRTS */
/*	Wait for the Transmission to complete */
	while(1){
		u8tmp	= enc_readcontrolregister(ENC_ECON1);
		u8tmp	= enc_readcontrolregister(ENC_ESTAT);
		u8tmp	= enc_readcontrolregister(ENC_EIR);
		if((u8tmp & 0x08) != 0)	break;
		/*	Wait for an interrupt */
		CS_IRQ_ENABLE
		WaitEvent(&NetworkEVENT);
	}
//Reset the TXRTS-Flag if set
	enc_bitfieldclear(ENC_ECON1,0x10);

	u8tmp	= enc_readcontrolregister(ENC_EIR);
	if((u8tmp & 0x02)!=0){
			while(1);
	}
	u8tmp	= enc_readcontrolregister(ENC_ESTAT);
	if((u8tmp & 0x02)!=0){
			while(1);
	}

	/*	Clear Interrupt */
	enc_bitfieldclear(ENC_EIR,0x0A);
	/*	Clear bit.3: TXIF	
	 * 	Clear bit.1: TXERIF */
/*	Reenable the receiver Interrupt */
	enc_bitfieldset(ENC_EIE,0x40);
/*	Set bit.6: PKTIE */	

/*	We cant be sure, that no other Network-Event took place meanwhile */
	SignalEvent(&NetworkEVENT);
/*	Read the ESTAT-Register */
//	u8tmp	= enc_readcontrolregister(ENC_ESTAT);
}
/*****************************************************************************/
/*	Send the data */
/*****************************************************************************/
void	ethernet_senddata(void* pFrame,u16 u16Length){
u8*	u8pData	= (u8*)pFrame;
	while(u16Length > 0){
		spi_rw(*u8pData);
		u16Length--;
		u8pData++;
	}
}
/*****************************************************************************/
/*	Send equal bytes */
/*****************************************************************************/
void	ethernet_sendbyte(u8 u8Value,u16 u16Cnt){
	while(u16Cnt > 0){
		spi_rw(u8Value);
		u16Cnt--;
	}	
}
/*****************************************************************************/
/*	Send a frame */
/*****************************************************************************/
void	ethernet_sendframe(void* pFrame,u16 u16Length){
	ethernet_waittxready(u16Length);
	ethernet_senddata(pFrame,u16Length);
	ethernet_waittxdone();
}
/*****************************************************************************/
/*	Ethernet: DropFrameContens */
/*****************************************************************************/
u16		ethernet_dropframecontents(u16 u16ByteLength){
u16	u16ReadLength;
u16	u16tmp;
/*	Determine how much we can drop */
	if(gu16EthernetFrameDataLeft < u16ByteLength){
		u16ReadLength	= gu16EthernetFrameDataLeft;
	}else{
		u16ReadLength	= u16ByteLength;
	}
/*	Update the number of bytes left */
	gu16EthernetFrameDataLeft	-= u16ReadLength;
/*	Check if thats all */
	if(gu16EthernetFrameDataLeft != 0){
	/*	We can still read from the packet: drop the data */
		u16tmp	= u16ReadLength;
		while(u16tmp > 0){
			spi_rw(0);
			u16tmp--;
		}
	}
		return	u16ReadLength;
}
/*****************************************************************************/
/*	Ethernet: DropFrameContens */
/*****************************************************************************/
u16		ethernet_truncateframe(u16 u16ByteLength){
/*	Try to truncate the number of bytes we still have in the package to the
 *  given value. This is usefull for higher level protocolls to  through away
 * 	padding bytes the lower level layers may have inserted */
u16	u16ReadLength;
	if(gu16EthernetFrameDataLeft == 0){
		/*	We have no data left in the frame */
		return	0;
	}
/*	Determine how much we can truncate */
	if(gu16EthernetFrameDataLeft < u16ByteLength){
		u16ReadLength	= gu16EthernetFrameDataLeft;
	}else{
		u16ReadLength	= u16ByteLength;
	}
/*	Adjust the amount of data we still have in the packet */
	gu16EthernetFrameDataLeft	= u16ReadLength;
	return	u16ReadLength;
}
/*****************************************************************************/
/*	Read from the frame */
/*****************************************************************************/
u16	ethernet_readframe(void* pBuffer,u16 u16Len){
u16	u16ReadLength;
u8*	u8pBuffer	= ((u8*)pBuffer);
u16	u16Read;
/*	Determine how much we can copy */
	if(gu16EthernetFrameDataLeft < u16Len){
		u16ReadLength	= gu16EthernetFrameDataLeft;
	}else{
		u16ReadLength	= u16Len;
	}
/*	Save the number of bytes to read */
	u16Read	= u16ReadLength;
/*	Update the number of bytes left */
	gu16EthernetFrameDataLeft	-= u16Read;
/*	Copy the data */
	while(u16ReadLength > 0){
		*u8pBuffer	= spi_rw(0);
		u8pBuffer++;
		u16ReadLength--;
	}
/*	Check if this is all */
	return	u16Read;
}
/*****************************************************************************/
/*	Gets the number of bytes left to read */
/*****************************************************************************/
inline	uint	ethernet_dataleft(void){
	return	gu16EthernetFrameDataLeft;
}
/*****************************************************************************/
