/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/* ETHERNET */
/*****************************************************************************/
#ifndef __ETHERNET
#define	__ETHERNET
#include "global.h"
#include "inttypes.h"
#include "system.h"
#include "network.h"
#include "spi.h"
#include "enc28j60.h"
/*****************************************************************************/
/* Structures */
/*****************************************************************************/
typedef struct	_MACADDRESS MACADDRESS;
typedef struct	_ETHERNETHEADER ETHERNETHEADER;
typedef struct	_ETHERNETFRAME ETHERNETFRAME;
/*****************************************************************************/
struct _MACADDRESS{
u8	u8Address[6];
};
struct _ETHERNETHEADER{
MACADDRESS	sDestination;
MACADDRESS	sSource;
u8		u8TypeHigh;
u8		u8TypeLow;
};
struct _ETHERNETFRAME{
MACADDRESS	sDestination;
MACADDRESS	sSource;
u8		u8TypeHigh;
u8		u8TypeLow;
u8		u8Data[];
};
/*****************************************************************************/
/* Macros */
/*****************************************************************************/
#define	SETMACADDRESS(ma,ad0,ad1,ad2,ad3,ad4,ad5)\
		ma.u8Address[0] = ad5;\
		ma.u8Address[1] = ad4;\
		ma.u8Address[2] = ad3;\
		ma.u8Address[3] = ad2;\
		ma.u8Address[4] = ad1;\
		ma.u8Address[5] = ad0;
#define	COPYMACADDRESS(dest,src)\
		dest.u8Address[0] = src.u8Address[0];\
		dest.u8Address[1] = src.u8Address[1];\
		dest.u8Address[2] = src.u8Address[2];\
		dest.u8Address[3] = src.u8Address[3];\
		dest.u8Address[4] = src.u8Address[4];\
		dest.u8Address[5] = src.u8Address[5];
#define	MACCOMPARE(mac1,mac2) ((mac1.u16Address[0] == mac2.u16Address[0]) && \
	(mac1.u16Address[1] == mac2.u16Address[1]) && \
	(mac1.u16Address[2] == mac2.u16Address[2]))
/*****************************************************************************/
/* Internal-API */
/*****************************************************************************/
/*****************************************************************************/
/* USER-API */
/*****************************************************************************/
/*	This are the functions the higher level protocols may use */
/************ SETUP **************/
/*	Initialize */
void	ethernet_init(void);
/*	Sets the MAC of the Device */
void	ethernet_setmac(MACADDRESS* psMac);
/************ READ FRAME **************/
/*	Returns the amount of data still in the packet */
u16		ethernet_dataleft(void);
/*	Reads data from the packet */
u16		ethernet_readframe(void* pBuffer,u16 u16Length);
/*	The given amount of data from the currently processed packet is droped */
u16		ethernet_dropframecontents(u16 u16ByteLength);
/*	Set the number of bytes remaining in the current packet to this (if possible) */
u16		ethernet_truncateframe(u16 u16ByteLength);
/************ WRITE FRAME **************/
/*	Writes an Ethernet-Header */
void*	ethernet_writeheader(void* pHeader,MACADDRESS* psSource,MACADDRESS* psDestination,u16 u16Protocol);
/*	Wait until we can transmit this amount of data */
void	ethernet_waittxready(u16 u16Length);
/*	Wait until the Hardware has send the package */
void	ethernet_waittxdone(void);
/*	Send the data */
void	ethernet_senddata(void* pFrame,u16 u16Length);
void	ethernet_sendbyte(u8 u8Value,u16 u16Cnt);
/*	Send a whole frame */
void	ethernet_sendframe(void* pFrame,u16 u16Length);
/*****************************************************************************/
/* global variabels */
/*****************************************************************************/
/*	Contains the number of bytes in the packet we want to send */
u16		gu16EthernetTxFrameData;
u16		gu16EthernetNextRxFrame;
/*	Contains the number of Data left in the currently processed
	frame */
u16	gu16EthernetFrameDataLeft;
/*****************************************************************************/
MACADDRESS	debugmac;
MACADDRESS	mymac;
MACADDRESS	zeromac;
MACADDRESS	broadmac;
#define	MYMAC0		0xDC
#define	MYMAC1		0xAC
#define	MYMAC2		0x00
#define	MYMAC3		0xc2
#define	MYMAC4		0x50
#define	MYMAC5		0x00
#define	DEBUGMAC0	0x73	
#define	DEBUGMAC1	0x33
#define	DEBUGMAC2	0xA0
#define	DEBUGMAC3	0x1B
#define	DEBUGMAC4	0x30
#define	DEBUGMAC5	0x00
/*****************************************************************************/
#endif
