/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	FAT	(16/32) - Driver */
/*****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include "inttypes.h"
#include "sd.h"
#include "fat.h"
/*****************************************************************************/
/*	FAT: */
/*****************************************************************************/
/*	Known limitations: A file/directory with a filesize of 0xFFFFFFFF
	cant be read!!!! */

u8		gu8ChrBuffer;
FATFILE	gsSearchDir;
u16		gu16SearchName[FAT_MAXFILELENGTH];
/*	TODO:
Ensure that the filename of a file isnt longer than FAT_MAXFILELENGTH
during the search!!! 
Define if the Maximum length is with path or without!!! */
/*****************************************************************************/
/*	FAT: StdIo */
/*****************************************************************************/
int 	fat_get(FILE* spFile){
/*	This is the STDIO-Interface to FAT-Files */
FATFILE* 	spFatFile	= fdev_get_udata(spFile);
s16			s16Ret;
	EnterCritical();
	s16Ret	= fat_readfile(spFatFile,&gu8ChrBuffer,1);
	if(s16Ret	== 1){
		s16Ret	= (s16)gu8ChrBuffer;
	}else{
		s16Ret	= EOF;
	}
	LeaveCritical();
	return	s16Ret;
}
/*****************************************************************************/
/*	FAT: SearcheFile */
/*****************************************************************************/
s8		fat_searchfile(FATFILE* spFile,u16* u16pWName){
/*	Searches for a given file */
u16*	u16pStart;
u16*	u16pCurrent;
u8		u8typ;
u8		u8ReturnFile	= 0;
	/*	Start with the root */
	fat_openroot(&gsSearchDir);
	/*	Search for the Directories */
	while(1){
nextsubdirectory:
		/*	Save the first character */
		u16pStart	= u16pWName;
		/*	Search for the next '/' */
		while(1){
			if(*u16pWName == 0){
				/*	The String finishes here. Thus we have
					to find a file now!!! */
					u8ReturnFile	= 1;
				break;
			}else if(*u16pWName == '/'){
				/*	Replace it with a Zero Termination */
				*u16pWName	= 0;
				break;
			}
			u16pWName++;
		}
		/*	Search for this Subdirectory or file */
		while(1){
nextentry:
			u16pWName	= u16pStart;
			u8typ		= fat_nextentry(&gsSearchDir);
			if(u8typ == FAT_EOC){
				/*	We reached the end of the directory.
					We didnt find the file!!! */
					return -1;
			}else if((u8ReturnFile == 0) && (u8typ != FAT_DIR)){
				/*	We shouldnt return a file, but we have an entry
					that belongs not to a directory - skip */
					goto nextentry;
			}
			/*	Get the Filename of the entry */
			fat_getlongfilename(&gsSearchDir,gu16SearchName,FAT_MAXFILELENGTH);
			u16pCurrent	= gu16SearchName;
			/*	Compare the Names (case insensitive) */
				while(1){
					if((*u16pCurrent | 0x20) != (*u16pWName | 0x20)){
						/* Names dont match: Try the next one */
							goto nextentry;
					}else if(*u16pWName == 0){
							/*	The end of one string is reached */
						if(*u16pCurrent==0){
							/*	Both strings are zero, we have a match!!! */
							if(u8ReturnFile == 0){
								/*	We have to open another subdirectory */
								/*	jump over the zero-termination - get
									next Subdirectory */
								u16pWName++;
								/*	We have to open the Subdirectory */
								fat_changedirectory(&gsSearchDir);
								goto nextsubdirectory;
							}else{
								/*	We found the file, return it */
								fat_openfile(&gsSearchDir,spFile);
								return	0;
							}
						}else{
							/*	Only one is zero, we have no match!!! */
							goto nextentry;
						}
					}else{
						/*	Get the next character */
						u16pCurrent++;
						u16pWName++;
					}
				}

		}//End while: Search Entry

	}//End while: Open Subdirectory
}
/*****************************************************************************/
/*	FAT: RewindFile */
/*****************************************************************************/
u8		fat_rewindfile(FATFILE* spFile){
/*	We should set the Position to the Beginning */
	spFile->u32CurrentCluster 	=		spFile->u32FirstCluster;
	spFile->u16OffsetCluster	=	0;
	spFile->u32Position			=	0;
	return	0;
}/*****************************************************************************/
/*	FAT: RewindDirectory */
/*****************************************************************************/
u8		fat_rewinddirectory(FATFILE* spDir){
/*	We should rewind the Directory, e.g. set the Current Read-Pointer
	to the beginning of the Directory */
	spDir->u32CurrentCluster 	=		spDir->u32FirstCluster;
/*	We open the directory. A call to "fat_nextentry" should return
	the FIRST and not the NEXT entry */
	spDir->u16OffsetCluster		=	0xFFFF;
/*	Start reading at the Beginning */
	spDir->u32Position			=	0;
	return	0;
}
/*****************************************************************************/
/*	FAT: OpenFile */
/*****************************************************************************/
s8		fat_openfile(FATFILE* spDir,FATFILE* spFile){
/*	We open the File which belongs to the current Directory-Entry:
	Get the Parameters of the file */
u32				u32FirstCluster;
FATDIRENT*		spDirent			=	(FATDIRENT*)gu8FATDirectoryEntry;
u32				u32CurrentCluster;
u16				u16OffsetCluster;
s8				s8Result;
/*	Save current position */
u32CurrentCluster	=	spDir->u32CurrentCluster;
u16OffsetCluster	=	spDir->u16OffsetCluster;
/*	Get the Short-Entry of the File */
		s8Result	= 	fat_loadnextshortentry(spDir);
	if(s8Result !=0){
		/*	There was an error */
		return s8Result;
	}
/*	Save the files first cluster (its also the current cluster at the
	beginning !!!) */
		u32FirstCluster	=	(u32)spDirent->u16FstClusLO;
	/* Even if its a FAT16-Volume we read the ClusHI-entry!!!! */
		u32FirstCluster	|=	((u32)spDirent->u16FstClusHI << 16);
	/*	First and Current Cluster are both the first cluster of the file */
	spFile->u32CurrentCluster 	=	u32FirstCluster;
	spFile->u32FirstCluster 	=	u32FirstCluster;
/*	We have to save the filesize */
	spFile->u32FileSize			=  spDirent->u32FileSize;
/*	We start Reading at the beginning of the File */
	spFile->u32Position			=	0;
/*	We havent read anything yet */
	spFile->u16OffsetCluster	=	0;
/*	Rewind the last read entry of the Directory */
	spDir->u32CurrentCluster		=	u32CurrentCluster;
	spDir->u16OffsetCluster			=	u16OffsetCluster;
	return	0;
}
/*****************************************************************************/
/*	FAT: ChangeDirectory */
/*****************************************************************************/
s8		fat_changedirectory(FATFILE* spDir){
/*	The current Directory-Entry should be used as the new Directory.
	We change */
u32				u32FirstCluster;
FATDIRENT*		spDirent			=	(FATDIRENT*)gu8FATDirectoryEntry;
s8				s8Result;
/*	Get the Short-Entry of the File */
		s8Result	= 	fat_loadnextshortentry(spDir);
	if(s8Result !=0){
		/*	There was an error */
		return s8Result;
	}
/*	Save the files first cluster */
		u32FirstCluster	=	(u32)spDirent->u16FstClusLO;
	/* Even if its a FAT16-Volume we read the ClusHI-entry!!!! */
		u32FirstCluster	|=	((u32)spDirent->u16FstClusHI << 16);
	/*	First and Current Cluster are both the first cluster of the file */
/*	If the FirstCluster is zero, we have a special situation: We have
	to go back to the root-Directory */
	if(u32FirstCluster == 0){
		u32FirstCluster	= FATVolume.u32RootClu;
	}
	spDir->u32CurrentCluster 	=	u32FirstCluster;
	spDir->u32FirstCluster 	=	u32FirstCluster;
/*	We have to save the filesize: We dont know how long the
	Directory is. Thus we set the size to this special value!!
	*/
	spDir->u32FileSize			= 	0xFFFFFFFF;
/*	We start Reading at the beginning of the Directory */
	spDir->u32Position			=	0;
/*	We open the directory. A call to "fat_nextentry" should return
	the FIRST and not the NEXT entry */
	spDir->u16OffsetCluster		=	0xFFFF;
	return	0;
}
/*****************************************************************************/
/*	FAT: OpenRoot */
/*****************************************************************************/
u8		fat_openroot(FATFILE* spDir){
/*	Sets the Directory to point to the root-Directory */
	spDir->u32FirstCluster		= FATVolume.u32RootClu;
	spDir->u32CurrentCluster	= FATVolume.u32RootClu;
	spDir->u16OffsetCluster		= 0xFFFF;
	spDir->u32FileSize			= 0xFFFFFFFF;
	spDir->u32Position			= 0;
	return	0;
}
/*****************************************************************************/
/*	FAT: */
/*****************************************************************************/
u8		fat_loadnextshortentry(FATFILE* spDir){
FATDIRENT*		spDirent			=	(FATDIRENT*)gu8FATDirectoryEntry;
	while(1){
	fat_readfile(spDir,gu8FATDirectoryEntry,32);
		if(gu8FATDirectoryEntry[0] == 0){
			/*	The end of the directory is reached */
			return	1;
		}else if(gu8FATDirectoryEntry[0] == 0xE5){
			/*	Empty entry */
			continue;
		}
		if((spDirent->u8Attr & FATDIRENT_ATTR_LONG_NAME) != FATDIRENT_ATTR_LONG_NAME){
		/*	We have found the next short-entry!!! */
			break;
		}
	}
	return	0;
}
/*****************************************************************************/
/*	FAT: */
/*****************************************************************************/
u8		fat_nextentry(FATFILE* spDir){
	/*	If the directory is opened for the first time, the Offset is set to
		0xFFFF. In this case we have to replace it with zero and to return
		the FIRST entry of the Directory */
	if(spDir->u16OffsetCluster	== 0xFFFF){
		/*	Return the first Directory-entry */
		spDir->u16OffsetCluster	= 0;
	}else{
		/*	We want to get the NEXT enrty. Thus the current one has to be
			steped over: Search for the next ShortFilename-entry: */
		fat_loadnextshortentry(spDir);
	}
	/*	Search for the next valid entry one */
	return	fat_getdirentry(spDir);
}
/*****************************************************************************/
/*	FAT: GetDirEntry */
/*****************************************************************************/
u8		fat_getdirentry(FATFILE* spDir){
	/*	Sets the read-position to the first valid entry */
FATDIRENT*		spDirent			=	(FATDIRENT*)gu8FATDirectoryEntry;
u32				u32CurrentCluster;
u16				u16OffsetCluster;
u8				u8Chksum;
u8				u8ChksumCalc		= 0;
u8				i;
	/*	Search for the next entry */
	while(1){
		/*	Save the current position 8 */
		u32CurrentCluster	=	spDir->u32CurrentCluster;
		u16OffsetCluster	=	spDir->u16OffsetCluster;
		/*	Read the next Directory-Entry */
		fat_readfile(spDir,gu8FATDirectoryEntry,32);
		/*	Analyse it */
		if(gu8FATDirectoryEntry[0] == 0xE5){
			/*	The directory-entry is free */
		}else if(gu8FATDirectoryEntry[0] == 0x00){
			/*	The end of the Directory is reached */
			return	FAT_EOC;
		}else if((spDirent->u8Attr & FATDIRENT_ATTR_LONG_NAME) == FATDIRENT_ATTR_LONG_NAME){
			if((gu8FATDirectoryEntry[0] & 0x40) != 0){
				/*	This is the start of a new Long-Directory entry:
					Save the checksum to detect Long-Filename Orphans */
				u8Chksum	= ((FATLONGDIRENT*)gu8FATDirectoryEntry)->u8Chksum;
				break;
			}else{
				/*	Just another long-directory entry */
			}
		}else if((spDirent->u8Attr & FATDIRENT_ATTR_VOLUME_ID) != 0){
			/*	This is a special file: It contains the volume-id of the drive */
		}else{
			/*	We should have a normal file now (only a short name!!):
				Rewind this enrtry and return */
			break;
		}
	}
	/*	We found the next Directory-Entry. We have to read the short-Directory entry
		to find out whether its a file or a directory */
	/*	Rewind the last read entry */
	spDir->u32CurrentCluster		=	u32CurrentCluster;
	spDir->u16OffsetCluster			=	u16OffsetCluster;
	/*	Retrive the short-filename entry */
	fat_loadnextshortentry(spDir);
	/*	We have to check for Long-Filename Orphans: Calculate the Checksum over the
		short directory-entry */
	for(i=0;i<11;i++){
		u8ChksumCalc	+=	gu8FATDirectoryEntry[i];
	}//DANGER
	/*	Rewind */
	spDir->u32CurrentCluster		=	u32CurrentCluster;
	spDir->u16OffsetCluster			=	u16OffsetCluster;
	/*	Decide whether its a Directory or a file */
	if((spDirent->u8Attr & FATDIRENT_ATTR_DIRECTORY) != 0){
		return	FAT_DIR;
	}else{
		return	FAT_FILE;
	}
}

/*****************************************************************************/
/*	FAT: Get short filename */
/*****************************************************************************/
s8		fat_getshortfilename(FATFILE* spDir,s8* s8Name){
/*	Writes the short filename (ASCII Characters) into the given string.
 *  This string must provide at least 12 characters */
u32				u32CurrentCluster;
u16				u16OffsetCluster;
s8				s8Result;
u8				u8chr;
u8				i,j;
/*	Save current position */
u32CurrentCluster	=	spDir->u32CurrentCluster;
u16OffsetCluster	=	spDir->u16OffsetCluster;
/*	Get the Short-Entry of the File */
		s8Result	= 	fat_loadnextshortentry(spDir);
	if(s8Result !=0){
		/*	There was an error */
		return s8Result;
	}
/*	Now we have the Short-Entry for the current file. Extract the
 * 	filename */
	/*	We have a short name */
	/*	We copy max 8 bytes up to the first space */
	for(i=0;i<8;i++){
		/*	Copy the first 8 bytes to the string. Because they are 
			ASCII, we have to convert */
		u8chr	=	gu8FATDirectoryEntry[i];
		if(u8chr == 0x20){
			/*	Its a space: The string is over */
			break;
		}else{
			/*	Convert to unicode and write to the buffer */
			s8Name[i]	= u8chr;
		}
	}//end for first 8 characters
	/*	OK, we still have to to the extension. We have an
		extension if the extension isnt equal to "   ". Thus
		we have to check that */
	if((gu8FATDirectoryEntry[8] == ' ') &&
		(gu8FATDirectoryEntry[9] == ' ') &&
		(gu8FATDirectoryEntry[10] == ' ')){
		/*	We dont have to add an extension */
	}else{
	/*	We add a dot */
		s8Name[i]		= '.';
		i++;
	/*	We copy max 3 bytes up to the first space */
		for(j=0;j<3;j++){
			/*	Copy the extension. Because they are 
				ASCII, we have to convert */
			u8chr	=	gu8FATDirectoryEntry[j+8];
			if(u8chr == 0x20){
				/*	Its a space: The string is over */
				break;
			}else{
				/*	Convert to unicode and write to the buffer */
				s8Name[i+j]	= u8chr;
			}
		}//end for copy extension
	}//end else extension
	/*	We have to add the zero-termination */
	s8Name[i+j]		= 0;
	/*	Rewind the Directory Entry */	
	spDir->u32CurrentCluster		=	u32CurrentCluster;
	spDir->u16OffsetCluster			=	u16OffsetCluster;
	return	0;
}
/*****************************************************************************/
/*	FAT: Get long filename */
/*****************************************************************************/
u8		fat_getlongfilename(FATFILE* spDir,u16* u16pBuffer,u16 u16Size){
/*	Copy the long filename of the current directory into the
	buffer */
/*	Save the current read-position of the directory */
u32				u32CurrentCluster	=	spDir->u32CurrentCluster;
u16				u16OffsetCluster	=	spDir->u16OffsetCluster;
u8				u8LongEntryCnt;
u16*			u16pNameWrPos;
u16* 			u16pWrPos;
u8				i;
u8				u8chr;
FATDIRENT*		spDirent			=	(FATDIRENT*)gu8FATDirectoryEntry;
/*	Read the current directory-entry and one more byte */
		fat_readfile(spDir,gu8FATDirectoryEntry,32);
	/*	The first Byte is important:
		0xE5:	Entry is free
		0x00:	Entry is free and nothing follows
		0x05:	Its actually 0xE5 (KANJI, japanees stuff)
				But: During a long directory entry, this seems to mean the 5st
				part nevertheless 
		0x01-... up to 0xXX | 0x40: Long Directory entry */
	if((spDirent->u8Attr & FATDIRENT_ATTR_LONG_NAME) == FATDIRENT_ATTR_LONG_NAME){
		if((gu8FATDirectoryEntry[0] |= 0x40) != 0){
			/*	We start with a Long-Directory Entry:
				Get the number of entrys */
			u8LongEntryCnt	= gu8FATDirectoryEntry[0] & 0x3F;
			/*	One enrtry can hold up to 13 characters:
				We have to reserve space for the comming entries */
				u16pNameWrPos	= u16pBuffer + (u8LongEntryCnt-1)*13;
			/*	We have to write a zero-termination for the case that
				the first LongFilename Entry is completly full */
				*(u16pNameWrPos+13)	= 0;
			while(1){
			/*	Copy the characters from the current Node */
				memcpy(u16pNameWrPos,&gu8FATDirectoryEntry[1],10);
				u16pNameWrPos	+=	5;
				memcpy(u16pNameWrPos,&gu8FATDirectoryEntry[14],12);
				u16pNameWrPos	+=	6;
				memcpy(u16pNameWrPos,&gu8FATDirectoryEntry[28],4);
				/*	We have to rewind the last 11 bytes and a whole entry of 13 bytes */
				u16pNameWrPos	-=	24;
				/*	One more entry copied */
				u8LongEntryCnt--;
				if(u8LongEntryCnt == 0)	break;
				/*	We have to read another enrtry */
				fat_readfile(spDir,gu8FATDirectoryEntry,32);
			}
		}else{
			/*	Dont know what that means... */

		}
	}else{
		/*	We have a short name */
		u16pWrPos	=	u16pBuffer;
		/*	We copy max 8 bytes up to the first space */
		for(i=0;i<8;i++){
			/*	Copy the first 8 bytes to the string. Because they are 
				ASCII, we have to convert */
			u8chr	=	gu8FATDirectoryEntry[i];
			if(u8chr == 0x20){
				/*	Its a space: The string is over */
				break;
			}else{
				/*	Convert to unicode and write to the buffer */
				*u16pWrPos++	= (u16)u8chr;
			}
		}//end for first 8 characters
		/*	OK, we still have to to the extension. We have an
			extension if the extension isnt equal to "   ". Thus
			we have to check that */
		if((gu8FATDirectoryEntry[8] == ' ') &&
			(gu8FATDirectoryEntry[9] == ' ') &&
			(gu8FATDirectoryEntry[10] == ' ')){
			/*	We dont have to add an extension */
		}else{
		/*	We add a dot */
			*u16pWrPos++		= '.';
		/*	We copy max 3 bytes up to the first space */
			for(i=0;i<3;i++){
				/*	Copy the extension. Because they are 
					ASCII, we have to convert */
				u8chr	=	gu8FATDirectoryEntry[i+8];
				if(u8chr == 0x20){
					/*	Its a space: The string is over */
					break;
				}else{
					/*	Convert to unicode and write to the buffer */
					*u16pWrPos++	= (u16)u8chr;
				}
			}//end for copy extension
		}//end else extension
		/*	We have to add the zero-termination */
		*u16pWrPos++		= 0;
	}//End Short Name
/*	Rewind this action */
	spDir->u32CurrentCluster		=	u32CurrentCluster;
	spDir->u16OffsetCluster			=	u16OffsetCluster;
/*	Ensure that the String has a zero-termination */
	u16pBuffer[u16Size-1]		=	0;
	return	0;			
}
/*****************************************************************************/
/*	FAT: Remove a file: Free its Clusters */
/*****************************************************************************/
u8		fat_rmfile(u32 u32Cluster){
/*	Removes the file, starting at the given cluster */
u32		u32NextCluster;
while(1){
	u32NextCluster	= fat_getentry(u32Cluster);
	/*	Free the current Cluster */
		fat_setentry(u32Cluster,0);
		if(FATVolume.u32LastFreeCluster > u32Cluster){
			FATVolume.u32LastFreeCluster	= u32Cluster;
		}
	/*	If the EOC is reached, we deleted the whole file */
	if(fat_entrytyp(u32NextCluster) == FAT_EOCCLUSTER)	return	0;
	/*	We still have to remove some clusters ... */
	u32Cluster	= u32NextCluster;
}
}
/*****************************************************************************/
/*	FAT: Write to a file on the FAT-System */
/*****************************************************************************/
u8		fat_writefile(FATFILE* spFile,u8* u8pBuffer,u16 u16Size){
/*	We write the buffer to the file and increment the current position */
u16		u16Offset;
u16		u16OffsetCluster;
u16		u16Left;
u32		u32HwSector;
u32		u32CurrentCluster;
u32		u32NewCluster;

	u16OffsetCluster	= spFile->u16OffsetCluster;
	u32CurrentCluster	= spFile->u32CurrentCluster;
do{

/*	Calculate the Hardware-Sector */
	u32HwSector		=	FATVolume.u32FirstDataHwSector;
	u32HwSector		+=	(u32CurrentCluster * FATVolume.u16HwSectorPerClus);
	u32HwSector		+=	(u16OffsetCluster >> FAT_HWSECTORBITCNT);
/*	Calculate the Offset in the current Hardware-Sector */
	u16Offset		=	u16OffsetCluster & FAT_HWSECTORBITMSK;

/*	Calculate the remaining bytes in the current Hardware-Sector */
	u16Left			= FAT_HWSECTORSIZE	- u16Offset;
	if(u16Size < u16Left){
		/*	We dont need to access another Hardware-Sector */
		u16Left	= u16Size;
	}
		u16Size	-= 	u16Left;
/*	Write the data */
	sd_wr(u32HwSector,u16Offset,u8pBuffer,u16Left);
	u8pBuffer				+= u16Left;
	u16OffsetCluster		+= u16Left;
	if(u16OffsetCluster == FATVolume.u16BytsPerClus){
		/*	We have reached the end of the Cluster, load a new one */
		u32NewCluster		= fat_getentry(u32CurrentCluster);
		/*	Test, if this is the end of the cluster-chain: in this case
			we have to allocate a new cluster for the file */
		if(fat_entrytyp(u32NewCluster) == FAT_EOCCLUSTER){
			/*	Allocate a new cluster */
			u32NewCluster	= fat_nextfreecluster();
			/*	The new cluster is the new EOC */
			fat_setentry(u32NewCluster,0xFFFFFFFF);
			/*	The current cluster has to point to the new cluster */
			fat_setentry(u32CurrentCluster,u32NewCluster);
		}
		u32CurrentCluster			= u32NewCluster;
		u16OffsetCluster			= 0;
	}
}while(u16Size!=0);
/*	Job done */
	spFile->u16OffsetCluster	= 	u16OffsetCluster;
	spFile->u32CurrentCluster	=	u32CurrentCluster;
	return	0;
}
/*****************************************************************************/
/*	FAT: Get next free Cluster */
/*****************************************************************************/
u32		fat_nextfreecluster(void){
u32		u32cluster;
u32		u32erg;
/*	Start with the last free cluster */
	u32cluster	= FATVolume.u32LastFreeCluster;
	while(1){
		u32erg	= fat_getentry(u32cluster);
		if(fat_entrytyp(u32erg) == FAT_FREECLUSTER){
			FATVolume.u32LastFreeCluster	= u32cluster;
			return	u32cluster;
		}
		/*	This entry isnt free, keep searching */
		u32cluster++;
	}
}
/*****************************************************************************/
/*	FAT: FAT-Entry Typ */
/*****************************************************************************/
u8		fat_entrytyp(u32 u32Entry){
	if(FATVolume.u8FATType == FAT16){
		/*	FAT 16 */
		u32Entry	&=	0xFFFF;
		if(u32Entry	== 0x0000)	return	FAT_FREECLUSTER;
		if(u32Entry	== 0xFFFF)	return	FAT_EOCCLUSTER;
		if(u32Entry	== 0xFFF7)	return	FAT_BADCLUSTER;
	}else{
		/*	FAT 32 */
		u32Entry	&=	0xFFFFFFF;
		if(u32Entry	== 0x0000000)	return	FAT_FREECLUSTER;
		if(u32Entry	== 0xFFFFFFF)	return	FAT_EOCCLUSTER;
		if(u32Entry	== 0xFFFFFF7)	return	FAT_BADCLUSTER;
	}
	return	FAT_USEDCLUSTER;
}
/*****************************************************************************/
/*	FAT: Set a FAT-Entry */
/*****************************************************************************/
u32		fat_setentry(u32 u32Cluster,u32 u32Value){
u32		u32HwSector;
u32		u32NextCluster;
u16		u16Offset;
u8		u8erg;
/*	It depends on FAT16/32 whether one FAT-Entry is 2 or 4 Bytes long
	Calculate the (Hardware)Sector-Address:
	Each Hardware-Sector has 2^FAT_HWSECTORBITCNT Bytes. Each FAT-Entry
	ocupies 2^2 or 2^1 Byte. So the number of FAT-Entrys per Hw-Sector
	is either 2^(FAT_HWSECTORBITCNT-1) or 2^(FAT_HWSECTORBITCNT-2)  */
		u32HwSector	=	FATVolume.u32FirstFATHwSector;
/*	Depending on which FAT get the entry */
		if(FATVolume.u8FATType == FAT16){
			/*	FAT16 Value */
			u32HwSector	+=	u32Cluster 		>>	(FAT_HWSECTORBITCNT-1);
			u16Offset	=	((u16)u32Cluster) << 1;
			u16Offset	&=	FAT_HWSECTORBITMSK;
			/*	Read the sector into the cache */
			u8erg		= sd_rdtocache(u32HwSector);
			/*	Set the entry */
			((u16*)gu8SdCache)[u16Offset]	= (u16)u32Value;
		}else{
			u32HwSector	+=	u32Cluster 		>>	(FAT_HWSECTORBITCNT-2);
			u16Offset	=	((u16)u32Cluster) << 2;
			u16Offset	&=	FAT_HWSECTORBITMSK;
			/*	Read the sector into the cache */
			u8erg		= sd_rdtocache(u32HwSector);
			/*	Get the entry */
			u32NextCluster	= ((u32*)gu8SdCache)[u16Offset];
			/*	Set the entry */
				((u16*)gu8SdCache)[u16Offset]	&=	0xF0000000;
				((u16*)gu8SdCache)[u16Offset]	|=	u32Value;
		}
/*	The Cache is dirty now */
	gu8SdCacheStatus				= SDCARD_CACHEDIRTY;
/*	Job done */
	return	0;
}
/*****************************************************************************/
/*	FAT: Get a FAT Entry */
/*****************************************************************************/
u32		fat_getentry(u32 u32Cluster){
u32		u32HwSector;
u32		u32Entry;
u32		u32NextCluster;
u16		u16Offset;
u8		u8erg;
/*	It depends on FAT16/32 whether one FAT-Entry is 2 or 4 Bytes long
	Calculate the (Hardware)Sector-Address:
	Each Hardware-Sector has 2^FAT_HWSECTORBITCNT Bytes. Each FAT-Entry
	ocupies 2^2 or 2^1 Byte. So the number of FAT-Entrys per Hw-Sector
	is either 2^(FAT_HWSECTORBITCNT-1) or 2^(FAT_HWSECTORBITCNT-2)  */
		u32HwSector	=	FATVolume.u32FirstFATHwSector;
/*	Depending on which FAT get the entry */
		if(FATVolume.u8FATType == FAT16){
			/*	FAT16 Value */
			u32HwSector	+=	(u32Cluster 		>>	(FAT_HWSECTORBITCNT-1));
			u16Offset	=	(((u16)u32Cluster) & 0xFF);
			/*	Read the sector into the cache */
			u8erg		= sd_rdtocache(u32HwSector);
			/*	Get the entry */
			u32NextCluster	= (u32) ((u16*)gu8SdCache)[u16Offset];
		}else{
			u32HwSector	+=	(u32Cluster 		>>	(FAT_HWSECTORBITCNT-2));
			u16Offset	=	(((u16)u32Cluster) & 0x7F);
			/*	Read the sector into the cache */
			u8erg		= sd_rdtocache(u32HwSector);
			/*	Get the entry */
			u32NextCluster	= ((u32*)gu8SdCache)[u16Offset];
			u32Entry		=	u32NextCluster;
			/*	The FAT32-Entry is only a 28-Bit entry */
			u32NextCluster	&=	0x0FFFFFFF;
		}
/*	Job done */
	return	u32NextCluster;
}
/*****************************************************************************/
/*	FAT: Read from a file on the FAT-System */
/*****************************************************************************/
s16		fat_readfile(FATFILE* spFile,u8* u8pBuffer,s16 s16Size){
/*	We read the given amount of data from the file and increment the
	position */
/*	We will return the read size in bytes. If this is less than the
	number we want to read, this means EOF. If its -1, we have an
	error!!! */
u16		u16Size;
u16		u16Offset;
u16		u16OffsetCluster;
u16		u16Left;
u32		u32HwSector;
u32		u32DataLeft;
s8		s8Result;
	/*	We dont want a negative size!!! */
	s16Size 	&= 0x7FFF;
	/*	We will test if we have enough stuff in the file */
	u32DataLeft	= spFile->u32FileSize - spFile->u32Position;
	if(u32DataLeft == 0){
		return	0;
	}
	if((u32)s16Size > u32DataLeft){
		/*	We have less stuff in the File than we want to read!!! */
		s16Size	= (s16)u32DataLeft;
	}
	u16Size	= (u16)s16Size;
	spFile->u32Position	+=	u16Size;
/*	Now we can read in the data */
	u16OffsetCluster	= spFile->u16OffsetCluster;
do{

/*	Calculate the Hardware-Sector */
	u32HwSector		=	FATVolume.u32FirstDataHwSector;
	u32HwSector		+=	((spFile->u32CurrentCluster-2) * FATVolume.u16HwSectorPerClus);
	u32HwSector		+=	(u16OffsetCluster >> FAT_HWSECTORBITCNT);
/*	Calculate the Offset in the current Hardware-Sector */
	u16Offset		=	u16OffsetCluster & FAT_HWSECTORBITMSK;

/*	Calculate the remaining bytes in the current Hardware-Sector */
	u16Left			= FAT_HWSECTORSIZE	- u16Offset;
	if(u16Size < u16Left){
		/*	We dont need to access another Hardware-Sector */
		u16Left	= u16Size;
	}
		u16Size	-= 	u16Left;
/*	Read in the data */
	s8Result	= sd_rd(u32HwSector,u16Offset,u8pBuffer,u16Left);
	if(s8Result != 0){
		/*	We have an I/O-Error!!! */
		return	-1;
	}
/*	Everything worked, goon */
	u8pBuffer				+= u16Left;
	u16OffsetCluster		+= u16Left;

	if(u16OffsetCluster == FATVolume.u16BytsPerClus){
		/*	We have reached the end of the Cluster, load a new one */
		spFile->u32CurrentCluster	= fat_getentry(spFile->u32CurrentCluster);
		if(fat_entrytyp(spFile->u32CurrentCluster) == FAT_EOCCLUSTER){
			/*	End of Clusterchain reached: We have no new data!!! */
			return	0;
		}
		u16OffsetCluster			= 0;
	}
}while(u16Size!=0);
/*	Job done */
	spFile->u16OffsetCluster	= 	u16OffsetCluster;
/*	Return the amount of data to read */
	return	s16Size;
}
/*****************************************************************************/
/*	Init the FAT-Stuff */
/*****************************************************************************/
s8		fat_init(void){
BPB*				spVolumeBPB			= (BPB*)(gu8SdCache + sizeof(BS));
BPBFAT32*			spVolumeBPBFAT32	= (BPBFAT32*)(gu8SdCache + sizeof(BS) + sizeof(BPB));
PARTITIONTABLE*		spPartitionTable	= (PARTITIONTABLE*)(gu8SdCache + 0x01BE);
u32					u32PartitionStart	= 0;
u8					u8FATtoHWSector;
s8					s8Result;
/*	Maybe the disk has a MBR. We have to check for that. A MBR looks like this
	(wikipedia):
Address
0x0000 	0 		Boot-Loader
0x01B8 	440 	Disc-Signature (since w2k)
0x01BC 	444 	Null
0x01BE 	446 	Partitiontable
0x01FE 	510 	55hex
0x01FF 	511 	AAhex */

 /*
Partition typ: 06 

DOS 3.31+ 16-bit FAT (over 32M)
Partitions, or at least the FAT16 filesystems created on them, are at most 2 GB 
for DOS and Windows 95/98 (at most 65536 clusters, each at most 32 KB). Windows NT
can create up to 4 GB FAT16 filesystems (using 64 KB clusters), but these cause
problems for DOS and Windows 95/98. Note that VFAT is 16-bit FAT with long filenames;
FAT32 is a different filesystem.
*/
	s8Result	= sd_rdtocache(0);
	if(s8Result != 0)	return	s8Result;
/*	We probe for a MBR. If we got one, it contains the partition-table. 
	We look for the first Volume (first entry) and try to load it. If this is
	a partition-table, this entry should contain the BR of the partition */
	u32PartitionStart	= spPartitionTable->u32StartingLBA;
	s8Result	= sd_rdtocache(u32PartitionStart);
	if(s8Result != 0)	return	s8Result;
/*	We got it read in. Check the magic cookie */
	if((gu8SdCache[510] != 0x55) || (gu8SdCache[511] != 0xAA)){
		/*	We havent got a valid BR!!! (hope so!)
		So our guess, that the sector 0 contains the partition-table
		was wrong. The next approach would be to parse the first sector
		directly as the BR */	
		u32PartitionStart	= 0;
		s8Result	= sd_rdtocache(0);
		if(s8Result != 0)	return	s8Result;
	}
/*	Check if it has the proper Format */
	if((gu8SdCache[510] != 0x55) || (gu8SdCache[511] != 0xAA)){
		return	-1;
	}
/*	We will fill in the Volume-Descriptor
	First we check if the 16 or the 32-bit fields
	are valid. If the 16-BitFields are zero, the coresponding 32-bit field
	should be used instead */
/*	Determine the number of Sectors */
	if(spVolumeBPB->u16TotSec16 != 0){
		FATVolume.u32TotSec	= spVolumeBPB->u16TotSec16;
	}else{
		FATVolume.u32TotSec	= spVolumeBPB->u32TotSec32;
	}
/*	Determine the size of the FAT (measured in sectors!) */
	if(spVolumeBPB->u16FATSz16 != 0){
		FATVolume.u32FATSz	= spVolumeBPB->u16FATSz16;
	}else{
		FATVolume.u32FATSz	= spVolumeBPBFAT32->u32FATSz32;
	}
/*	Fill in other stuff	*/
	FATVolume.u8SecPerClus			= spVolumeBPB->u8SecPerClus;
	FATVolume.u16HwSectorPerClus	= spVolumeBPB->u8SecPerClus;
	FATVolume.u16BytsPerSec			= spVolumeBPB->u16BytsPerSec;
/*	The FAT-Spec says that BytesPerSec is most likly 512, but MS-OS support
	1024, 2048 and 4096 as well. So we calculate how many HWSectors (512 byte)
	give one FATSector */
	u8FATtoHWSector					= (u8)(FATVolume.u16BytsPerSec / FAT_HWSECTORSIZE);
/*	Calculate the Bytes per Cluster */
	FATVolume.u16BytsPerClus		= spVolumeBPB->u16BytsPerSec * (u16)spVolumeBPB->u8SecPerClus;
/*	Save the FAT32-Root Sector */
	FATVolume.u32RootClu			= spVolumeBPBFAT32->u32RootClu;
/*	Calculate the Number of Sectors, the Root-Directory occupies:
	Each Direcotry Entry occupies 32 Byte. To round the value up,
	we add the FAT_HWSECTORSIZE-Value and substract one before the division */
	FATVolume.u32RootDirHwSectors
		= ((((u32)spVolumeBPB->u16RootEntCnt) << 5) + (FAT_HWSECTORSIZE - 1))
			/ FAT_HWSECTORSIZE;
//		= ((((u32)spVolumeBPB->u16RootEntCnt) << 5) + (FATVolume.u16BytsPerSec - 1))
//			/ FATVolume.u16BytsPerSec;
//	Would be the formular if we were interested in FAT-Sectors!!!
/*	Calculate the First FAT-Sector */
	FATVolume.u32FirstFATHwSector	=	(spVolumeBPB->u16RsvdSecCnt * u8FATtoHWSector) + u32PartitionStart;;
/*	Calculate the First Root-Directory Sector */
	FATVolume.u32FirstRootHwSector =
		FATVolume.u32FirstFATHwSector + (((u32)spVolumeBPB->u8NumFATs) *  FATVolume.u32FATSz);
/*	Calculate the First Data-Sector */
	FATVolume.u32FirstDataHwSector =
		FATVolume.u32FirstRootHwSector + FATVolume.u32RootDirHwSectors;
/*	We determine the number of Data-Sectors */
	FATVolume.u32DataHwSectors	=
		(FATVolume.u32TotSec * u8FATtoHWSector)	- FATVolume.u32FirstDataHwSector;
/*	We determine the number of Data-Clusters */
	FATVolume.u32DataClusters	=
		FATVolume.u32DataHwSectors / (FATVolume.u16BytsPerClus / FAT_HWSECTORSIZE);
/*	Now we have to determine the FAT-Type (12,16,32) */
	if(FATVolume.u32DataClusters < 4085){
		/*	We have a FAT12-Partition */
		FATVolume.u8FATType	= FAT12;
	}else if(FATVolume.u32DataClusters < 65525){
		/*	We have a FAT16-Partition */
		FATVolume.u8FATType	= FAT16;
	}else{
		/*	The Volume is FAT32 */
		FATVolume.u8FATType	= FAT32;
	}
/*	If we have a FAT16-Volume: We have to brows the Root-Directory for a
	valid root-folder!!! */
	if(FATVolume.u8FATType	== FAT16){
u32			u32Sector		= FATVolume.u32FirstRootHwSector;
u32			u32SectorOffset	= 0;
u8			i;
u8			u8res;
FATDIRENT*	spDirent;
		/*	Read the current sector */
	while(1){
		s8Result	= sd_rdtocache(u32Sector + u32SectorOffset);
		if(s8Result != 0)	return	s8Result;
		spDirent	= (FATDIRENT*)gu8SdCache;
		/*	Search for an short directory-entry called 'root' */
		for(i=0;i<FAT_HWSECTORSIZE/32;i++){
			u8res	= strcasecmp(spDirent->s8Name,"root       ");
			u8res	= spDirent->u8Attr & FATDIRENT_ATTR_LONG_NAME;
			if((spDirent->u8Attr & FATDIRENT_ATTR_DIRECTORY) != 0){
				/*	We have a directory-entry: We need a zero-termination
					to compare the strings */
				spDirent->u8Attr	= 0;
				if(strcasecmp(spDirent->s8Name,"root       ") == 0){
					/*	We found the entry */
					FATVolume.u32RootClu =	((u32)spDirent->u16FstClusLO | ((u32)spDirent->u16FstClusHI) << 16);
					return	0;
				}
			}
			/*	Get the next dirent */
			spDirent++; 
		}
		u32SectorOffset++;
		if(FATVolume.u32RootDirHwSectors == u32SectorOffset){
			/*	We reached the end of the root-directory */
			return	-1;
		}
	}
	}
/*	We setup the FAT-Volume properly */
	return	0;
}
/*****************************************************************************/

