/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
#ifndef FAT__
#define	FAT__
#include <stdio.h>
#include "system.h"
/*****************************************************************************/
/*	FAT	(16/32) - Driver */
/*****************************************************************************/
/*	Maximum Length a Filename can have */
#define		FAT_MAXFILELENGTH	64

/*	The Partition-Table Entry format (from http://www.ata-atapi.com/hiwtab.htm)
	+--- Bit 7 is the active partition flag, bits 6-0 are
    |    zero (when not zero this byte is also the drive
    |    number of the drive to boot so the active
    |    partition is always found on drive 80H,
    |    the first hard disk).
    |
    |    +--- Starting CHS in INT 13 call format.
    |    |
    |    |      +--- Partition type byte.
    |    |      |
    |    |      |    +--- Ending CHS in INT 13 call format.
    |    |      |    |
    |    |      |    |       +-- Starting LBA.
    |    |      |    |       |
    |    |      |    |       |        +-- Size in sectors.
    |    |      |    |       |        |
    v ---+----  v ---+----   v        v
   0. 1. 2. 3. 4. 5. 6. 7. 8.9.A.B. C.D.E.F.
   DL DH CL CH TB DH CL CH LBA..... SIZE....
   80 01 01 00 06 0e be 94 3e000000 0c610900  1st entry
   00 00 81 95 05 0e fe 7d 4a610900 724e0300  2nd entry
   00 00 00 00 00 00 00 00 00000000 00000000  3rd entry
   00 00 00 00 00 00 00 00 00000000 00000000  4th entry
*/
typedef	struct	_PARTITIONTABLE PARTITIONTABLE;
struct _PARTITIONTABLE{
u8		u8DriveNumber;
u8		u8StartingCHS[3];
u8		u8PartitionTyp;
u8		u8EndingCHS[3];
u32		u32StartingLBA;
u32		u32Size;
};
/*	BPB: BiosParameterBlock: The first Sector of the Volume
	The first Structure is the same for all three FAT-types, but
	the trailing stuff depends on whether 12/16/32 */
typedef	struct	_BS BS;
struct _BS{
u8		u8jmpBoot[3];
u8		u8OEMName[8];
};
typedef	struct	_BPB BPB;
struct _BPB{
u16		u16BytsPerSec;
u8		u8SecPerClus;
u16		u16RsvdSecCnt;
u8		u8NumFATs;
u16		u16RootEntCnt;
u16		u16TotSec16;
u8		u8Media;
u16		u16FATSz16;
u16		u16SecPerTrk;
u16		u16NumHeads;
u32		u32HiddSec;
u32		u32TotSec32;
};
typedef	struct	_BPBFAT32 BPBFAT32;
struct _BPBFAT32{
u32		u32FATSz32;
u16		u16ExtFlags;
u16		u16FSVer;
u32		u32RootClu;
u16		u16FSInfo;
u16		u16BkBootSec;
u8		u8reserved0[12];
u8		u8DrvNum;
u8		u8reserved1;
u8		u8BootSig;
u32		u32VolID;
u8		u8VolLab[11];
u8		u8FilSysType[8];
};
typedef	struct	_BPBFAT16 BPBFAT16;
struct _BPBFAT16{
u8		u8DrvNum;
u8		u8reserved1;
u8		u8BootSig;
u32		u32VolID;
u8		u8VolLab[11];
u8		u8FilSysType[8];
};
/*	This is a FAT-Directory entry */
typedef	struct	_FATDIRENT FATDIRENT;
struct _FATDIRENT{
s8		s8Name[11];
u8		u8Attr;
u8		u8NTRes;
u8		u8CrtTimeTenth;
u16		u16CrtTime;
u16		u16CrtDate;
u16		u16LstAccDate;
u16		u16FstClusHI;
u16		u16WrtTime;
u16		u16WrtDate;
u16		u16FstClusLO;
u32		u32FileSize;
};
/*	Values for the u8Attr - entry */
#define	FATDIRENT_ATTR_READ_ONLY	0x01
#define	FATDIRENT_ATTR_HIDDEN		0x02
#define	FATDIRENT_ATTR_SYSTEM		0x04
#define	FATDIRENT_ATTR_VOLUME_ID	0x08
#define	FATDIRENT_ATTR_DIRECTORY	0x10
#define	FATDIRENT_ATTR_ARCHIVE		0x20
#define	FATDIRENT_ATTR_LONG_NAME	0x0F
/*	FAT Long Directory Entries */
typedef	struct	_FATLONGDIRENT FATLONGDIRENT;
struct _FATLONGDIRENT{
u8		u8Ord;
u16		u16Name1[5];
u8		u8Attr;
u8		u8Type;
u8		u8Chksum;
u16		u16Name2[6];
u16		u16FstClusLO;
u16		u16Name3[2];
};
/*****************************************************************************/
/*	Interface stuff */
/*****************************************************************************/
/*	Values for the u8FATType-Field */
#define	FAT12		0
#define	FAT16		1
#define	FAT32		2
#define	FATINVALID	3
/*	This is the FAT-Volume Descriptor. It will be filled
	with all essential Information about the FAT-Volume */
typedef	struct	_VOLDESC VOLDESC;
struct _VOLDESC{
u8		u8FATType;
u8		u8SecPerClus;
u16		u16BytsPerSec;
u16		u16BytsPerClus;
u32		u32TotSec;
u32		u32FATSz;
u32		u32RootDirHwSectors;
u32		u32DataHwSectors;
u32		u32DataClusters;
u32		u32RootClu;
u16		u16HwSectorPerClus;
u32		u32FirstFATHwSector;
u32		u32FirstRootHwSector;
u32		u32FirstDataHwSector;
u32		u32LastFreeCluster;
};

typedef	struct	_FATFILE FATFILE;
struct _FATFILE{
/*	Size of the File */
u32		u32FileSize;
/*	Position in the File */
u32		u32Position;
/*	First Cluster of the File */
u32		u32FirstCluster;
/*	Current open cluster */
u32		u32CurrentCluster;
/*	Offset in the Cluster */
u16		u16OffsetCluster;
};
/*****************************************************************************/
/*	We have a Hardware-Sectorsize of 512 Byte which equals
	9 Bit Addresses */
#define	FAT_HWSECTORSIZE	512
#define	FAT_HWSECTORBITCNT	9
#define	FAT_HWSECTORBITMSK	0x1FF	//Bitmask

#define	FAT_USEDCLUSTER		0
#define	FAT_FREECLUSTER		1
#define	FAT_BADCLUSTER		2
#define	FAT_EOCCLUSTER		3


VOLDESC		FATVolume;
u8			gu8SdCache[512];
u8			gu8SdCacheStatus;
u8			gu8FATDirectoryEntry[33];


#define	FAT_EOC		1
#define	FAT_DIR		2
#define	FAT_FILE	3
/*****************************************************************************/
/* API */
/*****************************************************************************/
int 	fat_get(FILE* spFile);
s8		fat_searchfile(FATFILE* spFile,u16* u16pWName);
u8		fat_rewinddirectory(FATFILE* spDir);
u8		fat_rewindfile(FATFILE* spFile);
u8		fat_openroot(FATFILE* spDir);
s16		fat_readfile(FATFILE* spFile,u8* u8pBuffer,s16 s16Size);
u32		fat_getentry(u32 u32Cluster);
u32		fat_setentry(u32 u32Cluster,u32 u32Value);
u8		fat_entrytyp(u32 u32Entry);
u32		fat_nextfreecluster(void);
u8		fat_writefile(FATFILE* spFile,u8* u8pBuffer,u16 u16Size);
u8		fat_rmfile(u32 u32Cluster);
u8		fat_getlongfilename(FATFILE* spDir,u16* u16pBuffer,u16 u16Size);
s8		fat_getshortfilename(FATFILE* spDir,s8* s8Name);
u8		fat_getdirentry(FATFILE* spDir);
u8		fat_nextentry(FATFILE* spDir);
u8		fat_loadnextshortentry(FATFILE* spDir);
s8		fat_openfile(FATFILE* spDir,FATFILE* spFile);
s8		fat_changedirectory(FATFILE* spDir);
s8		fat_init(void);
/*****************************************************************************/
#endif
