/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	ICMP */
/*****************************************************************************/
#include	"icmp.h"

/*****************************************************************************/
/*	Setup */
/*****************************************************************************/
void	icmp_init(void){
	gpICMPEchoBuffer	= 0;
	gcICMPEchoStatus	= ICMP_ECHOREADY;
	/*	ARP-Querr */
	gsICMPArpQuery.pCallback	= &icmp_arpcallback;
}
/*****************************************************************************/
/*	Transmit-Handler */
/*****************************************************************************/
void	icmp_transmit(void){
ETHERNETHEADER*			spEthernetHeader		= (ETHERNETHEADER*)&gu8NetworkBuffer;
IPHEADER*				spIPHeader;
ICMPHEADER*				spICMPHeader;
u16						u16Checksum;
	if(gcICMPEchoStatus != ICMP_ECHOREADY){
/*	an ICMP-Echoreply should be send */
		switch(gcICMPEchoStatus){
/*---------------------------------------------------------------------------*/	
/*	The Echo needs a MAC */
			case ICMP_ECHOGETMAC:
				/*	Start the Query */
				gcICMPEchoStatus	= ICMP_ECHOWAITARPRESPONSE;
				arp_sendquery(&gsICMPArpQuery);
				break;
/*---------------------------------------------------------------------------*/	
/*	The Echo has a MAC and should be send */
			case ICMP_ECHOSEND:
/*	Send the Message via Ethernet-Boradcast, Protocol: IP (0x0800) */
spIPHeader			=\
	ethernet_writeheader(spEthernetHeader,&mymac,&gsICMPArpQuery.sMAC,0x0800);
/*	Write the IP-Header: 
	Protokoll:		1 (ICMP)
	Messagesize:	Header (IP,UDP,DHCP),two strings(64+128 byte), Body */	
spICMPHeader		=\
	ip_writeheader(spIPHeader,1,\
		sizeof(ICMPHEADER)+guiICMPEchoBufferLen,\
		&myip,&gsICMPArpQuery.sIP,guiICMPIPIdentification);
/*	Write the ICMP-Header */
		spICMPHeader->ucType			= 0;	//ECHO-Reply
		spICMPHeader->ucCode			= 0;
		spICMPHeader->ucChecksumLow		= 0;	
		spICMPHeader->ucChecksumHigh	= 0;
/*	Calculate the Checksum */
		u16Checksum	= ip_checksum(0,spICMPHeader,sizeof(ICMPHEADER));
		u16Checksum	= ip_checksum(u16Checksum,gpICMPEchoBuffer,guiICMPEchoBufferLen);	
/*	Write the Checksum into the header */
		spICMPHeader->ucChecksumLow		= LO(~u16Checksum);	
		spICMPHeader->ucChecksumHigh	= HI(~u16Checksum);
/*	Send the packet */
	EnterCritical();	//CRITICAL SECTION
	ethernet_waittxready(sizeof(ETHERNETHEADER)+sizeof(IPHEADER)+sizeof(ICMPHEADER)+guiICMPEchoBufferLen);
	ethernet_senddata(spEthernetHeader,sizeof(ETHERNETHEADER)+sizeof(IPHEADER)+sizeof(ICMPHEADER));
	ethernet_senddata(gpICMPEchoBuffer,guiICMPEchoBufferLen);
	ethernet_waittxdone();
	/*	The Buffer isn't needed any more */
	_free(gpICMPEchoBuffer);
	gcICMPEchoStatus	= ICMP_ECHOREADY;
	LeaveCritical();	//CRITICAL SECTION

			break;
		}
	}
}
/*****************************************************************************/
/*	ICMP-Arp Handler */
void	icmp_arpcallback(ARPQUERYNODE* pNode,char cCode){
/*	The ICMP-Arp Query has succeded or timed out */
	assert(gcICMPEchoStatus	== ICMP_ECHOWAITARPRESPONSE);
	switch(cCode){
/*---------------------------------------------------------------------------*/	
/*	We got the MAC */
		case	ARP_QUERYSUCCESS:
			/*	We can send the Echo now */
			gcICMPEchoStatus	= ICMP_ECHOSEND;
			/*	Activate the TX-Task */
			gu8NetworkSend = 1;
			break;
/*---------------------------------------------------------------------------*/	
/*	We haven't got the MAC */
		case	ARP_QUERYERROR:
			assert(gpICMPEchoBuffer!=0);
			/*	Release the data */
			_free(gpICMPEchoBuffer);
			gcICMPEchoStatus	= ICMP_ECHOREADY;
			break;
	}
}
/*****************************************************************************/
/*	ICMP-Packet */
/*****************************************************************************/
void	icmp_receivepacket(void){
/*	An ICMP-Packet was received */
ICMPHEADER*		spICMPHeader		= (void*)(((uchar*)&gu8NetworkBuffer)+sizeof(ETHERNETHEADER)+sizeof(IPHEADER));
IPHEADER*		spIPHeader			= (void*)(((uchar*)&gu8NetworkBuffer)+sizeof(ETHERNETHEADER));
uint			uiSize;
/*	Read the ICMP-Header */
	uiSize	= ethernet_readframe(spICMPHeader,sizeof(ICMPHEADER));
	if(uiSize < sizeof(ICMPHEADER)){
		/*	The Packet doesn't contain a full header */
		return;
	}
/*	Test, which message arrived */
	switch(spICMPHeader->ucType){
/*---------------------------------------------------------------------------*/	
/*	Echo Request (ID 8)*/
		case 8:
/*	We have to send an Reply */
	if(gcICMPEchoStatus != ICMP_ECHOREADY){
		/*	If there is an other echo-request
			pending, do nothing */
		return;
	}
/*	Get the size of the data left */
	guiICMPEchoBufferLen	= ethernet_dataleft();
/*	Try to allocate an buffer */
	gpICMPEchoBuffer 		= _malloc(guiICMPEchoBufferLen);
/*	Copy the data if we have an buffer */
	if(gpICMPEchoBuffer != 0){
		ethernet_readframe(gpICMPEchoBuffer,guiICMPEchoBufferLen);
		/*	Save the IP to reply to the sender */
		gsICMPArpQuery.sIP			= spIPHeader->sSourceIpAddress;
		guiICMPIPIdentification		= ((uint)spIPHeader->ucIdentificationLow) + (((uint)spIPHeader->ucIdentificationHigh)<<8);
		/*	Change the Status */
		gcICMPEchoStatus	= ICMP_ECHOGETMAC;
		/*	Activate the TX-Task */
		gu8NetworkSend = 1;
	}
			break;
/*---------------------------------------------------------------------------*/	
/*	Something not supported */
		default:
			/* Ignore the Message */
			break;
	}
}
/*****************************************************************************/
