/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	IP/UDP */
/*****************************************************************************/
#include "ip-udp.h"
#include "dhcp.h"
#include "icmp.h"
#include "tcp.h"
/*---------------------------------------------------------------------------*/
/*	Init */
void	ip_init(void){
/*	Setup IP-Addresses */	
	SETIPADDRESS(zeroip,0,0,0,0);
	SETIPADDRESS(broadip,255,255,255,255);
	SETIPADDRESS(myip,192,168,178,241);
	SETIPADDRESS(dhcpip,0,0,0,0);
}

/*	Get the number of data in an IP-Datagram */
uint	ip_datalength(IPHEADER* spIPHeader){
uint	uiTotalLength	= 	((uint)spIPHeader->ucTotalLengthLow) +
							(((uint)spIPHeader->ucTotalLengthHigh) << 8);
uchar	ucHeaderLength	=	(spIPHeader->ucVersHlen & 0x0F) << 2;

	return uiTotalLength - ucHeaderLength;
}
/*---------------------------------------------------------------------------*/
/*	Header & Checksum-Tools */
/*---------------------------------------------------------------------------*/
/*	IP-Checksum Calculation */
uint	ip_checksum16bit(uint uiChecksum, uint uiValue){
/*	One 16-bit word will be added to the checksum-calculation */
ulong	ulSum	=	uiChecksum;
/*	Calculate the header-checksum:
	treat the header as a array of 16-bit integers (signed)
	adding, one's complement, taking the one's complement of the result */
	ulSum		+=	(ulong)uiValue;

   /* Now add the upper bits on the lower as long as there are overflows */
    while ((ulSum >> 16)!= 0) {
        ulSum = ((ulSum & 0xFFFF) + (ulSum>>16));
    }
/*	This algorithm is from the internet!!! */
	return	(uint)ulSum;
}
/*---------------------------------------------------------------------------*/
/*	IP-Checksum Calculation */
uint	ip_constchecksum(uint uiChecksum, uint uiValue, uint uiLength){
/*	If a message contains some words of constant data, this can 
	be cosidered within the checksum using this function */
ulong	ulSum	=	uiChecksum;
/*	Calculate the header-checksum:
	treat the header as a array of 16-bit integers (signed)
	adding, one's complement, taking the one's complement of the result */
	ulSum		+=	((ulong)uiValue) * ((ulong)uiLength);

   /* Now add the upper bits on the lower as long as there are overflows */
    while ((ulSum >> 16)!= 0) {
        ulSum = ((ulSum & 0xFFFF) + (ulSum>>16));
    }
/*	This algorithm is from the internet!!! */
	return	(uint)ulSum;
}
/*---------------------------------------------------------------------------*/
/*	IP-Checksum Calculation */
uint	ip_checksum(uint uiChecksum, void* pBuffer, uint uiLength){
/*	This function is used to calculate an IP-like checksum. To
	calculate the number of diffrent Buffers, a checksum
	from a previous calculation of other data in the same
	packet can be given (zero otherwise) 
	
	ATTENTION:	An odd length is only allowed at the End!!!
	ATTENTION:	The Final Checksum is the 1-Complement of this !!!*/
ulong	ulSum	=	uiChecksum;
uchar*	ucpPos	=	(uchar*)pBuffer;
uchar	ucVal0;
uchar	ucVal1;
/*	Calculate the header-checksum:
	treat the header as a array of 16-bit integers (signed)
	adding, one's complement, taking the one's complement of the result */
	while(uiLength > 1){
		ucVal0	=	*ucpPos++;
		ucVal1	=	*ucpPos++;
		ulSum	+=	((((uint)ucVal0) << 8) + (ucVal1 << 0));
		uiLength	-= 2;
	}
/*	If there is an odd byte remaining, pad with zero */
	if(uiLength != 0){
		ulSum	+=	(((uint)*ucpPos)<<8);
	}

   /* Now add the upper bits on the lower as long as there are overflows */
    while ((ulSum >> 16)!= 0) {
        ulSum = ((ulSum & 0xFFFF) + (ulSum>>16));
    }
/*	This algorithm is from the internet!!! */
	return	(uint)ulSum;
}
/*---------------------------------------------------------------------------*/
/*	IP-Checksum Test */
uint	ip_verifychecksum(IPHEADER* pHeader){
/*	Calculates the Header-Checksum and verifies if everything is ok:
	returns 0 if ok, something else otherwise */
/*	Get the Headerlength *2 */
uchar	ucLen	=	(pHeader->ucVersHlen & 0xF0)>>3;
uchar	i;
ulong	ulTmp	= 0;
uint*	uipData	= (uint*)pHeader;
/*	Calculate the Checksum */
	for(i=0;i<ucLen;i++){
		ulTmp	+=	*uipData;
		uipData++;
	}
	return	(uint) ~((ulTmp && 0xFFFF) + (ulTmp>>16));
}
/*---------------------------------------------------------------------------*/
/*	UDP-Pseudoheader Checksum */
uint	udp_pseudoheaderchecksum(IPADDRESS* spSourceIP,IPADDRESS* spDestIP,uint uiLength){
uint	uiChecksum	= 0;
/* ucZero = 0, ucProto = 17, UDP */
uint	uiTmp0	=	17 << 8;
/*	Length of the UDP-Packet */
uint	uiTmp1	=	((uint)HI(uiLength+sizeof(UDPHEADER))) + (((uint)LO(uiLength+sizeof(UDPHEADER)))<<8);
/*	Calculate the Checksum due to the Pseudo-Header */
	uiChecksum	=	ip_checksum(uiChecksum,spSourceIP,sizeof(IPADDRESS));
	uiChecksum	=	ip_checksum(uiChecksum,spDestIP,sizeof(IPADDRESS));
	uiChecksum	=	ip_constchecksum(uiChecksum,uiTmp0,1);
	uiChecksum	=	ip_constchecksum(uiChecksum,uiTmp1,1);
	return	uiChecksum;
}
/*---------------------------------------------------------------------------*/
/* IP-Header */
void*	ip_writeheader(void* pHeader,char cProtocol,uint uiDataLength,IPADDRESS* spSourceIp,IPADDRESS* spDestinationIp,uint uiIdentification){
/*	Write the Ipheader */
IPHEADER*	spIPHeader	= pHeader;
uint		uiChecksum	= 0;
/*	upper 4 Bits: VERS (ipv4), 
	lower 4 Bits: HeaderLength: 5*32-Bit, no options used */
spIPHeader->ucVersHlen				=	0x45;	
/*	set the priority of the traffic: nothing special */
spIPHeader->ucServiceType			=	0;	
/*	write the length of the datagram (in octets): header and data */
spIPHeader->ucTotalLengthLow		=	LO(uiDataLength + sizeof(IPHEADER));
spIPHeader->ucTotalLengthHigh		=	HI(uiDataLength + sizeof(IPHEADER));
/*	write the Identification */
spIPHeader->ucIdentificationLow		=	LO(uiIdentification);
spIPHeader->ucIdentificationHigh	=	HI(uiIdentification);
/*	Fragmentation Control:
	FRAGMENT OFFSET (bit.3-15)
	FLAGS (bit.0-2): 	bit.0=0
						bit.1= DoNotFragmentBit
						bit.2= LastFragmentBit
						*/
spIPHeader->ucFlagsFragmentOffsetLow	= LO(0);
spIPHeader->ucFlagsFragmentOffsetHigh	= HI(0);
/*	TTL	*/
spIPHeader->ucTTL					=	IPTTL;
/*	Protocol */
spIPHeader->ucProtocol				=	cProtocol;
/*	Ip-Addresses */
spIPHeader->sSourceIpAddress		=	*spSourceIp;
spIPHeader->sDestinationIpAddress	=	*spDestinationIp;
/*	Header Checksum */
spIPHeader->ucHeaderChecksumLow		=	LO(0);//for calculation
spIPHeader->ucHeaderChecksumHigh	=	HI(0);
/*	Calculate the  ip-header */
uiChecksum	= ~ip_checksum(uiChecksum,spIPHeader,sizeof(IPHEADER));
spIPHeader->ucHeaderChecksumLow		=	LO(uiChecksum);
spIPHeader->ucHeaderChecksumHigh	=	HI(uiChecksum);
/*	Return a pointer to the Beginning of the data */
	return	(char*)pHeader + sizeof(IPHEADER);
}
/*---------------------------------------------------------------------------*/
/* UDP-Header */
void*	udp_writeheader(void* pHeader,uint uiSourcePort,uint uiDestinationPort,uint uiLength){
/*	Write an UDP-Header. The checksum is set to zero and has
	to be adjusted afterwards properly */
UDPHEADER*	pUDPHeader	=	(UDPHEADER*)pHeader;
	pUDPHeader->ucUDPSourcePortHigh			=	HI(uiSourcePort);	
	pUDPHeader->ucUDPSourcePortLow			=	LO(uiSourcePort);	
	pUDPHeader->ucUDPDestinationPortHigh	=	HI(uiDestinationPort);
	pUDPHeader->ucUDPDestinationPortLow		=	LO(uiDestinationPort);
	pUDPHeader->ucUDPMessageLengthHigh		=	HI(uiLength+sizeof(UDPHEADER));
	pUDPHeader->ucUDPMessageLengthLow		=	LO(uiLength+sizeof(UDPHEADER));
	pUDPHeader->ucUDPChecksumHigh			=	HI(0);
	pUDPHeader->ucUDPChecksumLow			=	LO(0);
/*	Return a Pointer to the End of the Header */
	return	((char*)pHeader)+sizeof(UDPHEADER);
}
/*---------------------------------------------------------------------------*/
/* UDP-Checksum */
void	udp_writechecksum(UDPHEADER* pUDPHeader, uint uiChecksum){
	pUDPHeader->ucUDPChecksumHigh			=	HI(uiChecksum);
	pUDPHeader->ucUDPChecksumLow			=	LO(uiChecksum);
}
/*****************************************************************************/
/* UDP-Packet */
/*****************************************************************************/
void	udp_receivepacket(void){
/*	The underlying protocol has a package for UDP */
uint		uiReadLength;
uint		uiSourcePort;
uint		uiDestinationPort;
UDPHEADER	sUDPHeader;
UDPHEADER*	spHeader	= &sUDPHeader;
	uiReadLength		= ethernet_readframe(spHeader,sizeof(UDPHEADER));
	if(uiReadLength	!=	sizeof(UDPHEADER)){
		/*	The received package doesn't contain a full header - drop */
		return;
	}
/*	We have an udp-header to process */

/*	Extract the Ports */
	uiSourcePort		=	spHeader->ucUDPSourcePortLow + \
		(((uint)spHeader->ucUDPSourcePortHigh)<<8);
	uiDestinationPort	=	spHeader->ucUDPDestinationPortLow + \
		(((uint)spHeader->ucUDPDestinationPortHigh)<<8);

/*	Check if it's DHCP-Stuff */
	if(uiSourcePort	== 67 && uiDestinationPort == 68){
		/*	It's an DHCP Answer */
		dhcp_receivepackage();
	}
	return;
}
/*****************************************************************************/
/*	IP-Packet	*/ 
/*****************************************************************************/
void	ip_receivepackage(void){
/*	The underlying hardware has received a package for ip and this has to be
	processed */
uint		uiReadLength;
uint		uiIpHeaderLength;
uchar		uiIpOptionsLength;


IPHEADER*	spHeader	= (void*)(((uchar*)&gu8NetworkBuffer)+sizeof(ETHERNETHEADER));
/*	Read the IP-Header */
	uiReadLength	=	ethernet_readframe(spHeader,sizeof(IPHEADER));
	if(uiReadLength	!=	sizeof(IPHEADER)){
		/*	The received package doesn't contain a full header - drop */
		return;
	}
/*	We have an ip-header to process */
	if((spHeader->ucVersHlen >> 4) != 4){
		/*	Invalid Version */
		return;
	}
/*	Calculate the Length of the header:
	The HLEN-Field gives the length in 32-bit-words */
	uiIpHeaderLength	= ((uint)(0xF & spHeader->ucVersHlen));
/*	Calculate the Length of options: they will be ignored */
	uiIpOptionsLength	= uiIpHeaderLength - sizeof(IPHEADER)/4;
	ethernet_dropframecontents(uiIpOptionsLength*4);

/*	Decide, wich protocol to invoke */
	switch(spHeader->ucProtocol){
		case	17:
			/*	case of UDP */
			udp_receivepacket();
			break;
		case	1:
			/*	case of IMCP */
			icmp_receivepacket();
			break;
		case	6:
			/*	case of TCP */
			tcp_receivepackage();
		default:
			/*	The protocol is not supported - drop the packet */
			break;
	}
	return;
}
/*****************************************************************************/
