/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include <assert.h>
#include <inttypes.h>
#include <string.h>
/*	If you get a warning from delay.x, ensure that the
	CPU frequency is set in the AVR-Studio Project settings */
#include <util/delay.h>
#include "global.h"
#include "inttypes.h"
#include "system.h"
#include "spi.h"
#include "sd.h"
#include "fat.h"
#include "vs10xx.h"
#include "enc28j60.h"
#include "ethernet.h"
#include "network.h"
#include "ip-udp.h"
#include "tcp.h"
TASKDECLARE(maintask);

TIMER	idle_timer;
TIMER	idle_time_printer_timer;

volatile	u32		u32idletime_calc	= 0;		//Register used to calculate the idle-time
volatile	u8		u8idletime			= 0;		//Last Value for the IDLE-time (1000: IDLE, 0: BUSY)


//-----------------------------------------------------------------------------
//Internal main / Idle-Task
//-----------------------------------------------------------------------------
int	main(){
	/*	Configure the ENC28J60 */
	ENC_CONFIGURE_CS
	ENC_CONFIGURE_INT

	VS10xx_CONFIGURE_XDCS
	VS10xx_CONFIGURE_XCS
	VS10xx_CONFIGURE_XRESET

	/* Rest is done by the SPI-Code */
	PORTA	|= _BV(1);
	SD_DESELECT
	SD_SELECT_INIT
	
	PORTB 	|= _BV(4);
	DDRB	|= _BV(4);
	/*	Configure the SCLK-Pin as (LOW) Output */
	DDRB	|= _BV(7);
	/*	Configure the MOSI-Pin as (LOW) Output */
	DDRB	|= _BV(5);


//-----------------------------------------------------------------------------
//Configure the Kernel
//-----------------------------------------------------------------------------
//Timer
gpTimerWaitQueue= 0;							//No timer yet
gpTimerRunQueue	= 0;							//No callback-function on the execute-queue
glTime			= 0;							//Global time is zero
glAlertTime		= TIMER_STOPPED;				//No timer
gcTimerStatus	= TIMERSTATUS_WAITING;			//Timer-Status
/* This main-routine will be the idle-task. It`s structure has to be filled out */
gIdleTask.pNextTask				= 0;			//No other task so far
/* Setup the task`s timer. It`s used for timeouts while the task waits for an event */
gIdleTask.Timer.pNextTimer		= 0;			//Not queued
gIdleTask.Timer.pCallback = eventtimer;			//This is the callback-handler for the event-timer
//the time is not important, it will be setup when used
gIdleTask.pEvent = (EVENT*)EVENT_UNSIGNALED;	//This points to the event the task is waiting for if any 
gIdleTask.cPriority	= IDLETASK_PRIORITY;		//The priority of the idle-task: has to be the lowest!!!
//gIdleTask.pStack	Will be updated when suspended
gIdleTask.cErrno				= 0;			//no error yet
gIdleTask.uSignature			= TASKSIGNATURE;//The Signature: used to check after a task change if the task`s stack has grown to much
/* We have a idle-task now. It`s the running one and it`s on the shedule-queue */
gpSheduleQueue	= &gIdleTask;
gpRunningTask	= &gIdleTask;
//-----------------------------------------------------------------------------
//System-Timer Configuration
//-----------------------------------------------------------------------------
/*	ATmega644 Datasheet */
/*	TCCR0A – Timer/Counter Control Register A	*/
/*                                              Timer/Counter
                                                Mode of                 Update of  TOV Flag
                                                                                  Set on(1)(2)
             Mode     WGM2    WGM1    WGM0      Operation           TOP  OCRx at
               0        0        0       0      Normal             0xFF Immediate    MAX
                                                PWM, Phase
               1        0        0       1                         0xFF   TOP      BOTTOM
                                                Correct
               2        0        1       0      CTC                OCRA Immediate    MAX
               3        0        1       1      Fast PWM           0xFF   TOP        MAX
               4        1        0       0      Reserved             –     –           –
                                                PWM, Phase
               5        1        0       1                         OCRA   TOP      BOTTOM
                                                Correct
               6        1        1       0      Reserved             –     –           –
               7        1        1       1      Fast PWM           OCRA   TOP        TOP */      
/* (normal mode)      COM0A1       COM0A0 Description
                      0            0      Normal port operation, OC0A disconnected.
                      0            1      Toggle OC0A on Compare Match
                      1            0      Clear OC0A on Compare Match
                      1            1      Set OC0A on Compare Match */
/*	We want the CTC mode */
	TCCR0A	|= 	_BV(WGM01);
/*	TCCR0B – Timer/Counter Control Register B
	CS02	CS01	CS00 	Description
	0 		0 		0 		No clock source (Timer/counter stopped)
	0 		0 		1 		clkT0S/(No prescaling)
	0 		1 		0 		clkT0S/8 (From prescaler)
	0 		1 		1 		clkT0S/32 (From prescaler)
	1 		0 		0		clkT0S/64 (From prescaler)
	1 		0		1		clkT0S/128 (From prescaler)
	1 		1 		0		clkT0S/256 (From prescaler)
	1 		1 		1		clkT0S/1024 (From prescaler) */
/*	We want a 64 prescaler */
	TCCR0B	|= 	_BV(CS02);
	OCR0A	=	(20000/(2*64))-1;	//Clear Timer after 1ms
/*	Enable the Interrupt */	
	TIMSK0	|= _BV(OCIE0A);
//-----------------------------------------------------------------------------
/*	TASK-Setup */
//-----------------------------------------------------------------------------
/*	Timer-Task	*/
gpTimerTask	=	InitTask(timertask_mem,TIMERTASK_MEMSIZE,&timertask,TIMERTASK_PRIORITY);
/*	Main-Task
	This is the Task that replaces the ordinary C-main entry-function */
gpMainTask	=	InitTask(maintask_mem,MAINTASK_MEMSIZE,&maintask,MAINTASK_PRIORITY);
//Idle-Task--------------------------------------------------------------------
/*	It's important to have the interrupts enabled here! Otherwise the cpu
	would hang if all tasks are blocked!!! Every task should enable the
	Interrupts as soon as possible to make it interruptable */
	EnableInterrupts();
idleloop:
/*	From now on this is the idle-task. This code is executed whenever no other 
	task is able to run. */
/*	We want to measure the IDLE-Time of the cpu. */
	while(1){
//	_delay_ms(1);
//	u32idletime_calc++;
/* Direct time measurement vie assembler-code:
 * 	we neglect the time needed to jump in the loop, etc */
/* Count up to 2000 (0x7D0). We have 8 Clocks in each loop,
 * this means 0.5us if running at 16Mhz. thus we waste 1ms in each loop */
		asm volatile(
				"ldi  r26,0xD0" "\n\t"
				"ldi  r27,0x07" "\n\t"
				"mov  r2,r26" "\n\t"
				"mov  r3,r27" "\n\t"
				"mov  r26,__zero_reg__" "\n\t"
				"mov  r27,__zero_reg__" "\n\t"
				"L_dl1%=:" "\n\t"
				"nop" "\n\t"
				"nop" "\n\t"
				"adiw  r26,1" "\n\t"
				"cp  r26,r2" "\n\t"
				"cpc  r27,r3" "\n\t"
				"brne L_dl1%=" "\n\t"
		             :
		             :
		            );
		u32idletime_calc++;
	}
		goto idleloop;
}
/*****************************************************************************/
//-----------------------------------------------------------------------------
//	IDLE-Time measurement
//-----------------------------------------------------------------------------
void	idle_timer_callback(TIMER* pTimer){
/*	CALLBACK-Funktion for the idle-time calculation */
	RestartTimer(pTimer,1000);
/*	Calculate the idle time in percent */
	EnterCritical();	//CRITICAL SECTION
	u8idletime			=	(u8)(u32idletime_calc / 10);
	u32idletime_calc	=	0;
	LeaveCritical();	//CRITICAL SECTION	
}
//-----------------------------------------------------------------------------


/*****************************************************************************/
/*	Audio playback from tcp-socket */
/*****************************************************************************/
s8		vs10xx_playfromsocket(TCPTCB* spTCB){
/*	DANGER: It is recommended that the RxBuffer is a multiple of 32 Byte and at least
 * 	32 Byte long!!! */
/*	Play from the Socket */
u8		u8i,u8j;
u8*		u8pReadPos;
u16		u16Available;
u16		u16Count;
u8 		u8PlayerAction;
s8		s8Return;
/*	Play endless loop */
while(1){
	while(1){
		/*	Get the number of bytes in the Buffer */
		EnterCritical();
		u16Available		= tcp_rxbytefull(spTCB);
		LeaveCritical();
		if(u16Available<=128){
			/*	Reset the Timer: We  send now */
			spTCB->s8TCPTxTimer	= 0;
			/*	Reset the Event */
			EVENTINIT(spTCB->sTCPRxEvent);
			tcp_queuetxtcb(spTCB);
			SignalEvent(&NetworkEVENT);
		}
		/*	We can play if we have at least 32 byte */
		if(u16Available >= 128)	break;
		/*	We have not enough data in the RX Buffer */
		/*	Check if the Connection is still alive */
		if((spTCB->u8TCPFlags & TCPFLAG_RXEOF) != 0){
			/*	The Remote End has signaled us EOF!!! */
			s8Return		= -1;
			goto	end;
		}
		if((spTCB->u8TCPFlags &	TCPFLAG_RESET) != 0){
			/*	The Connection is reseted */
			s8Return		= -1;
			goto	end;
		}
		if(WaitEventTimeout(&spTCB->sTCPRxEvent,TCP_MAXWRXAITTIME) == EVENT_TIMEOUT){
			/*	The Event timed out! */
			s8Return		= -1;
			goto	end;
		}	
	}
	if((PINA & _BV(1))== 0){
		goto	end;
	}
	EnterCritical();
	/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
	/*	We can adjust the Bus as we want it */
	SPCR0	= 	_BV(SPE0) | _BV(MSTR0) ;
	/*	Set SPI2X = 0!! (diffrent register!)  */
	SPSR0	= _BV(SPI2X);
	VS10xx_DESELECT_XDCS
	for(u8j=0;u8j<4;u8j++){
		/*	We have at least 32 Byte of data, send them to the VS10xx */
		EnterCritical();
		u8pReadPos	= spTCB->u8pRxBufferRead;
		LeaveCritical();
		/*	At the beginning of the stream (or if the buffer is not a
		 * 	multiple of 32 Bytes) we are not alligned, check this */
		u16Count	= (u16)tcp_rxbufferend(spTCB) - (u16)u8pReadPos;
		if(u16Count > 32){
			/*	If we have more than 32 Byte, restrict us nevertheless to
			 * 	send only 32!!! */
			u16Count	= 32;
		}
		/*	Play the Data */
		while(VS10xx_DREQ_STATUS == 0){
			/*	As long as the chip does not want data, we will wait */
			WaitEvent(&gVS10xxDREQEvent);
		}
		/*	Send our 32 Byte */
		for(u8i=0;u8i<u16Count;u8i++){
			spi_rw(*u8pReadPos);
			u8pReadPos++;
		}
		/*	Adjust the TCB */
		if(u8pReadPos == tcp_rxbufferend(spTCB)){
			/*	Test if we reached the end of the buffer */
			u8pReadPos	= tcp_rxbufferstart(spTCB);
		}
		/*	Write the new read Position back, update the number of free bytes */
		EnterCritical();
		spTCB->u8pRxBufferRead	=  u8pReadPos;
		spTCB->u16RxBufferFree	+=	u16Count;
		LeaveCritical();
	}
	VS10xx_SELECT_XDCS
	/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
	LeaveCritical();


}	
/*	We finished playing, close the connection */
end:
/*	close the connection */
	tcp_closetx(spTCB);
	tcp_closerx(spTCB);
	return	s8Return;
}
/*****************************************************************************/



/*****************************************************************************/
/*	Audio playback from stdio */
/*****************************************************************************/
s16		vs10xx_playfromfile(FATFILE* spFile){
/*	Communication is done via the global
 * 	variable gu8PlayerAction and the gPlayerUIEvent - Event */
s16		s16chr;
u8		u8i,u8j;
u8*		u8pReadPos;
u8 		u8PlayerAction;
u8*		u8pbuffer	=_malloc(512);
	while(1){
		/*	REMARK: The RX-Buffer should be a multiple of 32 Byte to get
		 * 	best performance */
		if((PINA & _BV(1))== 0)	goto	end;
		/*	Fill our buffer */
		if(fat_readfile(spFile,u8pbuffer,512)!=512){
			goto	end;
		}
		/*	Wait until the SPI-Bus is ready */
		WaitEvent(&gsSpiLockEvent);
		SPCR0	= 	_BV(SPE0) | _BV(MSTR0) ;
		/*	Set SPI2X = 0!! (diffrent register!)  */
		SPSR0	= _BV(SPI2X);
		VS10xx_DESELECT_XDCS
		/*	Send Data */
		u8pReadPos	= u8pbuffer;
		for(u8j=0;u8j<16;u8j++){
			while(VS10xx_DREQ_STATUS == 0){
				WaitEvent(&gVS10xxDREQEvent);
			}
			/*	Send our 32 Byte */
			for(u8i=0;u8i<32;u8i++){
				spi_rw(*u8pReadPos);
				u8pReadPos++;
			}
		}
		VS10xx_SELECT_XDCS
		/*	Release the SPI-Bus */
		SignalEvent(&gsSpiLockEvent);

	}//END Playloop
end:
	_free(u8pbuffer);
	return	0;
}
/*****************************************************************************/
/* http stuff */
/*****************************************************************************/
s16		http_readheader(FILE* spStdIoSocket){
/*	Read the http-header */
s16		s16Chr;
	while(1){
		s16Chr = fgetc(spStdIoSocket);
		if(s16Chr == 0x0D){
			s16Chr = fgetc(spStdIoSocket);
			if(s16Chr == 0x0A){
				s16Chr = fgetc(spStdIoSocket);
				if(s16Chr == 0x0D){
					s16Chr = fgetc(spStdIoSocket);
					if(s16Chr == 0x0A){
						break;
					} 
				} 
			} 
		}
		if(s16Chr == EOF){
			return	EOF;
		}
	}
		return	0;
}
/*****************************************************************************/
/*	*/
/*****************************************************************************/
void	removetermination(s8* s8pString){
/*	Removes the Termination (newline/carrige return from the string given */
s8*	s8pstr;
	s8pstr	= strchr(s8pString,'\n');
	if(s8pstr != 0)	*s8pstr = 0;
	s8pstr	= strchr(s8pString,'\r');
	if(s8pstr != 0)	*s8pstr = 0;	
}


TASKDECLARE(playertask);

s8	s8ShortName[13];
//u16	u16WStr[FAT_MAXFILELENGTH];
s8	s8URL[256];
s8	s8SessionID[64];
s8	s8ArtistURL[64];
//s8	s8URL[256];
//s8	s8SessionID[64];
//s8	s8ArtistURL[64];
//u8	u8test[sizeof(ETHERNETHEADER)];
s8	s8wav[5]	= ".wav";
s8	s8mp3[5]	= ".mp3";
s8	s8ogg[5]	= ".ogg";
s8	s8m3u[5]	= ".m3u";
s8	s8lfm[5]	= ".lfm";
//-----------------------------------------------------------------------------
//Main-Task
//-----------------------------------------------------------------------------
TASKBODY(maintask){
/* This is the system's main-task. The code-execution starts here */
		spi_init();
		vs10xx_init();	
		if(sd_init()!=0){
			while(1);
		}			//SD-Card
	/*	Init the Filesystem */
		if(fat_init()!=0){
			while(1);
		}
	
		/*	Network-Task */
		gpNetworkTask	= InitTask(networktask_mem,NETWORKTASK_MEMSIZE,&networktask,NETWORKTASK_PRIORITY);

		
		
		
		
		
		
			
		gu8NetworkSend	= 1;
		/*	Signal the tx-task that something's to do */
		SignalEvent(&NetworkEVENT);

/*	
	ethernet_writeheader(u8buffer,&mymac,&broadmac,0x0800);
	ethernet_waittxready(sizeof(ETHERNETHEADER)+1000);
	ethernet_senddata(u8buffer,sizeof(ETHERNETHEADER));
	ethernet_sendbyte(0x00,1000);
	ethernet_waittxdone();
*/
	EnableInterrupts();
	PORTA	|= _BV(1);

	IPADDRESS	sRemoteIP;
	TCPPORT		sRemoteTCPPort;
	SETIPADDRESS(sRemoteIP,192,168,178,1);
	SETTCPPORT(sRemoteTCPPort,8080);
	TCPTCB*		spTCBhttp;
	FILE*	spFileHttpSocket	= fdevopen(&tcp_put,&tcp_get);
	s16			s16Chr;
	s16		s16chr;
	u16		u16i,j;
	u8*		u8pReadPos;
	u8 		u8PlayerAction;
	

		
FATFILE		fatfile;
FATFILE		sDir;
FILE*		spStdIoFile;
u8			u8typ;
s8*			s8pstr;
/*	DANGER: We using the same space... */
s8*			s8pUserName		= s8SessionID;
s8*			s8pPassword		= s8URL;
	
while(1){
	fat_openroot(&sDir);
/*	We browse the directory structure. Depending on the file extension we will
	do one of the following actions:
	.mp3, .ogg: Play the file
	.m3u:	try to connect to the webstream
	.lfm:	try to connect to lastfm service
*/
	while(1){
next:
		u8typ		= fat_nextentry(&sDir);

		if(u8typ == FAT_EOC)	break;
		if(u8typ == FAT_FILE){
			/*	Sleep one second */
				sleep(1000);
				fat_getshortfilename(&sDir,s8ShortName);
				/*	Convert to lower-case letters */
				strlwr(s8ShortName);
				fat_openfile(&sDir,&fatfile);
				vs10xx_reset();	
				/*	Decide what to do */
				if(strstr(s8ShortName,s8wav)!=0){
					/*	Play wav-File */
					vs10xx_playfromfile(&fatfile);
				}else if(strstr(s8ShortName,s8mp3)!=0){
					/*	Play mp3-File */
					vs10xx_playfromfile(&fatfile);
				}else if(strstr(s8ShortName,s8ogg)!=0){
					/*	Play ogg-File */
					vs10xx_playfromfile(&fatfile);					
				}else if(strstr(s8ShortName,s8m3u)!=0){
					/*	Play m3u-File */
					/*	We open the File for StdIo-Access */
							spStdIoFile	= fdevopen(0,&fat_get);
							fdev_set_udata(spStdIoFile,&fatfile);
					/*	Now we read the web-address */
							fgets(s8URL,256,spStdIoFile);
					/*	Remove line terminations */
							removetermination(s8URL);
					/*	Open a Connection to the Proxy */
					spTCBhttp	= tcp_open(&sRemoteIP,&sRemoteTCPPort);
					/*	Configure the Socket for StdIo-Access */
					fdev_set_udata(spFileHttpSocket,spTCBhttp);
					/*	Send the http-querry to the remote server */
			        fprintf_P(spFileHttpSocket,PSTR("GET %s HTTP/1.1\r\nConnection: Close\r\n\r\n"),s8URL);
			        /*	Ensure that the whole querry gets send */
					tcp_flush(spTCBhttp);
					/*	Read in the Response-Header */
						if(http_readheader(spFileHttpSocket) == EOF){
								/*	close the connection */
								tcp_closetx(spTCBhttp);
								tcp_closerx(spTCBhttp);
								goto	next;
						}
					/*	Play the file */
						vs10xx_playfromsocket(spTCBhttp);
				}else if(strstr(s8ShortName,s8lfm)!=0){
					/*	We should connect to the lastfm service */
					/*	We open the File for StdIo-Access */
							spStdIoFile	= fdevopen(0,&fat_get);
							fdev_set_udata(spStdIoFile,&fatfile);
					/*	Now we read the username and password */
							fgets(s8pUserName,64,spStdIoFile);
							removetermination(s8pUserName);					
							fgets(s8pPassword,128,spStdIoFile);
							removetermination(s8pPassword);					
						/*	Send the http-querry to the remote server: try to get
						 * 	a Session */
					/*	Open a Connection to the Proxy */
					spTCBhttp	= tcp_open(&sRemoteIP,&sRemoteTCPPort);
					/*	Configure the Socket for StdIo-Access */
					fdev_set_udata(spFileHttpSocket,spTCBhttp);
					/*	Send the http-querry to the remote server */
					fprintf_P(spFileHttpSocket,PSTR("GET http://ws.audioscrobbler.com/radio/handshake.php?version=1.1.1&platform=linux&username=%s&passwordmd5=%s&debug=0&partner= HTTP/1.1\r\nConnection: Close\r\n\r\n"),s8pUserName,s8pPassword);
			        /*	Ensure that the whole querry gets send */
					tcp_flush(spTCBhttp);
					/*	Read in the Response-Header */
						if(http_readheader(spFileHttpSocket) == EOF){
								/*	close the connection */
								tcp_closetx(spTCBhttp);
								tcp_closerx(spTCBhttp);
								goto	next;
						}
					/*	Now we have to scan the response for the session-id and the url of the stream */
					/*	Read the session-ID first we read session=, then the ID */
						fgets(s8SessionID,9,spFileHttpSocket);
						fgets(s8SessionID,64,spFileHttpSocket);
						removetermination(s8SessionID);		
					/*	Now read the stream-URL (first we read stream_url=, then the url)*/
						fgets(s8URL,12,spFileHttpSocket);
						fgets(s8URL,256,spFileHttpSocket);
						removetermination(s8URL);		
					/*	Read the rest of the data */
						while(fgetc(spFileHttpSocket)!= EOF);
					/*	Close the connection */
						tcp_closetx(spTCBhttp);
						tcp_closerx(spTCBhttp);
				/*	Tune to a similar Artist Station */
					/*	Read the Artist string in */
						fgets(s8ArtistURL,64,spStdIoFile);
						removetermination(s8ArtistURL);								
					/*	try to tune this artist */
						/*	Open a Connection to the Proxy */
						spTCBhttp	= tcp_open(&sRemoteIP,&sRemoteTCPPort);
						/*	Configure the Socket for StdIo-Access */
						fdev_set_udata(spFileHttpSocket,spTCBhttp);
						/*	Send the http-querry to the remote server */
						fprintf_P(spFileHttpSocket,PSTR("GET http://ws.audioscrobbler.com/radio/adjust.php?session=%s&url=lastfm://artist/%s/similarartists&debug=0 HTTP/1.1\r\nConnection: Close\r\n\r\n"),s8SessionID,s8ArtistURL);
				        /*	Ensure that the whole querry gets send */
						tcp_flush(spTCBhttp);
						/*	Read in the Response-Header */
							if(http_readheader(spFileHttpSocket) == EOF){
									/*	close the connection */
									tcp_closetx(spTCBhttp);
									tcp_closerx(spTCBhttp);
									goto	next;
							}
						/*	Read in the first line of the response */
						fgets(s8SessionID,64,spFileHttpSocket);
						removetermination(s8SessionID);		
						/*	Read the rest of the data */
							while(fgetc(spFileHttpSocket)!= EOF);
						/*	Close the connection */
							tcp_closetx(spTCBhttp);
							tcp_closerx(spTCBhttp);
						/*	Play the stream now */
							spTCBhttp	= tcp_open(&sRemoteIP,&sRemoteTCPPort);
							/*	Configure the Socket for StdIo-Access */
							fdev_set_udata(spFileHttpSocket,spTCBhttp);
							/*	Send the http-querry to the remote server */
					        fprintf_P(spFileHttpSocket,PSTR("GET %s HTTP/1.1\r\nConnection: Close\r\n\r\n"),s8URL);
					        /*	Ensure that the whole querry gets send */
							tcp_flush(spTCBhttp);
							/*	Read in the Response-Header */
								if(http_readheader(spFileHttpSocket) == EOF){
										/*	close the connection */
										tcp_closetx(spTCBhttp);
										tcp_closerx(spTCBhttp);
										goto	next;
								}
							/*	Play the file */
								vs10xx_playfromsocket(spTCBhttp);

						
				}
		}
	}
	}//Browse Folder
}



