/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/* network stuff */
/*****************************************************************************/
#include "network.h"
#include "ethernet.h"
#include "arp.h"
#include "icmp.h"
#include "dhcp.h"
#include "ip-udp.h"
#include "tcp.h"
/*****************************************************************************/
/*	Interrupt Routine */
/*****************************************************************************/
ISR(INT1_vect){
/*	Signal the Ethernet-Event */
	CS_IRQ_DISABLE
	/*	We received something */
	gu8NetworkReceive =  1;
	SignalEvent(&NetworkEVENT);
}
u8	u8test[sizeof(ETHERNETHEADER)+sizeof(TCPHEADER)+sizeof(IPHEADER)];
/*****************************************************************************/
/* Network-Task */
/*****************************************************************************/
TASKBODY(networktask){
/*	This is the network-task. It waits for an network-event and procceeds it. */
u16	u16Length;
u16	u16Protocol;
u16	u16status;
u8	u8tmp;
u16	u16tmp;
u16	u16tmp0;
u16	u16tmp1;
u16	u16tmp2;
ETHERNETHEADER*		spEthernetHeader	= (ETHERNETHEADER*)&gu8NetworkBuffer;
//-----------------------------------------------------------------------------
/*	High-Level Network-Setup */
//-----------------------------------------------------------------------------
gu8NetworkTimer		= 0;
gu8NetworkSend		= 0;
gu8NetworkReceive	= 0;
/*	Setup the Event */
	EVENTINIT(NetworkEVENT);
/*	Setup Ethernet */
	ethernet_init();
/*	Setup IP-Layer */
	ip_init();
/*	Setup ARP */
	arp_init();
/*	Setup ICMP */
	icmp_init();
/*	Setup DHCP */
	dhcp_init();
/*	Setup TCP */
	tcp_init();
/*---------------------------------------------------------------------------*/	
/* Wait for something to do */
/*---------------------------------------------------------------------------*/	
while(1){
/*	Enable Interrupts globally */
	EnableInterrupts();
	u8tmp	= enc_readcontrolregister(ENC_EIR);
	u8tmp	= enc_readcontrolregister(ENC_EPKTCNT);
	if(u8tmp == 0){
		/*	We have no receive-event */
		/*	Enable the Network-Controller-Interrupt */
		u8tmp	= enc_readcontrolregister(ENC_EIE);
		if((u8tmp & 0x40) == 0){
				while(1);
		}
		CS_IRQ_ENABLE
		/*	Wait for an Network-Event to occure. */
			WaitEvent(&NetworkEVENT);
	}
/*---------------------------------------------------------------------------*/	
/*	Receive-Handler */
/*---------------------------------------------------------------------------*/	
/*	If the network controller generates an interrupt, the ISR sets the receive-flag */
while(1){
	/*	Check the PKTIF (bit.6) - Flag: if
	 * 	its not set, we dont have any packet to process */
	u8tmp	= enc_readcontrolregister(ENC_EIR);
	u8tmp	= enc_readcontrolregister(ENC_ESTAT);
	u8tmp	= enc_readcontrolregister(ENC_EPKTCNT);

	if(u8tmp == 0)	break;
	/*	We have a new Packet to process: Read it */
	/*	Setup the Read-Pointer */

		enc_writecontrolregister(ENC_ERDPTL,(u8)gu16EthernetNextRxFrame);
		enc_writecontrolregister(ENC_ERDPTH,(u8)(gu16EthernetNextRxFrame>>8));
	/*	Wait until the SPI-Bus is ready */
		WaitEvent(&gsSpiLockEvent);
		/*	Setup the SPI-Speed */
		enc_spiconfig();
		/*	Make the CS-Pin low */
		ENC_DESELECT_CS
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		asm volatile("nop"::);
		/*	Write Instruction : Consists of a 3-Bit Opcode (001: Read buffer memory) and 
		 * 	the 5-Bit code 11010 */
		spi_rw(0x3A);
		/*	Read in the next Packet Pointer, lowbyte first */
		gu16EthernetNextRxFrame	=	spi_rw(0x00);
		gu16EthernetNextRxFrame	|=	(((u16)spi_rw(0x00))<<8);
		/*	Read in the Status Code (4 Byte) */
		gu16EthernetFrameDataLeft	= ((u16)spi_rw(0x00));
		gu16EthernetFrameDataLeft	|= ((u16)spi_rw(0x00))<<8;
		u16status					= ((u16)spi_rw(0x00));
		u16status					|= ((u16)spi_rw(0x00))<<8;
		/*	Save the number of bytes in the frame (including CRC & padding) */
		/*	Check Bit.23: RX-OK (bit15-0 are already read in as the size of the package */
		if(gu16EthernetFrameDataLeft == 0){
				while(1);
		}
		
		if((u16status & 0x0080) == 0){
			u16status--;
			goto	drop_packet;
		}
		/*	Check CRC: bit.20*/
		if((u16status & 0x0010) != 0){
			u16status--;
			goto	drop_packet;
		}
		
		/*	We received a good frame */
		/*	We read the Ethernet-Header in */
		u16Length	= ethernet_readframe(spEthernetHeader,sizeof(ETHERNETHEADER));

		if(u16Length!=sizeof(ETHERNETHEADER)){
			/*	The Packet doesn't contain a full header -> through it away */
			goto	drop_packet;
		}					

			
		/*	We need the High-Level Protocoll */
		u16Protocol	=	(((u16)spEthernetHeader->u8TypeHigh)<<8)\
						+ (u16)spEthernetHeader->u8TypeLow;
		switch(u16Protocol){
			case	0x0800:
				/*	This belongs to IP (code 0x0800) */
				ip_receivepackage();
				break;
			case	0x0806:
				/*	This belongs to ARP (code 0x0806) */
				arp_receivepackage();
				break;
			default:
				/*	We don't know the protocol: drop */
				break;
		}//END switch: Protocol
drop_packet:
	/*	Make the CS-Pin high */
	ENC_SELECT_CS
	/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
	/*	Advance the Packet read-Pointer to the Beginning of the next Packet:
	 * 	DANGER: Errata !!!
	
		11. Module: Memory (Ethernet Buffer)
		The receive hardware may corrupt the circular receive buffer (including the Next Packet Pointer
		and receive status vector fields) when an even value is programmed into the ERXRDPTH:ERXRDPTL registers.
		Work around
		Ensure that only odd addresses are written to the ERXRDPT registers. Assuming that ERXND
		contains an odd value, many applications can derive a suitable value to write to ERXRDPT by subtracting
		one from the Next Packet Pointer (a value always ensured to be even because of hardware padding)
		and then compensating for a potential ERXST to ERXND wraparound. Assuming that the receive
		buffer area does not span the 1FFFh to 0000h memory boundary, the logic in Example 1 will
		ensure that ERXRDPT is programmed with an odd value:
	*/	
	u16tmp	= gu16EthernetNextRxFrame;
	if(u16tmp==0){
		/*	The next Address is zero */
		u16tmp	= ENC_RXND;
	}else{
		u16tmp--;
	}
	enc_writecontrolregister(ENC_ERXRDPTL,(u8)u16tmp);
	enc_writecontrolregister(ENC_ERXRDPTH,(u8)(u16tmp>>8));
	/* In addition to advancing the Receive Buffer Read Pointer, after each packet is fully processed, the host
	controller must write a ‘1’ to the ECON2.PKTDEC bit. Doing so will cause the EPKTCNT register to
	decrement by 1. After decrementing, if EPKTCNT is ‘0’, the EIR.PKTIF flag will automatically be cleared.
	Otherwise, it will remain set, indicating that additional packets are in the receive buffer and are waiting to be
	processed. Attempts to decrement EPKTCNT below 0 are ignored. Additionally, if the EPKTCNT register ever
	maximizes at 255, all new packets which are received will be aborted, even if buffer space is available. To
	indicate the error, the EIR.RXERIF will be set and an interrupt will be generated (if enabled). To prevent this
	condition, the host controller must properly decrement the counter whenever a packet is processed. */
	/*	Write a 1 to bit.6 (PKTDEC-Bit) */
	enc_bitfieldset(ENC_ECON2,0x40);

}//END while: Receive Packages	
/*****************************************************************************/
/* TRANSMIT-HANDLER */
/*****************************************************************************/
tcp_transmit();

	if(gu8NetworkSend == 1){
		gu8NetworkSend = 0;
/*---------------------------------------------------------------------------*/
/*	TCP Transmit */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/*	ARP Transmit */
/*---------------------------------------------------------------------------*/
		arp_transmit();
/*---------------------------------------------------------------------------*/
/* DHCP-Transmit */
/*---------------------------------------------------------------------------*/
		dhcp_transmit();
/*---------------------------------------------------------------------------*/
/* ICMP-Transmit */
/*---------------------------------------------------------------------------*/
		icmp_transmit();
/*****************************************************************************/
	}//End Transmit-Handler
/*****************************************************************************/
}//END while
}//END TASK
/*****************************************************************************/

