/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
#include <string.h>
#include <avr/io.h>
#include <util/delay.h>


#include "inttypes.h"
#include "sd.h"
#include "spi.h"

u8	gu8SdStatus				= SD_LOWCAPACITY;
u8	gu8SdCacheStatus		= SDCARD_CACHECLEAN;
u32	gu32SdCacheSecAdd		= 0;
u8	gu8SdCache[SDCARD_SECTORSIZE];
/*****************************************************************************/
/*	SD/MMC-Card: SPI Settings */
/*****************************************************************************/
void sd_spiconfig(void){
/*	This function is called before the SD Carde access the
 *  SPI Bus */
/*	use the following configuration:
	We set	SPR0 = SPR1 = 0, SPI2X=0 which
	equals 1/4 of the system frequency 
	SPIE = 0: No Interrupts used so far
	SPE	= 1: SPI enabled
	DORD = 0: MSB first 
	MSTR = 1: Master Operation
	CPOL = 0: SCK low when idle
	CPHA = 0: Sample on the leading edge
	(SPI2X = 0): Choose fosc/4 as the SCK-frequency
	SPR1 = 0
	SPR0 = 0 */
	SPCR0	= 	_BV(SPE0) | _BV(MSTR0) ;
/*	Set SPI2X = 0!! (diffrent register!)  */
	SPSR0	= _BV(SPI2X);	
	
}
/*****************************************************************************/
/*	SD/MMC-Card: Send a Command to the Card */
/*****************************************************************************/
s8	sd_sendcmd(u8 u8cmd,u32 u32arg){
/*	We have to own the SPI-Bus !!! */
/*	Sends a command to the SD-Card. In case of an timeout, we return -1.
	Otherwise we will return the R1-Response. Because the R1-Response has
	the highest bit always set to zero, both cases can be easyly distinguished */
/*	We have to send Commands to the SD-Card. Because we use the
	SPI-Mode, we will send them grouped as bytes (danger: Commands
	are transmitted MSB first)
	Structure:
	Bit.47:	0	(Start)
	Bit.46:	1	(Transmission)
	Bit.45-40: 6 Bit Command Index
	Bit.39-8:	32 Bit Command argument
	Bit.7-1:	CRC7
	Bit.0:	1	(End)
	The CRC7 uses the generator Polynomial G(x)= x^7+x^3+1
	M(x) = (first bit) * x^n + ... + (last bit) * x^0
	CRC7 = Remainder((M(x) * x^7)/G(x))
	(Specification, 4.5)
	*/
u8	u8timeout;
s8	s8erg;
/*	In the SD-Spec they say we have to provide clock AFTER a Command.
	But in the WEB there is a lot of code providing it BEFORE. (but after
	the game equals before, dosent it...) */
//	SD_DESELECT
	spi_rw(0xFF);	//8 Dummy-Clocks (Transmitt only 1s: no start-Bit!!!)
//	SD_SELECT

/*	Bit.47-40 (Command and Start/Transmissionbit */
	spi_rw(0x40 | u8cmd);
/*	Bit.39-8  (Argument) */
	spi_rw((u8)(u32arg >> 24));
	spi_rw((u8)(u32arg >> 16));
	spi_rw((u8)(u32arg >> 8));
	spi_rw((u8)(u32arg >> 0));
/*	Bit.7-1 (CRC) and Stop-Bit (CRC has to be valid only for the first
	command (CMD0) SD-Spec 7.2.2
	DANGER: The CMD8 CRC-Verification is always enabled!!! */
	spi_rw(0x95);
/*	We	wait for a response (value != 0xFF)
	DANGER: After a timeout-period we will stop!!! */
	u8timeout	=	11;	//Read a maximum of 10 bytes
	do{
		u8timeout--;
		if(u8timeout == 0){
			return	-1;
		}
		s8erg	= spi_rw(0xFF);
	}while((u8)s8erg == 0xFF);
/*	Responses:
	In SPI-Mode we have diffrent Responses.
	R1: One Byte long, MSB first transmitted
	Bit.0:	Idle State
	Bit.1:	Erase Reset
	Bit.2:	Illegal Command
	Bit.3:	Com CRC Error
	Bit.4:	Erase sequence error
	Bit.5:	Address error
	Bit.6:	Parameter error
	Bit.7:	Always zero

	R3 (OCR register): response length 48 bits
	(manual-rs-mmcv1.0) (SD-Datasheet says the first byte is equal to
		the R1-Response)
	Bit.47:	0
	Bit.46:	0
	Bit.45-40:1, Reserved
	Bit.39-8:	32-Bit OCR-Register
	Bit.7-1:	reserved
	Bit.0:	1	(End-Bit)
	OCR-Register: (Operating Conditions Register)
	Bit.6-0:	Reserved (zero both)
	Bit.7:		1.65-1.95V (High-Voltage=0, Low-Voltage=1)
	Bit.14-8:	2.0-2.6	(zero both)
	Bit.23-15:	2.7-3.6 (one both)
	Bit.30-24:	Reserved (zero both)
	Bit.31		Card power-up status bit
	(SD-Spec says the following:
	Bit.15	2.7-2.8V
	...
	Bit.23	3.5-3.6V
	Bit.30	Card Capacity Status
	Bit.31	CardPowerUp
	*/
	/*	DANGER:
		After we wrote a command, we still have
		to provide a Clock-Signal (See Chapter 4.4. SD-Card Spec (simplified version))
	*/
/*	CRC:
	In SPI-Mode the CRC is optional. But in the first Command
	it looks like we have to provide it (Code in the web says so)
	(card isnt in spi-mode yet)
	*/
	return	s8erg;
}
/*****************************************************************************/
/*	SD/MMC-Card: Write (High Level, through cache!!!) */
/*****************************************************************************/
u8	sd_wr(u32 u32SecAdd,u16 u16Offset,u8* u8pData,u16 u16Length){
/*	The SD-Card is organized in Sectors. We writ to the given Sector at the
	given Offset the Data from the buffer with given Length 
	ATTENTION: u16Offset + u16Length mustnt exceed SDCARD_SECTORSIZE */
u8	u8erg	= 0;
	if(gu32SdCacheSecAdd == u32SecAdd){
		/*	Cache-Hit: We can write the data into the cache:
			MIND: Even if we could write a whole sector of 512
			bytes, we have to write it in the cache to keep
			the data integer */
	}else if((u16Offset == 0) && (u16Length == SDCARD_SECTORSIZE)){
		/*	We can write the data directly from the buffer */
			return	sd_wrsec(u32SecAdd,u8pData);
	}else{
		/*	Cache-Miss: Read the given Sector into the Cache */
		u8erg	= sd_rdtocache(u32SecAdd);
	}
	/*	We can copy the data to the cache now */
		gu8SdCacheStatus	= SDCARD_CACHEDIRTY;
		memcpy(gu8SdCache + u16Offset,u8pData,u16Length);
		return	u8erg;
}
/*****************************************************************************/
/*	SD/MMC-Card: Write a Sector */
/*****************************************************************************/
u8	sd_wrsec(u32 u32add,u8* u8pData){
/*	We want to write a block of data to the card */
u16	i;
u8	u8erg;

	/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
	/*	Configure SPI like we need it */
	sd_spiconfig();
/*	Select the card. */
	SD_SELECT
/*	Send the Command: Single-Block write */
	u8erg	=  sd_sendcmd(24,u32add);
	if(u8erg != 0){
		/*	We have a error-condition */
		goto	end;
	}
/*	We have to wait a while... 
	And from the Hitachi HB28E016MM2... Spec (and Internet-Code) we have to send
	8 Clock-Pulses as well.
	We have to send a Start-Token (7.3.3.2 SD-Spec, SingleBlockRead/Write, MultipleBlockRead)
	0xFE (equals a simple Start-Bit...)
	*/
	for(i=0;i<100;i++){
		spi_rw(0xFF);			//DANGER
	}	

	spi_rw(0xFF);	//Dummy-Clocks...
	spi_rw(0xFE);	//Data-Start-Token
/*	Send the Datablock: 512 bytes */
	for(i=0;i<SDCARD_SECTORSIZE;i++){
		SPDR0 =	*u8pData;
		u8pData++;
		while((SPSR0 & _BV(SPIF0)) == 0);
	}	
/*	We have to append a 16-Bit CRC (is ignored in SPI-mode) */
	spi_rw(0);
	spi_rw(0);
/*	We wait for a Data-Response-Token! */
	do{
		u8erg	= spi_rw(0xFF);
	}while((u8erg == 0xFF) || (u8erg == 0x00));
/*	SD-Spec 7.3.3.1:
	Bit.7-5: Doesnt matter
	Bit.4-2: Status: 010 Accepted, 101 CRC-error, 110 Write-error,
	Bit.0: 1
	*/
	if((u8erg & 0x0F) != 5){
		/*	We have an error */
		u8erg = 0xff;
		goto	end;
	}
/*	Clock-Control (chapter 4.4 SD-Spec): After a write-transaction we have to provide
	8 additional Cycles */
	spi_rw(0xFF);
/*	Everything is ok!! */
	u8erg	= 0;
end:
/*	Deselect SD Interface */
	SD_DESELECT
	/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
	return	u8erg;
}
/*****************************************************************************/
/*	SD/MMC-Card: Read (High Level, through cache!!!) */
/*****************************************************************************/
u8	sd_rd(u32 u32SecAdd,u16 u16Offset,u8* u8pData,u16 u16Length){
/*	The SD-Card is organized in Sectors. We read from the given Sector at the
	given Offset the Data into the buffer with given Length 
	ATTENTION: u16Offset + u16Length mustnt exceed SDCARD_SECTORSIZE */
u8	u8erg	= 0;
	if(gu32SdCacheSecAdd == u32SecAdd){
		/*	Cache-Hit: We can retrive the data from the cache
			MIND: Even if we could read a whole sector of 512
			bytes, we have to read it from the cache to keep
			the data integer and a copy should be more efficient
			nevertheless */
	}else if((u16Offset == 0) && (u16Length == SDCARD_SECTORSIZE)){
		/*	We can read the data directly into the buffer */
			return	sd_rdsec(u32SecAdd,u8pData);
	}else{
		/*	We have to read into the cache first */
		u8erg	= sd_rdtocache(u32SecAdd);
	}
	/*	We can copy the data from the cache now */
		memcpy(u8pData,gu8SdCache + u16Offset,u16Length);
		return	u8erg;
}
/*****************************************************************************/
/*	SD/MMC-Card: Read a sector into the cache */
/*****************************************************************************/
s8	sd_rdtocache(u32 u32SecAdd){
s8	s8erg;
	if(gu8SdCacheStatus	== SDCARD_CACHEDIRTY){
		/*	We have to write back the cache first, because its
			dirty */
		s8erg =	sd_wrsec(gu32SdCacheSecAdd,gu8SdCache);
		/*	Clear the dirty-flag */
		gu8SdCacheStatus	= SDCARD_CACHECLEAN;
	}
	/*	We can can read in the whole block to the cache */
		gu32SdCacheSecAdd	= u32SecAdd;
		s8erg =	sd_rdsec(u32SecAdd,gu8SdCache);
	return	s8erg;
}
/*****************************************************************************/
/*	SD/MMC-Card: Read a whole sector */
/*****************************************************************************/
s8	sd_rdsec(u32 u32SecAdd,u8* u8pData){
u16	i;
u16	u16timeout;
s8	s8erg;
/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
/*	Configure SPI like we need it */
	sd_spiconfig();
/*	Select the card. */
	SD_SELECT
//	for(i=0;i<100;i++){
//		spi_rw(0xFF);			//DANGER
//	}	
/*	We retrive the 512-Byte Block Address */
	if(gu8SdStatus == SD_LOWCAPACITY){
	/*	The Low-Capacity Cards measure in 1-Byte blocks (See 7.3.1.3 Note 10
		of the SD-Spec */
		u32SecAdd	<<= 9;
	}
/*	Send the Command: Single-Block read */
	s8erg	=  sd_sendcmd(17,u32SecAdd);
	if(s8erg != 0){
		/*	We have a error-condition */
		goto	end;
	}
/*	We have to wait a while until the card sends a Data-Token
	(0xFE) DANGER: After a timeout-period we will stop!!! */
	u16timeout	=	10000;	//Read a maximum of 10 bytes
	do{
		if(u16timeout-- == 0){
			s8erg	=	-1;
			goto	end;
		}
		s8erg	= spi_rw(0xFF);
	}while((s8erg == -1) || (s8erg == 0)); //DANGER
/*	Check if it`s an error-token: If everything is succesfull, the card
	returns 0xFE (First Byte Start-Token). Otherwise it returns a Data
	Error-Token */
	if((s8erg & 0xF0) == 0){
		/*	Data Error Token (SD-Spec 7.3.3.3)
			Bit.7-4 : 0
			Bit.3	Out of range
			Bit.2	Card ECC Failed
			Bit.1	CC Error
			Bit.0	Error */
		s8erg	=	-1;
		goto	end;
	}
/*	Read in the block */
	for(i=0;i<SDCARD_SECTORSIZE;i++){
		SPDR0 =	0xFF;
		while((SPSR0 & _BV(SPIF0)) == 0);
		*u8pData	= SPDR0;
		 u8pData++;
	}
/*	We ignore the CRC-Checksum (16-Bit) */
	spi_rw(0xFF);	//Dummy-Clocks...
	spi_rw(0xFF);	//Dummy-Clocks...
/*	Everything is ok!! */
/*	Clock-Control (chapter 4.4 SD-Spec): After a read-transaction we have to provide
	8 additional Cycles */
	spi_rw(0xFF);
	SD_DESELECT
	spi_rw(0xFF);
/*	Everything is ok */
	s8erg	= 0;
end:
/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
	return	s8erg;
}
/*****************************************************************************/
/*	Init */
/*****************************************************************************/
s8	sd_init(void){
/*	The sdcard needs the following configuration:
	We use the maximum SPI-frequency possible 
	SPIE = 0: No Interrupts used so far
	SPE	= 1: SPI enabled
	DORD = 0: MSB first 
	MSTR = 1: Master Operation
	CPOL = 0: SCK low when idle
	CPHA = 0: Sample on the leading edge
	(SPI2X = 0): Choose fosc/4 as the SCK-frequency
	SPR1 = 0
	SPR0 = 0 */
//	SPCR	= 	_BV(SPE) | _BV(MSTR);
/*	Set SPI2X = 1!! (diffrent register!)  */
//	SPSR	= 0;
/*	Setting up the Chip-Select Line */
	SD_SELECT_INIT
	SD_DESELECT
	return	sd_reset();
}
/*****************************************************************************/
/*	SD/MMC-Card: Reset */
/*****************************************************************************/
s8	sd_reset(void){
/*	Chapter 6 (SDCARD-Spec): Powering up the SD-Card: We should supply power to 
	the card within 250 ms. We should supply at least 74 Clock-cycles while
	CMD-line is high (SPI-Mode: CS high as well) Because I dont know if the
	SPI-DataOutput Line is stable during SPI-Operation, we will send the Clock
	-Pulses by Software. The MMC-Spec requires that the Clock is less than 400khz */
s8	s8erg;
u8	i;

/*	Wait until the SPI-Bus is ready */
	WaitEvent(&gsSpiLockEvent);
/*	Configure SPI like we need it */
	sd_spiconfig();
/*	MMCA 4.1 Application Note:
1.1.1Power-up sequence
1.Apply power to the bus, communication voltage range (2.0V to 3.6V)
	*/
/*
2.Set clock to 400kHz, or less
3.Wait for 1msec, then wait for 74 more clock cycles
*/

/*	Disable the SPI */
	SPCR0	&= 	~_BV(SPE0);
/*	DeSelect the SD-Card (CS high) */
	SD_DESELECT
/*	Keep the CMD-Line High */
	SPI_SELECT_MOSI
	
/*	Send the Pulses */
	for(i=0;i<80;i++){
		_delay_ms(2.5);
		SPI_SELECT_CLK
		_delay_ms(2.5);
		SPI_DESELECT_CLK
	}
/*
4.Send CMD0 to reset the bus, keep DAT3 (CS) low during this step
*/
/*	Enable the SPI */
	SPCR0	|= 	_BV(SPE0);
/*	We transmit CMD0:	*/
	do{
	SD_SELECT
	s8erg	= sd_sendcmd(0,0);
	SD_DESELECT
	}while(s8erg != 1);

	if((s8erg != 1) && (s8erg != 0)){
		/*	The card doesnt response with a idle-message. ERROR */
		s8erg	= -1;
		goto	error;
	}
/*
5.Send CMD1, with the intended voltage range in the argument (either High Voltage or Dual Voltage)
6.Receive R3 (Provides the content of the OCR register)
7.Repeat steps 5 and 6 as long as the OCR busy bit is �0�
8.From the R3 response argument the host can learn if the card is a high voltage or dual voltage card.
*/
/*	The SD-Spec says (Chapter 7.2.7: Reset Sequence) That for the thick 2.1mm cards we can use
	CMD1. For other cards ACMD41 has to be used !!!! ACMD41 can also be used to destinguish
	between SD and MMC cards!!! */
	do{
		SD_SELECT
		s8erg	= sd_sendcmd(1,0);
/*	The Result is a R3-Respons (MMC-spec). The first byte is already read in,
	we have to check the other ones
	The SD-Card spec says, that the answere is only a R1-Response.... So
	we check thatL:
	We wait for the IDLE-Bit to vanish... */
		for(i=0;i<5;i++){
			spi_rw(0xFF);
		}
		SD_DESELECT
	}while(s8erg != 0);
/*	Determine if the card is a high-capacity card */
//sd_status	= SD_LOWCAPACITY
//	SD_SELECT
//	u8erg	=  sd_sendcmd(58,0);
//	u8erg	=  spi_rw(0xFF);
//	u8erg	=  spi_rw(0xFF);
//	u8erg	=  spi_rw(0xFF);
//	SD_DESELECT
/*	*/
//	SD_SELECT
//	u8erg	=  sd_sendcmd(55,0);
//	SD_DESELECT
//	SD_SELECT
//	u8erg	=  sd_sendcmd(41,0);
//	SD_DESELECT


/*	Setup the block-length for Block-Data-Transfer: 512 Byte:
	I got an error by just using this command after the CMD0 and
	CMD1 initialization */
	SD_SELECT
	s8erg	=  sd_sendcmd(16,512);
	spi_rw(0xFF);	//Dummy-Clocks...
	SD_DESELECT

/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
/*	Fill the Cache with sector 0 */
/*	Reset done!!! */
	s8erg	= sd_rdsec(0,gu8SdCache);
	return s8erg;
/*	In case of an error */
error:
/*	Release the SPI-Bus */
	SignalEvent(&gsSpiLockEvent);
	return s8erg;
}
/*****************************************************************************/
