/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/* KERNEL */
/*****************************************************************************/
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <assert.h>
#include <string.h>
#include "inttypes.h"
#include "global.h"
#include "system.h"
//-----------------------------------------------------------------------------
//Make malloc Multitasking (-:
//-----------------------------------------------------------------------------
inline	void*	_malloc(size_t size){
void*	pVoid;
		EnterCritical();
		pVoid	= malloc(size);
		LeaveCritical();
		return	pVoid;
}
inline void	_free(void* data){
		EnterCritical();
		free(data);
		LeaveCritical();
}
//-----------------------------------------------------------------------------
//Errno-Functions
//-----------------------------------------------------------------------------
void	SetErrno(char cErrno){
/*	Set the errno of the running task */
	EnterCritical();	//CRITICAL SECTION
	gpRunningTask->cErrno	= cErrno;
	LeaveCritical();	//CRITICAL SECTION
}
char	GetErrno(void){
/*	Get the errno of the running task */
char	cErrno;
	EnterCritical();	//CRITICAL SECTION
	cErrno		= gpRunningTask->cErrno;
	LeaveCritical();	//CRITICAL SECTION
	return	cErrno;
}
//-----------------------------------------------------------------------------
//SwitchTask
//-----------------------------------------------------------------------------
/*This is the main-function of the system: It does the task-switch and checks
  if the switched of task is still "sane". This is highly hardware and 
  compiler dependened. This switching-procedure is working with the winavr
  gcc package */
void	SwitchTask(void){
		asm volatile("push r2"::);
		asm volatile("push r3"::);
		asm volatile("push r4"::);
		asm volatile("push r5"::);
		asm volatile("push r6"::);
		asm volatile("push r7"::);
		asm volatile("push r8"::);
		asm volatile("push r9"::);
		asm volatile("push r10"::);
		asm volatile("push r11"::);
		asm volatile("push r12"::);
		asm volatile("push r13"::);
		asm volatile("push r14"::);
		asm volatile("push r15"::);
		asm volatile("push r16"::);
		asm volatile("push r17"::);
		asm volatile("push r28"::);
		asm volatile("push r29"::);
		gpRunningTask->pStack = (void*)SP;
		/*	Debug-Code: Test, if the Task-Header has it's Signature. */
			if(gpRunningTask->uSignature != TASKSIGNATURE){
				asm	volatile("end0:	brne end0"::);
			}
		/*	Debug-Code: Test if r1 is zero */
			asm volatile("clr r10"::);
			asm volatile("or r1,r10"::);
			asm	volatile("end:	brne end"::);
		/*	Debug-Code: Test if the Interrupts are disabled */
			asm volatile("in r28,__SREG__"::);
			asm volatile("cli"::);
			asm	volatile("andi r28,0x80"::);
			asm volatile("end2: brne end2"::);
		SP	=	(uint)((TASK*)gpSheduleQueue)->pStack;
		gpRunningTask	=	gpSheduleQueue;
		asm volatile("pop r29"::);
		asm volatile("pop r28"::);
		asm volatile("pop r17"::);
		asm volatile("pop r16"::);
		asm volatile("pop r15"::);
		asm volatile("pop r14"::);
		asm volatile("pop r13"::);
		asm volatile("pop r12"::);
		asm volatile("pop r11"::);
		asm volatile("pop r10"::);
		asm volatile("pop r9"::);
		asm volatile("pop r8"::);
		asm volatile("pop r7"::);
		asm volatile("pop r6"::);
		asm volatile("pop r5"::);
		asm volatile("pop r4"::);
		asm volatile("pop r3"::);
		asm volatile("pop r2"::);
		asm volatile("ret"::);
}
//-----------------------------------------------------------------------------
//QueueTask
//-----------------------------------------------------------------------------
void	QueueTask(QUEUE* pQueue,TASK* pTask){
/* Adds a Task to a Queue */
TASK*	pCurrent = (TASK*)pQueue;
		while(pCurrent->pNextTask != 0){
		/* Search as long as the end of the List isn't reached */	
			if(pCurrent->pNextTask->cPriority > pTask->cPriority){
				/* Insert the task befor the first task with a lower priority*/
				break;
			}
			pCurrent = pCurrent->pNextTask;
			if(pCurrent->pNextTask == 0) break;
		}
/* Add the Task to the List */
	pTask->pNextTask 	= pCurrent->pNextTask;
	pCurrent->pNextTask = pTask;
/*	Test if the leader of the Shedule-list has changed */
	if(gpSheduleQueue != gpRunningTask){
		/* The running task has to be switched */
		SwitchTask();
	}
	return;
}
//-----------------------------------------------------------------------------
//UnqueueTask
//-----------------------------------------------------------------------------
void	UnqueueTask(QUEUE* pQueue,TASK* pTask){
/*	Removes a Task from a Queue */
TASK**	ppCurrentTask	=	&(pQueue->pNextTask);
	while(*ppCurrentTask != 0){
		if(*ppCurrentTask == pTask){
			/*	We got the task in the Queue. Kick it out */
			*ppCurrentTask	= pTask->pNextTask;
			break;
		}
	}
}
//-----------------------------------------------------------------------------
//SheduleTask
//-----------------------------------------------------------------------------
void	SheduleTask(TASK* pTask){
/*	Adds a Task to the Shedule List and makes a Context-Switch if the Task has
	the highest priority now */

	EnterCritical();		//CRITICAL SECTION ENTERED
	QueueTask((QUEUE*)&gpSheduleQueue,pTask);
	LeaveCritical();		//LEAVING CRITICAL SECTION

	return;
}
//-----------------------------------------------------------------------------
//SuspendTask
//-----------------------------------------------------------------------------
void	SuspendTask(void){
/*	Stops the current task from running. It is simply deleted from the
	Shedule-List. The only way to resume it is to call SheduleTask. */

	EnterCritical();		//CRITICAL SECTION ENTERED
	gpSheduleQueue = gpSheduleQueue->pNextTask;
	SwitchTask();
	LeaveCritical();		//LEAVING CRITICAL SECTION

	return;
}
//-----------------------------------------------------------------------------
//TaskChangePriority
//-----------------------------------------------------------------------------
uchar	TaskChangePriority(uchar ucPriority){
/*	Changes the Priority of the running task. If the new priority is less
	than the old, the control maybe turned to another task. Therfor the
	task is first suspended and then resheduled */
uchar	ucOldPriority;

	EnterCritical();		//CRITICAL SECTION ENTERED
//BAD: cPriority, but value is uchar!!!!
	/*	Change the Priority */
	ucOldPriority				= gpSheduleQueue->cPriority;
	gpSheduleQueue->cPriority	= ucPriority;
	/*	Remove the task from the Shedule-List */
	gpSheduleQueue	= gpSheduleQueue->pNextTask;
	/*	Shedule the task */
	QueueTask((QUEUE*)&gpSheduleQueue,(TASK*)gpRunningTask);
	LeaveCritical();		//LEAVING CRITICAL SECTION
	
	return	ucOldPriority;
}


//-----------------------------------------------------------------------------
//WaitEvent
//-----------------------------------------------------------------------------
char	WaitEvent(EVENT* pEvent){
/*	Simply waits for an Event, no timeout */
		return	WaitEventTimeout(pEvent,0);
}
//-----------------------------------------------------------------------------
//WaitEventTimeout
//-----------------------------------------------------------------------------
char	WaitEventTimeout(EVENT* pEvent,long lTimeout){
/*	Checks if pEvent is signaled. If not we wait until either the Event
	gets signaled or a Timeout occures. A Timeout of 0 indicates no 
	Timeout*/
	EnterCritical();			//CRITICAL SECTION ENTERED
/* Let the task wait for an event to occure */
	if((int)pEvent->pNextTask == EVENT_SIGNALED){
		/* The event is signaled. Untrigger it and return */
		pEvent->pNextTask = (TASK*)EVENT_UNSIGNALED;
		LeaveCritical();		//LEAVING CRITICAL SECTION
		return	EVENT_SIGNALED;
	}else{
		/* 	We have to wait for the event: Check Timer-Settings */
		if(lTimeout != 0){
			/*	The timer has to be started */
			StartTimer((TIMER*)\
						((char*)gpRunningTask+sizeof(TASKHEADER)),lTimeout);
		}
		/*Save the Event-Pointer */
		gpRunningTask->pEvent	=	pEvent;
		/*Make the next Task from the Shedule-
			List runable */
		gpSheduleQueue = gpSheduleQueue->pNextTask;
		/*Queue the current Task for the event. A Task Switch is performed
		  afterwards */
		QueueTask((QUEUE*) pEvent,(TASK*)gpRunningTask);
		/*	We're still in the CriticalSection !!! */
			assert((SREG | SREG_I) != 0); 
		LeaveCritical();		//LEAVING CRITICAL SECTION
		/*	The Returncode (EVENT_SIGNALED or EVENT_TIMEOUT) is 
			in the Event-Pointer */
		return	(char)((uint)gpRunningTask->pEvent);
	}
}
//-----------------------------------------------------------------------------
//SignalEvent
//-----------------------------------------------------------------------------
void	SignalEvent(EVENT* pEvent){
/* Signales an event */
TASK*	pTask;

	if(pEvent == 0)	return;
	EnterCritical();		//CRITICAL SECTION ENTERED
	/*	The pointer (is assumed to) point to an event... */
	switch((uint)(pEvent->pNextTask)){
		case EVENT_UNSIGNALED:
			/* There is no task waiting */
			pEvent->pNextTask = (TASK*) EVENT_SIGNALED;
			break;
		case EVENT_SIGNALED:
			/* The event is already signaled */
			break;
		default:
			/* There is someone waiting */
			pTask 				= pEvent->pNextTask;
			pEvent->pNextTask 	= pEvent->pNextTask->pNextTask;
			pTask->pEvent		=(EVENT*)EVENT_SIGNALED;
			/*Stopp the timer (if it's waiting)*/
			StopTimer((TIMER*)((char*)pTask+sizeof(TASKHEADER)));
			/*Shedule the task*/
			SheduleTask(pTask);
	}
	LeaveCritical();		//LEAVING CRITICAL SECTION

	return;
}
//-----------------------------------------------------------------------------
//ClearEvent
//-----------------------------------------------------------------------------
void	ClearEvent(EVENT* pEvent){
/*	If no task is waiting for the event and the event is signaled, the
	call clears the Event */
	EnterCritical();		//CRITICAL SECTION ENTERED
	if((u16)(pEvent->pNextTask) == EVENT_SIGNALED){
		/*	No task waiting, clear the event */
		pEvent->pNextTask = (TASK*) EVENT_UNSIGNALED;
	}
	LeaveCritical();		//LEAVING CRITICAL SECTION
}

//-----------------------------------------------------------------------------
//InitTask
//-----------------------------------------------------------------------------
TASK*	InitTask(void* pMem,size_t uMemSize,void* pEntryFunction,char cPriority){
/*	The given Buffer will be used to setup a TASK-Struckture. The rest of Memory
	will be used as the task's Stack */
/*	The Task-Structure is placed at the Beginning of the Buffer.  Because the
	stack grows in the Inverse Direction, it will be placed at the end !!*/
TASK*	pTask	=	(TASK*)pMem;
		EnterCritical();		//CRITICAL SECTION ENTERED
/*	Priority of the Task: The task with the "lowest number == highest priority"
	that is runable will run! */
		pTask->cPriority 		= cPriority;
/*	This is the Task's Event-Timer. It is used to wait for an event with
	timeout. Because the systems timer-mechanism is based on Callback-Routines,
	this has to be set to the correct function that is used to stop a task from
	waiting for an event any longer */
		pTask->Timer.pCallback	= eventtimer;
		pTask->pEvent			= 0;
/*	This signature is only needed for Debug-Reasons: If too much Stack-Space is
	used, it will be overwritten (most likely), so an invalid Signature is a 
	hint to Stack Overflow! This is checked after a new task is sheduled! */
		pTask->uSignature 		= TASKSIGNATURE;
/*	The initial Stack-Pointer value is placed at the end of the buffer 
	(- sizeof(TASK) because this structure occupies space...), leaving out
	a so-called Switchframe (this stuff is put into the registers when the
	task is sheduled for the first time! */
		pTask->pStack	= (void*)(((char*)pTask)+(uMemSize-SWITCHFRAME_SIZE-sizeof(TASK)));
/*	Put the "Return-Address" at the End of the Switchframe: This is the
	address of the entry-function. The Code-Execution starts there !!! */
		*((char*)pTask+(uMemSize-sizeof(TASK)-1)) = (char)((uint)pEntryFunction);
		*((char*)pTask+(uMemSize-sizeof(TASK)-2)) = (char)((uint)pEntryFunction >> 8);
/*	Shedule the Task - if it's runable, run it! */
		SheduleTask(pTask);
		LeaveCritical();		//LEAVING CRITICAL SECTION
	return	pTask;
} 


//-----------------------------------------------------------------------------
//CreateTask
//-----------------------------------------------------------------------------
TASK*	CreateTask(void* pEntryFunction,char cPriority,size_t uMemSize){
TASK*	pTask;	
		pTask	= (TASK*)_malloc(uMemSize);
		if(pTask == 0){
						return 0;
		}

		EnterCritical();		//CRITICAL SECTION ENTERED
/* We got the space, create the task */
		pTask->cPriority 		= cPriority;
		pTask->Timer.pCallback	= eventtimer;
		pTask->pEvent			= 0;
		pTask->uSignature 		= TASKSIGNATURE;
		pTask->pStack	= (void*)(((char*)pTask)+(uMemSize-SWITCHFRAME_SIZE));
		*((char*)pTask+(uMemSize-1)) = (char)((uint)pEntryFunction);
		*((char*)pTask+(uMemSize-2)) = (char)((uint)pEntryFunction >> 8);
		SheduleTask(pTask);
		LeaveCritical();		//LEAVING CRITICAL SECTION

		return	pTask;
}
//-----------------------------------------------------------------------------
//Timer0 ISR - Used by the system-timer
//-----------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect){
/*	One System-Tick is expired */
TIMER**	ppTimer	= (TIMER**)&gpTimerRunQueue;
	glTime++;
	glTime &=	TIMER_MAXTIMEMASK;
	if(glTime == glAlertTime){
	/* We have an alert */	
		while(glTime == glAlertTime){
		/*	The Timer-Function has to be executed:
			Transfer the timer to the run-list */
			while(*ppTimer != 0){
				ppTimer = &((*ppTimer)->pNextTimer);
			}
			*ppTimer = (TIMER*)gpTimerWaitQueue;
			gpTimerWaitQueue		= gpTimerWaitQueue->pNextTimer;
			if(gpTimerWaitQueue != 0){
			/*	There is another timer, take it's alert-time
				Because the global Time wrapps round, the highest
				bit has to be cleared. */
				glAlertTime	= (gpTimerWaitQueue->ulTime & TIMER_MAXTIMEMASK);
			}else{
			/* There is no other timer, disable the alerts */
				glAlertTime				= TIMER_STOPPED;
			}
			(*ppTimer)->pNextTimer	= 0;
		}
		if(gcTimerStatus == TIMERSTATUS_WAITING){
		/*	The timer is waiting: Shedule the Timer-Task and
			adjust the Status to avoid reentrance! The ISR may
			add new Timers to the Run-List, but it won't start the
			task again! */
			gcTimerStatus = TIMERSTATUS_ALERT;
			SheduleTask((TASK*)gpTimerTask);
		}else{
		/*	The timer-task is already running, nothing to be done */
		}
	}
}
//-----------------------------------------------------------------------------
//StopTimer
//-----------------------------------------------------------------------------
void	StopTimer(TIMER* pTimer){
/*	Searches for the given timer in the waitlist. The runlist can't
	contain the timer, because the timer-task has the highest priority and
	the timer-functions can't be suspended!
	Can't be called within an ISR: They may interrupt the Timer-Task!! */
TIMER**	ppTimer	= (TIMER**)&gpTimerWaitQueue;

	EnterCritical();		//CRITICAL SECTION ENTERED
	/* Search the timer in the wait-queue */
	while(*ppTimer != 0){
		if(*ppTimer ==	pTimer){
			/* We found it, delete it from the list */
			*ppTimer	= pTimer->pNextTimer;
			break;
		}
		ppTimer = &((*ppTimer)->pNextTimer);
	}
	/* 	Adjust the Alert-Time:
		See ISR, the highest Bit has to be cleared */
	if(gpTimerWaitQueue==0){
		glAlertTime	=	TIMER_STOPPED;
	}else{
	/*There is a timer left, take it's alarm-time */
		glAlertTime = (gpTimerWaitQueue->ulTime & TIMER_MAXTIMEMASK);
	}
	LeaveCritical();		//LEAVE THE CRITICAL SECTION
	return;
}
//-----------------------------------------------------------------------------
//StartTimer
//-----------------------------------------------------------------------------
void	StartTimer(TIMER* pTimer,long lTime){
/*	Starts a timer with the given time */
TIMER**	ppTimer	= (TIMER**)&gpTimerWaitQueue;
ulong	ulWaitTime;
/*	The glTime-Variable is a rotating Variable. After the Maximum Time
	is reached, its reseted and starts with zero again. It's restricted
	to signed values and is never negativ, so the highest bit is always
	zero. To ensure that the timestamp of the timer is always bigger
	than the global time, the timestamp is unsigned and may have its
	highest bit set. */

	EnterCritical();		//CRITICAL SECTION ENTERED
	/*	Set the time left to wait */
	ulWaitTime	= (ulong)lTime + (ulong)glTime;
	pTimer->ulTime	= ulWaitTime;
	/*	Insert the Timer into the list, ordering it
		after the time left to wait */
	while(*ppTimer != 0){
		/* REMARK: The Timestamp is always higher than the
		   global time! */
		if(((*ppTimer)->ulTime-((ulong)glTime)) > ((ulong)lTime)) break;
		ppTimer = &((*ppTimer)->pNextTimer);
	}
	pTimer->pNextTimer = *ppTimer;
	*ppTimer = pTimer;
	/* 	Adjust the Alert-Time:
		See ISR, the highest Bit has to be cleared */
	glAlertTime = (gpTimerWaitQueue->ulTime & TIMER_MAXTIMEMASK);
	LeaveCritical();		//LEAVE THE CRITICAL SECTION

	return;
}
//-----------------------------------------------------------------------------
//RestartTimer
//-----------------------------------------------------------------------------
char	RestartTimer(TIMER* pTimer,long lTime){
/*	Restarts a timer, that means it substracts the timer
	already expired from the wait-time. If that's a positive
	number, the function restarts the timer. Otherwise it
	simply returns */
ulong	ulLastTime;
ulong	ulGlobalTime;
long	lTimeExpired;

	EnterCritical();			//CRITICAL SECTION ENTERED
/*	Remark: The function requires, that the timer is restarted before
	the maximum length a timer can have is expired. It takes simply the
	absulute value of the Time Difference between the global time and
	the last timer-time. The */
		ulLastTime		= (pTimer->ulTime & TIMER_MAXTIMEMASK);
		ulGlobalTime	= (ulong)glTime;
		if(ulGlobalTime < ulLastTime){
		/*	The global Time is lower than the last timeout. This 
			means that the global Time has been reset meanwile. Thus
			adjust this by setting the highest bit: The Global Time
			is now bigger again */
			ulGlobalTime |= TIMER_MAXTIMEFLAG;
		}
		lTimeExpired	= (long)(ulGlobalTime-ulLastTime);

		if(lTimeExpired>=lTime){
			/* The time is already expired */
			LeaveCritical();		//LEAVE THE CRITICAL SECTION
			return TIMER_EXPIRED;
		}else{
			/* There is time left to wait */
			StartTimer(pTimer,lTime-lTimeExpired);
			LeaveCritical();		//LEAVE THE CRITICAL SECTION
			return	TIMER_STARTED;
		}
}
//-----------------------------------------------------------------------------
//Timer-Task
//-----------------------------------------------------------------------------
TASKBODY(timertask){
/* This Task is used by the timer to execute the callback-functions */
void	(*pCallback)(TIMER* pTimer);
TIMER*	pTimer;
loop:
		EnableInterrupts();
		while(gpTimerRunQueue != 0){
		/* As long as there is a function to execute, I will do this */
			EnterCritical();		//CRITICAL CODE
			pTimer			= (TIMER*)gpTimerRunQueue;
			pCallback		= gpTimerRunQueue->pCallback;
			gpTimerRunQueue	= gpTimerRunQueue->pNextTimer;
			LeaveCritical();		//END CRITICAL CODE

			pCallback(pTimer);
		}
		/*	Sleep for a while and adjust the Timer-Mode that the Task
			isn't running any more */

	EnterCritical();	//CRITICAL CODE
	gcTimerStatus	= TIMERSTATUS_WAITING;
	SuspendTask();
	LeaveCritical();	//END CRITICAL CODE

		goto	loop;
}
//-----------------------------------------------------------------------------
//Sleep
//-----------------------------------------------------------------------------
void	sleep(u32 u32Timeout){
	EnterCritical();			//CRITICAL SECTION ENTERED
/*	Setup the callback-function for the timer of the task */
	gpSheduleQueue->Timer.pCallback	= sleeptimer;
/*	The timer has to be started */
	StartTimer((TIMER*)\
				((char*)gpRunningTask+sizeof(TASKHEADER)),u32Timeout);
/*	Suspend the task */
	SuspendTask();
/*	TIMER EXPIRED!!!! */

/*	Restore the default callback-function for the timer of the task */
	gpSheduleQueue->Timer.pCallback	= eventtimer;
	LeaveCritical();		//LEAVE THE CRITICAL SECTION
}
//-----------------------------------------------------------------------------
//Sleep Timout-Handler
//-----------------------------------------------------------------------------
void	sleeptimer(TIMER* pTimer){
/* 	This timer is used internally to timeout the sleep-period for a task.
	After the timer is expired, the task should wake up. Thus shedule the
	task again */
TASK*	pTask;
	EnterCritical();			//CRITICAL SECTION ENTERED
	/*	The Timer is at the Beginning of the Task-Structure, after
		the <TASKHEADER> Therfor the Task-Pointer can be constructed
		out of the Timer-Pointer. */
	pTask	= (TASK*)((char*)pTimer-sizeof(TASKHEADER));
	/*	Shedule the Task */
		SheduleTask(pTask);
	/*	REMARK: The Timer-Task has the highest Priority, no task-switch
		can occure here !!! */
	LeaveCritical();		//LEAVE THE CRITICAL SECTION
}
//-----------------------------------------------------------------------------
//Event Timout-Handler
//-----------------------------------------------------------------------------
void	eventtimer(TIMER* pTimer){
/* 	This timer is used internally to timeout an event. Every task has
	a build-in timer which is used by <WaitEvent> if a timeout is given.
	The Task will be taken out of the EventQueue and will be sheduled.
	The Status will be adjusted to indicate, that a timeout has occured */
TASK*	pTask;
	
	EnterCritical();			//CRITICAL SECTION ENTERED
	/*	The Timer is at the Beginning of the Task-Structure, after
		the <TASKHEADER> Therfor the Task-Pointer can be constructed
		out of the Timer-Pointer. */
	pTask	= (TASK*)((char*)pTimer-sizeof(TASKHEADER));
	/*	Remove the Task from the Queue */
		UnqueueTask((QUEUE*)pTask->pEvent,pTask);
	/*	Adjust it's Status to Timeout */
		pTask->pEvent	= (EVENT*)EVENT_TIMEOUT;
	/*	Shedule the Task */
		SheduleTask(pTask);
	/*	REMARK: The Timer-Task has the highest Priority, no task-switch
		can occure here !!! */
	LeaveCritical();		//LEAVE THE CRITICAL SECTION

}
//-----------------------------------------------------------------------------
