/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/*	Include other Files if nececcary */
/*****************************************************************************/
#include	<stdio.h>
#include	"inttypes.h"
#include	"global.h"

#ifndef	__SYSTEM
#define	__SYSTEM
/*****************************************************************************/
/* structures */
/*****************************************************************************/
typedef struct _TASK TASK;
typedef struct _TASKHEADER TASKHEADER;
typedef struct _QUEUE QUEUE;
typedef struct _EVENT EVENT;
typedef struct _TIMER TIMER;
//-----------------------------------------------------------------------------
struct _TIMER{
TIMER*	pNextTimer;
ulong	ulTime;
void*	pCallback;
};
/* The Task-Structure */
struct _TASK{
TASK*	pNextTask;
TIMER	Timer;
EVENT*	pEvent;
uchar	cPriority;
void*	pStack;
char	cErrno;
uint	uSignature;
};
/* Header between the Task and the Timer Structure */
struct _TASKHEADER{
TASK*	pNextTask;
};
struct _QUEUE{
TASK*	pNextTask;
};
struct _EVENT{
TASK*	pNextTask;
};
/*****************************************************************************/
/* Constants */
/*****************************************************************************/
//-----------------------------------------------------------------------------
//Task-related Stuff
//-----------------------------------------------------------------------------
#define	TASKSIGNATURE		0xACDC
#define	TIMERTASK_PRIORITY	0x00
#define	TIMERTASK_MEMSIZE	(250 + sizeof(TASK))
//#define	TIMERTASK_MEMSIZE	(200 + sizeof(TASK))
char	timertask_mem		[TIMERTASK_MEMSIZE];

#define	MAINTASK_PRIORITY	0x40
//#define	MAINTASK_MEMSIZE 	(400 + sizeof(TASK))
#define	MAINTASK_MEMSIZE 	(300 + sizeof(TASK))
char	maintask_mem		[MAINTASK_MEMSIZE];

#define	NETWORKTASK_PRIORITY		0x20
//#define	NETWORKTASK_MEMSIZE 		(350	+ sizeof(TASK))
#define	NETWORKTASK_MEMSIZE 		(300	+ sizeof(TASK))
char	networktask_mem	[NETWORKTASK_MEMSIZE];
volatile	TASK*	gpNetworkTask;


#define	IDLETASK_PRIORITY	0xFF
#define	SWITCHFRAME_SIZE	((18+1)*sizeof(char)+sizeof(void*))
//-----------------------------------------------------------------------------
//Timer-related Stuff
//-----------------------------------------------------------------------------
#define	TIMER_STOPPED -1
#define	TIMER_EXPIRED 1
#define	TIMER_STARTED 0
#define	TIMER_MAXTIMEMASK	0x7FFFFFFF
#define	TIMER_MAXTIMEFLAG	0x80000000
#define	TIMERSTATUS_WAITING	0
#define	TIMERSTATUS_ALERT	1
//-----------------------------------------------------------------------------
//Events and Queues
//-----------------------------------------------------------------------------
#define	EVENT_UNSIGNALED 0	//Has to be zero - the functions expects that!
#define	EVENT_SIGNALED	 1
#define	EVENT_TIMEOUT	 2
/*****************************************************************************/
/* Macros */
/*****************************************************************************/
//The Task-Declaration Macro
#define TASKDECLARE(_entryfunction) void _entryfunction(void)\
 __attribute__ ((noreturn));
#define TASKBODY(_entryfunction) void _entryfunction(void)\
 __attribute__ ((noreturn)); \
 void _entryfunction(void)
//-----------------------------------------------------------------------------
//Timer-Macros
#define	TIMERSETCALLBACK(timer,callback)	(timer)->pCallback = callback;
//-----------------------------------------------------------------------------
//Event-Macros
#define	EVENTINIT(event)			event.pNextTask = (TASK*)EVENT_UNSIGNALED;
#define	EVENTINITSIGNALED(event)	event.pNextTask = (TASK*)EVENT_SIGNALED;
/*****************************************************************************/
/* API */
/*****************************************************************************/
//TIMER
char	WaitEventTimeout(EVENT* pEvent,long lTimeout);
void	StartTimer(TIMER* pTimer,long lTime);
char	RestartTimer(TIMER* pTimer,long lTime);
void	StopTimer(TIMER* pTimer);
//INTERNAL
TASKDECLARE(timertask);
void	eventtimer(TIMER* pTimer);
void	sleeptimer(TIMER* pTimer);
//TASK
uchar	TaskChangePriority(uchar ucPriority);
void	sleep(u32 u32Timeout);
void	SwitchTask(void) __attribute__ ((noinline)) __attribute__ ((naked));
void	QueueTask(QUEUE* pQueue,TASK* pTask);
void	UnqueueTask(QUEUE* pQueue,TASK* pTask);
void	SheduleTask(TASK* pTask);
void	SuspendTask(void);
TASK*	CreateTask(void* pEntryFunction,char cPriority,size_t uMemSize);
TASK*	InitTask(void* pMem,size_t uMemSize,void* pEntryFunction,char cPriority);
//EVENT
char	WaitEvent(EVENT* pEvent);
char	WaitEventTimeout(EVENT* pEvent,long lTimeout);
void	SignalEvent(EVENT* pEvent);
void	ClearEvent(EVENT* pEvent);
//SECURITY
static inline void EnterCritical(void) __attribute__((always_inline));
static inline void LeaveCritical(void) __attribute__((always_inline));
static inline void EnableInterrupts(void) __attribute__((always_inline));
//Errno
void	SetErrno(char cErrno);
char	GetErrno(void);
//memory
inline	void*	_malloc(size_t size);
inline	void	_free(void* data);
//-----------------------------------------------------------------------------
//Macro-Functions for Interrupt-Handling and Secure Operating
//-----------------------------------------------------------------------------
inline	void	EnterCritical(void){
/*	Enters a critical region: Interrupts will be globaly disabled and
	the privious status is saved */
	asm	volatile("in r0,__SREG__"::);
	asm	volatile("push r0"::);
	asm	volatile("cli"::);
}
inline	void	LeaveCritical(void){
/*	Leaves a critical region: The Interrupt-Status will be restored */
	asm	volatile("pop r0"::);
	asm	volatile("out __SREG__,r0"::);
}
inline	void	EnableInterrupts(void){
	SREG	|=	_BV(SREG_I);	//Enable Interrupts Globaly;
}
//-----------------------------------------------------------------------------
//Global System-Variables
//-----------------------------------------------------------------------------
//Timer
volatile	TIMER*	gpTimerWaitQueue;
volatile	TIMER*	gpTimerRunQueue;
volatile	long	glTime;
volatile	long	glAlertTime;
volatile	char	gcTimerStatus;
//Tasks
volatile	TASK*	gpTimerTask;
volatile	TASK*	gpMainTask;
volatile 	TASK	gIdleTask;
volatile	TASK*	gpSheduleQueue;
volatile	TASK*	gpRunningTask;
/*****************************************************************************/
/*	TIMER:
	Callback-Funktion has to be from the typ:
	void	mycallback(TIMER* pTimer)
	*/
/*****************************************************************************/
#endif
