/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/* TCP */
/*****************************************************************************/
#include "tcp.h"
//-----------------------------------------------------------------------------
//SEND
//-----------------------------------------------------------------------------
void	tcp_send(TCPTCB* spTCB){
/*	We should send a TCP-packet */
u8 		u8Flags		= 0;
//u16		u16PacketLength;
//u16		u16TxBufferLength;
u8*		u8pRead;
u8*		u8pCurrentRead;
u8*		u8pWrite;
u8*		u8pStart;
u8*		u8pStop;
u16		u16Free;
u16		u16SendLength;
u16		u16Checksum;
u16		u16WindowSize;
u16		u16Offset;
u16		u16Tmp;
/*****************************************************************************/
	u8pRead			= spTCB->u8pTxBufferRead;
	u8pWrite		= spTCB->u8pTxBufferWrite;
	u8pStart		= spTCB->TxBuffer;
	u8pStop			= spTCB->TxBuffer+TCP_TXBUFFERLENGTH;
	u16Free			= spTCB->u16TxBufferFree;
	u16Offset		= spTCB->u16TxBufferOffset;

/*****************************************************************************/
/*	Set the Timer to zero: The TCP-Timer wont try to modify this TCB!!! */
	EnterCritical();
	spTCB->s8TCPTxTimer			= 0;
	LeaveCritical();	
/*****************************************************************************/
/* Assemble the Package: 1st Overall initialization */
/*****************************************************************************/
/*----- Construct the header ------------------------------------------------*/
ETHERNETHEADER*	spEthernetHeader	= (ETHERNETHEADER*)&gu8NetworkBuffer;
TCPHEADER*		spTCPHeader			= (TCPHEADER*) ((char*)&gu8NetworkBuffer + sizeof(ETHERNETHEADER) + sizeof(IPHEADER));
/*	Send the Message via Ethernet-Boradcast, Protocol: IP (0x0800) */
	ethernet_writeheader(spEthernetHeader,&mymac,&spTCB->sRemoteMac,0x0800);
/*	Write the TCP-Header: */
/*	Set the Source/Destination ports */
	spTCPHeader->sSourcePort	= spTCB->sLocalTcpPort;
	spTCPHeader->sDestPort		= spTCB->sRemoteTcpPort;
/*	Set the acknowledge-number of the frame */
	tcp_setacknumber(spTCPHeader,spTCB->u32LocalReceived);
/*	Set the Data-Offset: We use no options. The upper
	4 bit give the Data-Offset as a multiple of 4 bytes */
	spTCPHeader->u8DataOffset	= (20/4) << 4;
/*	Set the window-size */
	if(TCP_MINWINDOWSIZE > spTCB->u16RxBufferFree){
		/*	We have less than the minimum Window-Size: Advertise a
			zero-window */
		u16WindowSize	= 0;
	}else{
		/*	Advertise all we have */
		u16WindowSize	= spTCB->u16RxBufferFree;
	}
	tcp_setwindowsize(spTCPHeader,u16WindowSize);
/*	Set the urgent-pointer */
	tcp_seturgentpointer(spTCPHeader,0);
/*****************************************************************************/
/* Assemble the Packages: 2nd Individual Data */
/*****************************************************************************/
	while(1){
		/*	We have to calculate the current read position in the Buffer */
		u16Tmp	= u8pRead + u16Offset - u8pStop;
		if((s16)u16Tmp < 0){
			/*	Even with Offset we havent reached the end of the Buffer */
			u8pCurrentRead 	= u8pRead + u16Offset;
		}else{
			/*	With the offset we reached the end of the Buffer. Wrap */
			u8pCurrentRead	= u8pStart + u16Tmp;
		}
		/*	We send a Package from the current read position  up to the
			end if possible */
		u16SendLength		=	(u16)(u8pStop - u8pCurrentRead);
		if((TCP_TXBUFFERLENGTH - u16Free - u16Offset) < u16SendLength){
			/*	We have less stuff in the buffer. We can only
				read up to the write-pointer */
			u16SendLength	=	(u16)(u8pWrite - u8pCurrentRead);
		}
		/*	Check for the maximum Packet Size */
		if(u16SendLength > TCP_MAXPACKETLENGTH){
			u16SendLength = TCP_MAXPACKETLENGTH;
		}
		/*	Check for the Remote Window: If its smaller than the current
			Send-Length, we have to send less data */
		if(u16SendLength > spTCB->u16RemoteWindowSize){
			u16SendLength	= spTCB->u16RemoteWindowSize;
			if(spTCB->u16RemoteWindowSize == 0){
				/*	We should send zero Data. This means the
					other side is advertising a zero window.
					Lets make a zero window-test next time
					and dont send anything now */
				spTCB->u16RemoteWindowSize	= 1;
				break;
			}
		}
		/*	We dont want to transmitt more data then buffer space
			on the remote side. Thus decrease the Remote Window */
			//DANGER: In case of Retransmit...!!!
			spTCB->u16RemoteWindowSize	-=	u16SendLength;
	/*************************************************************************/
	/*	TCP-Finit-State Machine */
	/*************************************************************************/
		switch(spTCB->u8TCPState){
	/*----- TCP_SYNSENT --------------------------------------------------------*/
			case TCP_SYNSENT:
			/*	We send a SYN and wait for an SYN&ACK. But maybe we timed out
			 * 	and have to try it again */
	/*----- TCP_OPEN --------------------------------------------------------*/
			case TCP_OPEN:
			/*	We should establish a new connection:
				send a syn over the ether */
				u8Flags				= TCP_SYN;
			/*	We don't send any data yet */
				u16SendLength		= 0;
			/*	Start the Retransmission-Timer */
				EnterCritical();
				spTCB->s8TCPTxTimer	= TCP_SYNSENDTIMEOUT;
				LeaveCritical();
			/*	Change state */
				spTCB->u8TCPState	= TCP_SYNSENT;
				break;
	/*----- TCP_SYNSENT1 --------------------------------------------------------*/
			case TCP_SYNSENT1:
			/*	We received a SYN/ACK, now we have to send an ACK back 
			 * 	After that we are in the ESTABLISHED-State! */
					spTCB->u8TCPState	= TCP_ESTABLISHED;
				/*	Signal the TX-Event: The connection is open!!! */
					SignalEvent(&spTCB->sTCPTxEvent);
				/*	Send the ACK */
					u8Flags				= TCP_ACK;
				/*	We don't send any data yet */
					u16SendLength		= 0;
					break;
	/*----- TCP_ESTABLISHED -----------------------------------------------------*/
			case	TCP_ESTABLISHED:
			/*	The connection is established: simply send data */
				u8Flags				= TCP_ACK;
				break;

	/*----- TCP_RESET0 --------------------------------------------------------*/
			case TCP_RESET0:
			/*	We should reset the connection */
				spTCB->u8TCPState	= TCP_RESET;
	/*----- TCP_RESET --------------------------------------------------------*/
			case TCP_RESET:
			/*	We received an RST-Flag and should transmitt s.th.
				our we detected an abnormal condition and want to
				reset the connection. Lets do that */
				u8Flags				= TCP_RST;
			/*	Don't send any data! */
				u16SendLength		= 0;
				break;							
	/*----- TCP_SYNRECEIVED --------------------------------------------------------*/
			case TCP_SYNRECEIVED:
			/*	We Received a SYN and send back an ACK/SYN. Now the 
				retransmission-timer timed out and we should send it
				again */
				u8Flags				= 0;
	/*----- LISTEN --------------------------------------------------------------*/
			case TCP_LISTEN:
			/*	Step 2 of the 3-way-handshake: send ack and syn */
				spTCB->u8TCPState	= TCP_SYNRECEIVED;
				u8Flags				= (TCP_ACK | TCP_SYN);
			/*	We don't send any data yet */
				u16SendLength		= 0;
				break;
	/*----- TCP_FINWAIT2 --------------------------------------------------------*/
			case	TCP_FINWAIT2:
			/*	We send a FIN and received an ACK. Now we wait for an FIN
				of the other site. But we should retransmitt.
				DANGER: Because we already accepted that the remote
				end has acknowledged our FIN, our Sequence-Number is
				increased. To enter this state its neccessary that all our
				data are received and that the remote side acknowledged all of
				them. Thus we cant have any data to transmitt any more.!!! */
				if(TCP_TXBUFFERLENGTH != u16Free ){
					/*	We have an ERROR */
					while(1);
				}

				/*	Simply retransmitt the ACK (With FIN? DANGER!!!) */
				u8Flags				= TCP_ACK;
				break;
	/*----- TCP_FINWAIT1 --------------------------------------------------------*/
			case	TCP_FINWAIT1:
			/*	We already send our last package. But now we have to retransmit!!
				This means we will send the data again like in the ESTABLISHED-
				state, we only have to add the FIN to the last Segment!!! 
				Thus we work like FINWAIT 0!!! */
	/*----- TCP_FINWAIT0 --------------------------------------------------------*/
			case	TCP_FINWAIT0:
			/*	The connection should be closed as soon as all data
				are send. Thus this state acts as ESTABLISHED as long
				as we have data */
				/*	Check if this is our last package */
				if(u16SendLength == (TCP_TXBUFFERLENGTH - u16Free - u16Offset)){
					/*	We have no more data after this!!!
						We will send a FIN-Flag */
					u8Flags				= TCP_FIN | TCP_ACK | TCP_PSH;
					/*	Change the State: We send our last package */
					spTCB->u8TCPState	= TCP_FINWAIT1;
					/*	We send a FIN. We will start a Retransmission-Timer, 
						because we wait for an ACK!!! */
					EnterCritical();
					spTCB->s8TCPTxTimer			= TCP_FINACKNOWLEDGETIMEOUT;
					LeaveCritical();
				}else{
					/*	Like Established */
					u8Flags				= TCP_ACK;
				}
				break;
	/*----- TCP_TIMEDWAIT --------------------------------------------------------*/
			case	TCP_TIMEDWAIT:
			/*	We should acknowledge the FIN of the other side */
					u8Flags				= TCP_ACK;
				/*	There should be no data left in the buffer */
				if(TCP_TXBUFFERLENGTH != u16Free ){
					while(1);
				}
			/*	We change state to CLOSE 
				DANGER DANGER DANGER!!!! */
					spTCB->u8TCPState		= TCP_CLOSED;
				break;
	/*----- TCP_CLOSEWAIT -------------------------------------------------------*/
			case	TCP_CLOSEWAIT:
			/*	Like ESTABLISHED, exepct that we already received
				a FIN (If this is the first time we send s.th. back
				after we received the FIN, this will also acknowledge
				the FIN of the other site!!! */
				u8Flags		= TCP_ACK;
				break;
	/*----- TCP_LASTACK --------------------------------------------------------*/
			case	TCP_LASTACK:
			/*	We already send our last package. But now we have to retransmit!!
				This means we will send the data again like in the TCP_CLOSEWAIT0-
				state */
	/*----- TCP_CLOSEWAIT0 -------------------------------------------------------*/
			case	TCP_CLOSEWAIT0:
			/*	We received a FIN from the other site (we were in the
				TCP_CLOSEWAIT-State) and now our user wants us to
				close the connection as well. But first we have to
				transmitt all the data we have!!! */
				/*	Check if this is our last package */
				if(u16SendLength == (TCP_TXBUFFERLENGTH - u16Free - u16Offset)){
					/*	We have no more data after this!!!
						We will send a FIN-Flag */
					u8Flags					= TCP_FIN | TCP_ACK | TCP_PSH;
					/*	Change the State: We send our last package */
					spTCB->u8TCPState		= TCP_LASTACK;
					/*	We send a FIN. We will start a Retransmission-Timer, 
						because we wait for an ACK!!! */
					EnterCritical();
					spTCB->s8TCPTxTimer		= TCP_FINACKNOWLEDGETIMEOUT;
					LeaveCritical();
				}else{
					/*	Like Established */
					u8Flags				= TCP_ACK;
				}
				break;
	/*----- DEFAULT -------------------------------------------------------------*/
			default:
			/*	In all other cases don't send anything */
					while(1);
		}
/* Check if we should set the PSH-Flag */
		if((spTCB->u8TCPFlags	& TCPFLAG_TXPSH) != 0){ 
			u8Flags	|= TCP_PSH;
		}
		
		/*	Set the Flags */
		spTCPHeader->u8Flags		= u8Flags;
	/*************************************************************************/
	/*	Fill in remaining Header-Fields  */
		/*	Write the IP-Header: 
			Protokoll:		6 (TCP)
			Messagesize:	Header (IP,TCP),TCP-Message */	
		IPHEADER*		spIPHeader		= (IPHEADER*)((u8*)&gu8NetworkBuffer + sizeof(ETHERNETHEADER));
			ip_writeheader(spIPHeader,6,\
				u16SendLength + sizeof(TCPHEADER),\
				&myip,&spTCB->sRemoteIp,0);
		/*	Set the sequence-number of the current frame */
		tcp_setseqnumber(spTCPHeader,spTCB->u32RemoteReceived + u16Offset);
		/*----- Calculate the Checksum ----------------------------------------------*/
		/*	Set the checksum to zero for the checksum-calculation */
		spTCPHeader->u8ChecksumHi	= 0;
		spTCPHeader->u8ChecksumLo	= 0;
		/*	Add the Pseudo-Header Checksum */
		u16Checksum	= ip_checksum(0,&myip,sizeof(IPADDRESS));							//Source IP
		u16Checksum	= ip_checksum(u16Checksum,&spTCB->sRemoteIp,sizeof(IPADDRESS));		//Dest IP
		u16Checksum	= ip_checksum16bit(u16Checksum,6);									//protocol-number
		u16Checksum	= ip_checksum16bit(u16Checksum,u16SendLength + sizeof(TCPHEADER));	//TCP-Length: Header+Data
		/*	Add the Data */
		u16Checksum	= ip_checksum(u16Checksum,u8pCurrentRead,u16SendLength);
		/*	Add the TCP-Header */
		u16Checksum	= ~ip_checksum(u16Checksum,spTCPHeader,sizeof(TCPHEADER));			//Dest IP
		/*	Write the calculated checksum into the header */
		spTCPHeader->u8ChecksumHi	= HI(u16Checksum);
		spTCPHeader->u8ChecksumLo	= LO(u16Checksum);
	/*****************************************************************************/
	/* Send the packet */
	/*****************************************************************************/
	/*	Wait for the CS8900 to send the Data */
		ethernet_waittxready(sizeof(ETHERNETHEADER)+sizeof(IPHEADER)+sizeof(TCPHEADER)+u16SendLength);
	/*	Send the Header over the ether */
		ethernet_senddata(spEthernetHeader,sizeof(ETHERNETHEADER)+sizeof(IPHEADER)+sizeof(TCPHEADER));
	/*	Send the Data over the ether */
		ethernet_senddata(u8pCurrentRead,u16SendLength);
	/*	Wait until the job is done */
		ethernet_waittxdone();
	/*****************************************************************************/
	/*	We send some Data, the Offset has changed!! */
		u16Offset	+=	u16SendLength;
	/*	Check if we send everything we could */
		if((TCP_TXBUFFERLENGTH - u16Free - u16Offset) == 0){
			/*	We have nothing more to transmit */
			break;
		}else if(u16SendLength == 0){
			/*	The last packet contained no data:
				we stop transmitting stuff */
			break;
		}
		if(u16Free + u16Offset > TCP_TXBUFFERLENGTH){
			while(1);
		}

	}//SEND-Loop (while)
/*****************************************************************************/
/* Update the TCB */
/*****************************************************************************/
	spTCB->u16TxBufferOffset	= u16Offset;
	/*	(Re)Start the TCP-Timer: Retransmission!!! */
	if(spTCB->u16TxBufferFree != TCP_TXBUFFERLENGTH){
		/*	We have stuff in the Buffer. We have to
			start a Retransmission Timeout */
		EnterCritical();
		spTCB->s8TCPTxTimer			= TCP_RETRANSMITTIMEOUT;
		LeaveCritical();
	}else{
		/*	We have no stuff in the Buffer - ensure that the
			Transmitter-Task knows that!!! */
		SignalEvent(&spTCB->sTCPTxEvent);
		/*	We stop retransmission for a while if the 
		 *  ReceiveBuffer is not empty - this means we have nothing to send
		 *  but still data to process, e.g. we may shut up for a while (-: */
		if(spTCB->u16RxBufferFree != TCP_RXBUFFERLENGTH){
			EnterCritical();
			spTCB->s8TCPTxTimer			= 0;
			LeaveCritical();
		}else{
			EnterCritical();
			spTCB->s8TCPTxTimer			= TCP_RETRANSMITTIMEOUT;
			LeaveCritical();
		}
	}
	EnterCritical();
	if(spTCB->u8TCPState == TCP_SYNSENT){
		/*	In case of establishing a connection we want a different
		 * (shorter) timeout */
		spTCB->s8TCPTxTimer			= TCP_SYNSENDTIMEOUT;
	}else{
		spTCB->s8TCPTxTimer			= TCP_RETRANSMITTIMEOUT;		
	}
	LeaveCritical();
/*****************************************************************************/
	if(spTCB->u8TCPState == TCP_TIMEDWAIT){
	/*	The connection is closed!!! */
	/*	Signal Event */
		SignalEvent(&spTCB->sTCPTxEvent);
	}else if(spTCB->u8TCPState == TCP_RESET){
	/*	Stop the Retransmission-Timer */
		EnterCritical();
		spTCB->s8TCPTxTimer			= 0;
		LeaveCritical();
	/*	Signal Event */
		//SignalEvent(&spTCB->sTCPTxEvent);
	}
/*****************************************************************************/
/* DONE */
/*****************************************************************************/
}
/*****************************************************************************/
