/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/* TCP: Small helper functions */
/*****************************************************************************/
#include "tcp.h"
/*****************************************************************************/
/*	Functions to get Infos from the TCB */
inline	u16	tcp_rxbytefree(TCPTCB* spTCB){
/*	Get the number of free bytes in the rx-Buffer */
	return	spTCB->u16RxBufferFree;
}
inline	u16	tcp_txbytefree(TCPTCB* spTCB){
/*	Get the number of free bytes in the tx-Buffer */
	return	spTCB->u16TxBufferFree;
}
inline	u16	tcp_rxbytefull(TCPTCB* spTCB){
/*	Get the number of  bytes currently in the rx-Buffer */
	return	TCP_RXBUFFERLENGTH - spTCB->u16RxBufferFree;
}
inline	u16	tcp_txbytefull(TCPTCB* spTCB){
/*	Get the number of  bytes currently in the tx-Buffer */
	return	TCP_TXBUFFERLENGTH - spTCB->u16TxBufferFree;
}
inline u8*	tcp_txbufferstart(TCPTCB* spTCB){
/*	Returns a pointer to the first address IN the buffer */
	return	spTCB->TxBuffer;
}
inline u8*	tcp_txbufferend(TCPTCB* spTCB){
/*	Returns a pointer to the address one AFTER the last byte in the buffer */
	return	spTCB->TxBuffer + TCP_TXBUFFERLENGTH;
}
inline u8*	tcp_rxbufferstart(TCPTCB* spTCB){
/*	Returns a pointer to the first address IN the buffer */
	return	spTCB->RxBuffer;
}
inline u8*	tcp_rxbufferend(TCPTCB* spTCB){
/*	Returns a pointer to the address one AFTER the last byte in the buffer */
	return	spTCB->RxBuffer + TCP_RXBUFFERLENGTH;
}
inline u8*	tcp_txbufferreadpos(TCPTCB* spTCB){
/*	Returns a pointer to the address we are reading from */
	return	spTCB->u8pTxBufferRead;
}
inline u8*	tcp_txbufferwritepos(TCPTCB* spTCB){
/*	Returns a pointer to the address we are reading from */
	return	spTCB->u8pTxBufferWrite;
}
inline u8*	tcp_rxbufferreadpos(TCPTCB* spTCB){
/*	Returns a pointer to the address we are reading from */
	return	spTCB->u8pRxBufferRead;
}
inline u8*	tcp_rxbufferwritepos(TCPTCB* spTCB){
/*	Returns a pointer to the address we are reading from */
	return	spTCB->u8pRxBufferWrite;
}
/*****************************************************************************/
/* Functions to work with the TCP-Header */
inline void	tcp_setacknumber(TCPHEADER* spHeader,u32 u32AckNumber){
/*	Write the ACK-Number to the tcp-header */
	spHeader->u8AckNumber0	=	(u8)(u32AckNumber>>0);
	spHeader->u8AckNumber1	=	(u8)(u32AckNumber>>8);
	spHeader->u8AckNumber2	=	(u8)(u32AckNumber>>16);
	spHeader->u8AckNumber3	=	(u8)(u32AckNumber>>24);
}
inline void	tcp_setseqnumber(TCPHEADER* spHeader,u32 u32SeqNumber){
/*	Write the SEQ-Number to the tcp-header */
	spHeader->u8SeqNumber0	=	(u8)(u32SeqNumber>>0);
	spHeader->u8SeqNumber1	=	(u8)(u32SeqNumber>>8);
	spHeader->u8SeqNumber2	=	(u8)(u32SeqNumber>>16);
	spHeader->u8SeqNumber3	=	(u8)(u32SeqNumber>>24);
}
inline void	tcp_setwindowsize(TCPHEADER* spHeader,u16 u16WindowSize){
/*	Write the WindowSize to the tcp-header*/
	spHeader->u8WindowHi	=	(u8)(u16WindowSize>>8);
	spHeader->u8WindowLo	=	(u8)(u16WindowSize>>0);
}
inline void	tcp_seturgentpointer(TCPHEADER* spHeader,u16 u16UrgentPointer){
/*	Write the WindowSize to the tcp-header*/
	spHeader->u8UrgentHi	=	(u8)(u16UrgentPointer>>8);
	spHeader->u8UrgentLo	=	(u8)(u16UrgentPointer>>0);
}
inline uint	tcp_windowsize(TCPHEADER* spHeader){
uint	uiRet;
	uiRet	= (((u16)spHeader->u8WindowHi) << 8) + (((u16)spHeader->u8WindowLo) << 0);
	return	uiRet;
}
inline ulong	tcp_acknumber(TCPHEADER* spHeader){
/*	Gets the Acknowledge-Number out of the TCP-Header */
u32	u32Ret;
	u32Ret	= ((u32)spHeader->u8AckNumber0)	\
				+ (((u32)spHeader->u8AckNumber1) << 8)	\
				+ (((u32)spHeader->u8AckNumber2) << 16)	\
				+ (((u32)spHeader->u8AckNumber3) << 24);
	return	u32Ret;
}
inline ulong	tcp_seqnumber(TCPHEADER* spHeader){
/*	Gets the Sequence-Number out of the TCP-Header */
u32	u32Ret;
	u32Ret	= ((u32)spHeader->u8SeqNumber0)	\
				+ (((u32)spHeader->u8SeqNumber1) << 8)	\
				+ (((u32)spHeader->u8SeqNumber2) << 16)	\
				+ (((u32)spHeader->u8SeqNumber3) << 24);
	return	u32Ret;
}
/*****************************************************************************/
/*	Functions to manage the TCBs */
/*****************************************************************************/
void	tcp_queuetcb(TCPTCB* spTCB){
/*	Queue the TCB on the queue */
	EnterCritical();	//CRITICAL
	spTCB->spNext	= gspTcpQueue;
	gspTcpQueue		= spTCB;
	LeaveCritical();	//CRITICAL
}
//-----------------------------------------------------------------------------
void	tcp_unqueuetcb(TCPTCB* spTCB){
/*	Remove the TCB from the global queue */
	if(spTCB->spTxNext	!= (TCPTCB*)1){
		/*	Check if the TCB is removed from the
			TX-Queue */
		while(1);	//DEBUG
	}
/*	Search the queue for the item */
TCPTCB** sppTCBItem		= &gspTcpQueue;

EnterCritical();	//CRITICAL
	while(*sppTCBItem != spTCB){
		if(*sppTCBItem == 0){
			/*	We reached the end of the
				queue. Thus the TCB isnt
				queued any more!!! */
			goto	end;
		}	
		sppTCBItem	=	&((*sppTCBItem)->spNext);
	}
		*sppTCBItem	= spTCB->spNext;
end:
LeaveCritical();	//CRITICAL
}
//-----------------------------------------------------------------------------
char	tcp_connectioncnt(TCPPORT* spPort){
/*	Returns the number of connections which have the given local port */
TCPTCB* spTCB	= gspTcpQueue;
char	cCnt	= 0;
	EnterCritical();	//CRITICAL
	while(spTCB != 0){
		if(TCPPORTCOMPARE(spTCB->sLocalTcpPort,(*spPort))){
			/* This entry has the given local port */
			cCnt++;
		}
		spTCB	= spTCB->spNext;
	}
	LeaveCritical();	//CRITICAL
	return cCnt;
}
//-----------------------------------------------------------------------------
void	tcp_queuetxtcb(TCPTCB* spTCB){
/*	Queue the TCB on the tx-queue if not done yet */
	EnterCritical();	//CRITICAL
	if(spTCB->spTxNext == (TCPTCB*)1){
		/*	The TCB isn't queued on the tx-list yet */
		spTCB->spTxNext	= gspTcpSendQueue;
		gspTcpSendQueue	= spTCB;
	}
/*	Ensure that the Network-Task sends data */
	gu8NetworkSend	= 1;
	LeaveCritical();	//CRITICAL
}
//-----------------------------------------------------------------------------
u16		tcp_getlocalport(void){
/*	Retrieves a usable local TCP-Port */
	gu16TcpDefaultLocalPort++;
	if(gu16TcpDefaultLocalPort == TCP_LASTLOCALPORT){
			/*	We have to wrap the Portnumber */
			gu16TcpDefaultLocalPort	= TCP_FIRSTLOCALPORT;
	}	
	return	gu16TcpDefaultLocalPort;
}
/*****************************************************************************/
