/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
/* TCP */
/*****************************************************************************/
#include "tcp.h"
			/*	DANGER: We have to ensure that if in the Established-
				State we get an SYN&ACK, we repeat the Acknowledge!!! */
/*	Have to setup a SYN-Send Timeout!!! */
/*	DANGER: Open: Fin-States, etc. not considered */

/*****************************************************************************/
/*	Init the TCP */
/*****************************************************************************/
void	tcp_init(void){
/*	Setup TCP */
	/* TCBs */
		gspTcpQueue			= 0;
		gspTcpSendQueue		= 0;
		gspTcpListenList	= 0;
	/*	Setup the Timer */
		TIMERSETCALLBACK(&gstcp_timer,tcp_timer_callback);
		StartTimer(&gstcp_timer,100);
	/* ARP */
		/*	Setup the Callback */
		gsTCPArpQuery.pCallback	= &tcp_arpcallback;
		/*	Init the TCP-Arp-Event */
		EVENTINIT(gsTCPArpEvent);
	/*	Default Port */
		gu16TcpDefaultLocalPort	= TCP_FIRSTLOCALPORT;
}
/*****************************************************************************/
/*	TCP:	Timer */
/*****************************************************************************/
void	tcp_timer_callback(TIMER* pTimer){
TCPTCB*		spTCB;
s8			s8Time;
	spTCB	= gspTcpQueue;	//Start at the beginning
	while(spTCB != 0){
		s8Time	= spTCB->s8TCPTxTimer - 1;	
		if(s8Time < 0){
			/* The Timeout-value was zero, no timeout needed */
		}else{
			/*	Write back the new value */
			spTCB->s8TCPTxTimer	= s8Time;
			if(s8Time == 0){
				/* The timer timed out, add the node to the TX-Queue */
				/*  We have to retransmit: Thus we set the Send-Offset
					back to zero!!! */
				spTCB->u16TxBufferOffset	= 0;
				tcp_queuetxtcb(spTCB);
				/*	We have a retransmission: Increment the
					counter
					This counter gets reseted as soon as the remote
					end acknowledges Data or as soon as the romte
					side sends us valid data!!! */
				if(spTCB->u8TCPRetransmissionCounter++ == TCP_MAXRETRANSMISSIONS){
					/*	The maximum Number of Retransmissions has been
						used up. We reset the connection */
					/*	Reset the connection */
					spTCB->u8TCPState	=	TCP_RESET0;
				}
				/*	Signal the NetworkTask: We have something to transmit */
				SignalEvent(&NetworkEVENT);
			}
		}
		/*	Get next node */
		spTCB	= spTCB->spNext;
	}
	/*	Restart the timer again. NOTE: We dont need exact timing, thus
		we simply start the timer, instead of restarting it. */
	StartTimer(pTimer,TCP_TICK);
}
/*****************************************************************************/
/*	TCP-Arp Handler */
void	tcp_arpcallback(ARPQUERYNODE* pNode,u8 u8Code){
/*	The TCP-Arp Query has succeded or timed out */
	switch(u8Code){
/*---------------------------------------------------------------------------*/	
/*	We got the MAC */
		case	ARP_QUERYSUCCESS:
			gsTCPArpStatus	= TCPARPSTATUS_SUCCESS;
			break;
/*---------------------------------------------------------------------------*/	
/*	We haven't got the MAC */
		case	ARP_QUERYERROR:
			/*	The Status defaults to ERROR!!! */
			break;
	}
/*	Signal the Event */
		SignalEvent(&gsTCPArpEvent);
}
/*****************************************************************************/


/*****************************************************************************/
/*	TCP USER API */
/*****************************************************************************/
//-----------------------------------------------------------------------------
//OPEN
//-----------------------------------------------------------------------------
TCPTCB*	tcp_open(IPADDRESS* spRemoteIP,TCPPORT* spRemotePort){
/*	Try to open a connection */
TCPTCB*	spTCB;
/*	Allocate a TCB */
		spTCB	= (TCPTCB*)_malloc(sizeof(TCPTCB));
		if(spTCB == 0){
		/*	No space for a TCB ! */
			return 0;
		}
		//------	INITIALIZE NEW TCB
		EnterCritical();
			/*	We have to make a ARP-Querry */
		gspTcpArpTCB		= spTCB;
			/*	The ARP-Query IP has to be filled in */
		gsTCPArpQuery.sIP	= *spRemoteIP;
		gsTCPArpStatus		= TCPARPSTATUS_ERROR;
		arp_sendquery(&gsTCPArpQuery);
		LeaveCritical();
		/*	Wait for ARP-Result */
		WaitEvent(&gsTCPArpEvent);
		if(gsTCPArpStatus == TCPARPSTATUS_ERROR){
			return	0;
		}
		/*	We have the corresponding MAC */
		EnterCritical();
			/*	The timer-task has a higher priority, we have to
				keep at least the variables it access exclusive */
		spTCB->spNext					=	0;			//Last one in the chain
		spTCB->spTxNext					=	(TCPTCB*)1;	//Nothing to transmit so far
		spTCB->spListenNext				=	0;			//No more task on the listen-list!!!
		/*	We have to initialize the buffer */
		spTCB->u8pRxBufferRead			= spTCB->RxBuffer;;
		spTCB->u8pRxBufferWrite			= spTCB->RxBuffer;;
		spTCB->u16RxBufferFree			= TCP_RXBUFFERLENGTH;
		spTCB->u8pTxBufferRead			= spTCB->TxBuffer;
		spTCB->u8pTxBufferWrite			= spTCB->TxBuffer;
		spTCB->u16TxBufferFree			= TCP_TXBUFFERLENGTH;
		/*	We havent send anything yet, the offset is zero */
		spTCB->u16TxBufferOffset		=	0;
		/*	We have to initialize the TCP-Events */
		EVENTINIT(spTCB->sTCPTxEvent);
		EVENTINIT(spTCB->sTCPRxEvent);
		/*	Init the Timer */
		spTCB->s8TCPTxTimer					= 0;
		/*	No Flags set so far!! */
		spTCB->u8TCPFlags					=	0;
		/* 	Save the connection-data: IP,MAC,Port */
		spTCB->sRemoteMac					=	gsTCPArpQuery.sMAC;
		spTCB->sRemoteIp					=	*spRemoteIP;
		spTCB->sRemoteTcpPort				=	*spRemotePort;
		/*	This is our local Initial Sequence-number */
		spTCB->u32RemoteReceived				=	0;
		/*	Save the remote window size */
		spTCB->u16RemoteWindowSize			=	0;
		/*	remote initial sequence number */
		spTCB->u32LocalReceived				=	0;
		/*	Add the TCB to the global TCB-Queue */
		tcp_queuetcb(spTCB);
	/*	Try to connect */
		while(1){
			/*	We should send a SYN to establish a connection */
			spTCB->u8TCPState					= TCP_OPEN;
			/*	We have no Retransmission so far */
			spTCB->u8TCPRetransmissionCounter	= 0;
			/*	Get a free local port: If the connection was not successfull, we
			 * 	try a different port */
			SETTCPPORT(spTCB->sLocalTcpPort,tcp_getlocalport());
			/*	Add the TCB to the send-queue: we have to send an SYN/ACK back */
			tcp_queuetxtcb(spTCB);
			/*	Ensure that we send our SYN */
			SignalEvent(&NetworkEVENT);
			/*	Wait for a TX-Event */
			while(1){
				/*	Wait until we are either established or reseted */
				WaitEvent(&spTCB->sTCPTxEvent);
				if(spTCB->u8TCPState	== TCP_ESTABLISHED){
					/*	The connection looks established */
					goto	end;
				}else if(spTCB->u8TCPState	== TCP_RESET){
					break;
				}
			}
		}
end:
		LeaveCritical();
/*	Return the TCB*/
	return	spTCB;
}


//-----------------------------------------------------------------------------
//ACCEPT
//-----------------------------------------------------------------------------
TCPTCB*	tcp_accept(TCPLISTEN* spTCPListen){
/*	Accept an incomming connection. If there is no connection, the
	function will block and wait for it ! */
TCPTCB*	spTCB;

	EnterCritical();	//CRITICAL
	while(spTCPListen->spNextTCB == 0){
		/*	Wait for a connection */
		WaitEvent(&spTCPListen->sListenEvent);
	}
/*	We have a TCB. Take the first one from the list */
	spTCB					= spTCPListen->spNextTCB;
	spTCPListen->spNextTCB	= spTCB->spListenNext;
	assert(spTCB != 0);
	LeaveCritical();	//CRITICAL
/*	Return the TCB*/
	return	spTCB;
}
//-----------------------------------------------------------------------------
//LISTEN
//-----------------------------------------------------------------------------
char	tcp_listen(TCPLISTEN* spTCPListen, char cBacklog){
/*	The calling server-process wants to listen on a port / the given struct
	has to be initialised ! */
s8	s8return	= 0;
TCPLISTEN**	sppListenNext	= &gspTcpListenList;
	EnterCritical();
/*	Walk through the list and check, if the port is already in use */
	while(*sppListenNext != 0){
		/*	compare the Local Port */
		if(TCPPORTCOMPARE((*sppListenNext)->sTcpPort,spTCPListen->sTcpPort)){
			/*	We got a match -> can't listen on this port */
			s8return	= -1;
			goto		ret;
		}
			/*	Get the next node */
		sppListenNext	= &((*sppListenNext)->spNext);
	}

/*	we have no match, insert the new node into the list */
	*sppListenNext				= spTCPListen;
	spTCPListen->spNext			= 0;			//No more entries!!!
	spTCPListen->spNextTCB		= 0;			//No connection yet!!!
	spTCPListen->u8MaxTCB		= cBacklog;		//Number of maximum incomming connections
	spTCPListen->sListenEvent.pNextTask = (TASK*)EVENT_UNSIGNALED;
/*	Success */
ret:
	LeaveCritical();
	return	s8return;
}
//-----------------------------------------------------------------------------
//STOP LISTEN
//-----------------------------------------------------------------------------
char	tcp_stoplisten(TCPLISTEN* spTCPListen){
/*	The caller wants to stop listening on a given port */
	EnterCritical();
TCPLISTEN**	sppListenNext	= &gspTcpListenList;
TCPTCB*		spTCB			= spTCPListen->spNextTCB;
TCPTCB*		spTCBFree;
/*	Walk through the list and find the entry */
	while(*sppListenNext != 0){
		/*	compare the Local Port */
		if(*sppListenNext == spTCPListen){
			/*	we found the entry, remove it */
			*sppListenNext	= spTCPListen->spNext;
			break;
		}
		/*	Get the next node */
		sppListenNext	= &(spTCPListen->spNext);
	}
/*	Now we have to release all TCBs that may be on the list */
	while(spTCB != 0){
		spTCBFree	= spTCB;
		spTCB		= spTCB->spNext;
		_free(spTCBFree);
	}
/*	job done */
	LeaveCritical();
	return 0;
}
//-----------------------------------------------------------------------------
//CLOSE: Reset
//-----------------------------------------------------------------------------
void	tcp_reset(TCPTCB* spTCB){
/*	Reset The Connection */
	EnterCritical();
	spTCB->u8TCPState	= TCP_RESET0;
	/*	Ensure that our RST-Message will be send */
	tcp_queuetxtcb(spTCB);
	gu8NetworkSend	= 1;
	SignalEvent(&NetworkEVENT);
	/*	Wait for the TX-Event */
		while((spTCB->u8TCPState != TCP_CLOSED) && (spTCB->u8TCPState != TCP_RESET)){
			WaitEvent(&spTCB->sTCPTxEvent);
		}
	LeaveCritical();
}

//-----------------------------------------------------------------------------
//CLOSE: RX
//-----------------------------------------------------------------------------
void	tcp_closerx(TCPTCB* spTCB){
/*	Wait until the Remote end closes as well  */
	EnterCritical();
	/*	Wait for the TX-Event */
		while((spTCB->u8TCPState != TCP_CLOSED) && (spTCB->u8TCPState != TCP_RESET)){
			if(WaitEventTimeout(&spTCB->sTCPTxEvent,20) == EVENT_TIMEOUT){
				tcp_reset(spTCB);
			}
		}
	/*	Remove the TCB from the queue */
		tcp_unqueuetcb(spTCB);
	/*	Release the Memory */
		_free(spTCB);
	LeaveCritical();
}
//-----------------------------------------------------------------------------
//CLOSE: TX
//-----------------------------------------------------------------------------
void	tcp_closetx(TCPTCB* spTCB){
/*	Close the TCP-Connection in one direction: We dont want to
	send any more */
	EnterCritical();
/*****************************************************************************/
/* TCP-Finit-State-Machine */
/*****************************************************************************/
	switch(spTCB->u8TCPState){
/*----- TCP_ESTABLISHED -----------------------------------------------------*/
		case	TCP_ESTABLISHED:
		/*	The connection is established: We should close our
			side as soon as all data are transmitted */
			spTCB->u8TCPState	= TCP_FINWAIT0;
			break;
/*----- TCP_CLOSEWAIT -------------------------------------------------------*/
		case	TCP_CLOSEWAIT:
		/*	we already received a FIN: Now we should
			close as soon as all our data are transmitted */
			spTCB->u8TCPState	= TCP_CLOSEWAIT0;
			break;
/*----- DEFAULT -------------------------------------------------------------*/
		default:
		/*	Other Status */
			goto	ret;
	}
/*	We force a transmitt now: */
		tcp_queuetxtcb(spTCB);
		/*	Signal the NetworkTask: We have something to transmit */
		SignalEvent(&NetworkEVENT);
ret:
	LeaveCritical();
}
/*****************************************************************************/
/*	Internal Functions used by the Network-stack */
/*****************************************************************************/
/*****************************************************************************/
/*	TCP:	Transmit-Handler */
/*****************************************************************************/
void	tcp_transmit(void){
TCPTCB*					spTCB;
/*	Check if we have s.th. to transmit */
	while(gspTcpSendQueue != 0){
		/*	We have something to do: Take the
			first one from the list */
		EnterCritical();
		spTCB			= gspTcpSendQueue;
		gspTcpSendQueue	= gspTcpSendQueue->spTxNext;
		spTCB->spTxNext	= (TCPTCB*)1;	//We ain't queued any more
		LeaveCritical();
		/*	Send stuff */
		tcp_send(spTCB);
	}
}
/*****************************************************************************/
/*	TCP StdIo-Interface */
/*****************************************************************************/
/*****************************************************************************/
/*	TCP: Put */
/*****************************************************************************/
int 	tcp_put(char cChr, FILE * spFile){
TCPTCB*	spTCB		= fdev_get_udata(spFile);
u8*		u8pWrite;
s8		s8Return	= 0;
	EnterCritical();
	while(spTCB->u16TxBufferFree == 0){
		/*	We have to wait until there is free space in the buffer */
		if(WaitEventTimeout(&spTCB->sTCPTxEvent,TCP_MAXWTXAITTIME) == EVENT_TIMEOUT){
			/*	The Event timed out! We reset the Connection! */
			tcp_reset(spTCB);
			s8Return		= -1;
			goto	end;
		}	/*	Check if the Connection is still alive */

	}
	/*	We have free space now. Lets write the Character to the Buffer */
	u8pWrite		= spTCB->u8pTxBufferWrite;
	 *u8pWrite++	= cChr;
	if(u8pWrite == (spTCB->TxBuffer+TCP_TXBUFFERLENGTH)){
		/*	We reached the end of the buffer!! wrap */
		u8pWrite	= spTCB->TxBuffer;
	}
	spTCB->u8pTxBufferWrite		= 	u8pWrite;
	spTCB->u16TxBufferFree--;
	if(spTCB->u16TxBufferFree == 0){
		tcp_queuetxtcb(spTCB);
		/*	The Buffer is full. Lets send the data */
		gu8NetworkSend	= 1;
		/*	Signal the tx-task that something's to do */
		SignalEvent(&NetworkEVENT);

	}
end:
	LeaveCritical();
	return	s8Return;
} 
/*****************************************************************************/
/*	TCP: Get */
/*****************************************************************************/
int 	tcp_get(FILE * spFile){
TCPTCB*	spTCB	= fdev_get_udata(spFile);
u8*		u8pRead;
int		s8Return	= 0;

	EnterCritical();
	while((TCP_RXBUFFERLENGTH - spTCB->u16RxBufferFree) == 0){
		/*	Check if the Connection is still alive */
		if((spTCB->u8TCPFlags & TCPFLAG_RXEOF) != 0){
			/*	The Remote End has signaled us EOF!!! */
			s8Return		= -1;
			goto	end;
		}
		if((spTCB->u8TCPFlags &	TCPFLAG_RESET) != 0){
			/*	The Remote End has signaled us EOF!!! */
			s8Return		= -1;
			goto	end;
		}
		tcp_queuetxtcb(spTCB);
		SignalEvent(&NetworkEVENT);
		if(WaitEventTimeout(&spTCB->sTCPRxEvent,TCP_MAXWRXAITTIME) == EVENT_TIMEOUT){
			/*	The Event timed out! We reset the Connection! */
			tcp_reset(spTCB);
			s8Return		= -1;
			goto	end;
		}	
	}
	/*	We have received something */
	u8pRead			= spTCB->u8pRxBufferRead;
	s8Return		=  *u8pRead++;
	if(u8pRead == (spTCB->RxBuffer+TCP_RXBUFFERLENGTH)){
		/*	We reached the end of the buffer!! wrap */
		u8pRead	= spTCB->RxBuffer;
	}
	spTCB->u8pRxBufferRead		= 	u8pRead;
	spTCB->u16RxBufferFree++;

	if(spTCB->u16RxBufferFree == TCP_RXBUFFERLENGTH){
		/*	Reset the Timer: We  send now */
		EnterCritical();
		spTCB->s8TCPTxTimer	= 0;
		LeaveCritical();
		/*	We have enough space, send an acknowledge */
		tcp_queuetxtcb(spTCB);
		/*	Signal the NetworkTask: We have something to transmit */
		SignalEvent(&NetworkEVENT);
	}
end:
	LeaveCritical();
	return	s8Return;
} 
/*****************************************************************************/
/*	TCP: Flush */
/*****************************************************************************/
void	tcp_flush(TCPTCB* spTCB){
	EnterCritical();
	/* Ensure that the PSH-Flag gets set */
	if(spTCB->u16TxBufferFree != TCP_TXBUFFERLENGTH){
	/* We flush only if there are remaining data to send */
		/*	Ensure that the PSH-Flag gets set */
		spTCB->u8TCPFlags		|= TCPFLAG_TXPSH;
		/*	We have to specify when we want to append a PSH-Flag */
		//!!!!!!!!!!!!!!!!!
		/*	Ensure we transmit */
		tcp_queuetxtcb(spTCB);
		/*	Signal the tx-task that something's to do */
		SignalEvent(&NetworkEVENT);
	}
	LeaveCritical();
}
/*****************************************************************************/
/*	TCP: BufferIO */
/*****************************************************************************/
/*****************************************************************************/
/*	TCP: Copy File */
/*****************************************************************************/
s8		tcp_copyfile(TCPTCB* spTCB,FATFILE* spFatFile){
u8*		u8pRead;
u8*		u8pWrite;
u8*		u8pStart;
u8*		u8pStop;
u16		u16Free;
u16		u16Written;
u16		u16WriteLength;
s8		s8Return	= 0;
		EnterCritical();
/*	Send everything which is in the buffer so far */
while(spTCB->u16TxBufferFree != TCP_TXBUFFERLENGTH){
	/*	Inform the TCP that we have s.th. to send!! */
		tcp_queuetxtcb(spTCB);
		/*	The Buffer is full. Lets send the data */
		gu8NetworkSend	= 1;
		/*	Signal the tx-task that something's to do */
		SignalEvent(&NetworkEVENT);
		if(WaitEventTimeout(&spTCB->sTCPTxEvent,TCP_MAXWTXAITTIME) == EVENT_TIMEOUT){
			/*	The Event timed out! We reset the Connection! */
			tcp_reset(spTCB);
			s8Return		= -1;
			goto	end;
		}	/*	Check if the Connection is still alive */
		if((TCPFLAG_RESET & spTCB->u8TCPFlags) != 0){
			/*	The Remote-End has reseted the connection...
				thus we have to stop transmitting data */
			s8Return		= -1;
			goto	end;
		}
}
/*	Set the Pointers to the beginning of the buffer */
spTCB->u8pTxBufferRead		= spTCB->TxBuffer;
spTCB->u8pTxBufferWrite		= spTCB->TxBuffer;

		while(1){
			/*	Wait until we have enough free space */
				while(1){
					/*	We have to wait until there is free space in the buffer */
					u16Free = spTCB->u16TxBufferFree;
						if(u16Free >= TCP_MINBUFFERFREE)	break;
						if(WaitEventTimeout(&spTCB->sTCPTxEvent,TCP_MAXWTXAITTIME) == EVENT_TIMEOUT){
							/*	The Event timed out! We reset the Connection! */
							tcp_reset(spTCB);
							s8Return		= -1;
							goto	end;
						}
					/*	Check if the connection has been reseted */
					if((spTCB->u8TCPFlags & TCPFLAG_RESET) != 0){
						/*	Our Connection has been reseted!! */
						s8Return		= -1;
						goto	end;
					}  
				}
				u8pRead			= spTCB->u8pTxBufferRead;
				u8pWrite		= spTCB->u8pTxBufferWrite;
				u8pStart		= spTCB->TxBuffer;
				u8pStop			= spTCB->TxBuffer+TCP_TXBUFFERLENGTH;
			/*	We have free space!! */
			while(1){
				/*	Copy data */
				u16WriteLength	=	(u16)(u8pStop - u8pWrite);
				if(u16Written != 512){
					u16Written	+= 1;
					u16Written	-= 1;
				}
				if(u16Free < u16WriteLength){
					/*	We have less space in the buffer. */
					u16WriteLength	=	(u16)(u8pRead - u8pWrite);
				}
				u16Written	= 	fat_readfile(spFatFile,u8pWrite,u16WriteLength);
				u8pWrite				+=	u16Written;
				/*	We have less free space now */
				u16Free					-=	u16Written;
				/*	Check if we reached the end of the buffer */
				if(u8pWrite == u8pStop){
					u8pWrite	= u8pStart;
				}
				if(u16Free == 0){
					/*	We have no free space left (thus the
						buffer didnt wrap around and we have some
						free space at the beginning */
					break;
				}
				if(u16Written != u16WriteLength){
					/*	We reached the EOF */
					break;
				}
			}
			/*	Write back the modified stuff */

	if((u16Free == TCP_TXBUFFERLENGTH) && (spTCB->u8pTxBufferRead != u8pWrite)){
			while(1);
	}

			spTCB->u8pTxBufferWrite	= u8pWrite;
			spTCB->u16TxBufferFree	= u16Free;


			/*	Inform the TCP that we have s.th. to send!! */
				tcp_queuetxtcb(spTCB);
				/*	The Buffer is full. Lets send the data */
				gu8NetworkSend	= 1;
				/*	Signal the tx-task that something's to do */
				SignalEvent(&NetworkEVENT);
			/*	Check if we copied everything */
			if(u16Written != u16WriteLength) {
				/*	We reached the EOF */
				goto	end;
			}
		}
end:
		LeaveCritical();
		return	s8Return;
	
}
/*****************************************************************************/
/* TO DO */
//In the receive FinitState Machine we have to check for valid ACK-Numbers as
// well!! (not only for the Flags...)
// Default Port management has to be improved:
// Check for ports already in use!!!!
//PSH-Flag management
