/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
//TCP
/*****************************************************************************/
#ifndef __TCP
#define	__TCP
#include <avr/interrupt.h>
#include <stdio.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <assert.h>
#include "global.h"
#include "system.h"
#include "network.h"
#include "ethernet.h"
#include "ip-udp.h"
#include "arp.h"
#include "icmp.h"
#include "fat.h"
/*****************************************************************************/
#define		TCP_MAXPACKETLENGTH	64
#define		TCP_MINPACKETLENGTH 32

#define		TCP_MINWINDOWSIZE 	128
#define		TCP_RXWAKEUP 		128
#define		TCP_RXBUFFERLENGTH	512

#define		TCP_TXBUFFERLENGTH	64
#define		TCP_MINBUFFERFREE 	32		//The TCP-Tx Event is signaled if this amount of buffer is availibel
#define		TCP_MAXWTXAITTIME 	2000	//We wait not longer than this long for a TX-Event
#define		TCP_MAXWRXAITTIME 	2000	//We wait not longer than this for a RX-Event
#define		TCP_TICK			50		//Timer Value for the TCP-Clock (ms)
/*	This Timeout-Values have to be multiplied by the
	TCP_TICK-Value to get the time in ms!!! */
#define	TCP_ZEROWINDOWTIMEOUT		200
#define	TCP_SYNSENDTIMEOUT			2

#define	TCP_BADSEQUENCENUMBERANSWER	10


#define	TCP_RETRANSMITTIMEOUT		10
#define	TCP_ACKNOWLEDGETIMEOUT		10

#define	TCP_FINACKNOWLEDGETIMEOUT	20
/*	The maximum count ouf Retransmission. After this number of
	Retransmissions we will reset the connection */
#define	TCP_MAXRETRANSMISSIONS		5
/*	The first default-port the tcp uses during a
	open-call */
#define	TCP_FIRSTLOCALPORT		10000
#define	TCP_LASTLOCALPORT		60000
/*****************************************************************************/
/*	Structs used by the tcp-code */
typedef	struct	_TCPHEADER	TCPHEADER;
typedef struct 	_TCPPORT	TCPPORT;
typedef struct	_TCPLISTEN	TCPLISTEN;
typedef struct	_TCPTCB		TCPTCB;

struct _TCPPORT{
u8	u8PortHi;
u8	u8PortLo;
};
struct _TCPHEADER{
	TCPPORT	sSourcePort;
	TCPPORT	sDestPort;
	u8	u8SeqNumber3;
	u8	u8SeqNumber2;
	u8	u8SeqNumber1;
	u8	u8SeqNumber0;
	u8	u8AckNumber3;
	u8	u8AckNumber2;
	u8	u8AckNumber1;
	u8	u8AckNumber0;
	u8	u8DataOffset;
	u8	u8Flags;
/*	TCP-Header Flags */
#define	TCP_FIN	0x01
#define	TCP_SYN	0x02
#define	TCP_RST	0x04
#define	TCP_PSH	0x08
#define	TCP_ACK	0x10
#define	TCP_URG	0x20
	u8	u8WindowHi;
	u8	u8WindowLo;
	u8	u8ChecksumHi;
	u8	u8ChecksumLo;
	u8	u8UrgentHi;
	u8	u8UrgentLo;
};
struct _TCPLISTEN{
	TCPLISTEN*	spNext;
	TCPTCB*		spNextTCB;
	TCPPORT		sTcpPort;
	u8			u8MaxTCB;
	EVENT		sListenEvent;
};

struct _TCPTCB{
	/*	The TCBs are organised in a SLL */
	TCPTCB*		spNext;
	/*	This List is used to manage the transmit of data */
	TCPTCB*		spTxNext;
	/*	The Listen-List */
	TCPTCB*		spListenNext;
	/*	This are the TCP-Parameters */
	s8			s8TCPTxTimer;
	u8			u8TCPRetransmissionCounter;
	u8			u8TCPState;
/*	The TCP-Finit-State-Machine */
#define	TCP_LISTEN				0x01
#define	TCP_OPEN				0x02
#define	TCP_SYNSENT				0x03
#define	TCP_SYNSENT1			0x05
	
#define	TCP_SYNRECEIVED			0x06
#define	TCP_ESTABLISHED			0x07
#define	TCP_FINWAIT0			0x10
#define	TCP_FINWAIT1			0x12
#define	TCP_FINWAIT2			0x13
#define	TCP_TIMEDWAIT			0x14
#define	TCP_CLOSEWAIT			0x20
#define	TCP_CLOSEWAIT0			0x21
#define	TCP_LASTACK				0x22
#define	TCP_CLOSED				0x30
#define	TCP_RESET0				0x40
#define	TCP_RESET				0x41
	u8			u8TCPFlags;
	/*	The Remote End has send a FIN: EOF reached */
#define	TCPFLAG_RXEOF			0x01
#define	TCPFLAG_RESET			0x02
#define	TCPFLAG_RXPSH			0x04
#define	TCPFLAG_TXPSH			0x08
	/*	Connection-Parameters */
	MACADDRESS	sRemoteMac;			//Remote MAC
	IPADDRESS	sRemoteIp;			//Remote IP
	TCPPORT		sRemoteTcpPort;		//Remote Port
	TCPPORT		sLocalTcpPort;		//Local Port
	/*	If  we have send this octed we
	 *  should send a PSH-Flag (address means TX-Buffer address!!! */
	u32		u32PshPosition;
	/*	The last bytenumber the remote end has acknowledged */
	u32		u32RemoteReceived;
	/*	The windowsize the remote end told us in the last packet */
	u16		u16RemoteWindowSize;
	/*	The last bytenumber we want to acknowledge (we received) */
	u32		u32LocalReceived;
	/*	TCP Events */
	EVENT	sTCPTxEvent;
	EVENT	sTCPRxEvent;
	/*	TCP Send Buffer */
	u8*		u8pTxBufferRead;
	u8*		u8pTxBufferWrite;
	u16		u16TxBufferFree;
	u16		u16TxBufferOffset;
	/*	TCP Receive Buffer */
	u8*		u8pRxBufferRead;
	u8*		u8pRxBufferWrite;
	u16		u16RxBufferFree;
	u8		RxBuffer[TCP_RXBUFFERLENGTH];
	u8		TxBuffer[TCP_TXBUFFERLENGTH];
};
/*****************************************************************************/
/*	Macros */
#define	SETTCPPORT(TCP,port)	\
		TCP.u8PortHi	= HI(port);\
		TCP.u8PortLo	= LO(port);

#define	TCPPORTCOMPARE(port1,port2) ((port1.u8PortLo == port2.u8PortLo) &&\
		(port1.u8PortHi == port2.u8PortHi))
/*****************************************************************************/
/* API */
/*****************************************************************************/
/* Init */
void		tcp_init(void);
/* Buffer IO */
s8			tcp_copyfile(TCPTCB* spTCB,FATFILE* spFatFile);
void		tcp_flush(TCPTCB* spTCB);
/* StdIO */
int 		tcp_get(FILE * spFile);
int 		tcp_put(char cChr, FILE * spFile);
/* Network-Stack */
void		tcp_receivepackage(void);
void		tcp_transmit(void);
void		tcp_send(TCPTCB* spTCB);
/* User API */
TCPTCB*		tcp_open(IPADDRESS* spRemoteIP, TCPPORT* spRemotePort);
void		tcp_reset(TCPTCB* spTCB);
void		tcp_closetx(TCPTCB* spTCB);
void		tcp_closerx(TCPTCB* spTCB);
char		tcp_stoplisten(TCPLISTEN* spTCPListen);
char		tcp_listen(TCPLISTEN* spTCPListen, char cBacklog);
TCPTCB*		tcp_accept(TCPLISTEN* spTCPListen);


void			tcp_timer_callback(TIMER* pTimer);
/*	ARP */
void	tcp_arpcallback(ARPQUERYNODE* pNode,u8 u8Code);

/* Functions from tcp-utils, internal use only */
void			tcp_queuetxtcb(TCPTCB* spTCB);
char			tcp_connectioncnt(TCPPORT* spPort);
void			tcp_unqueuetcb(TCPTCB* spTCB);
void			tcp_queuetcb(TCPTCB* spTCB);
inline void		tcp_setacknumber(TCPHEADER* spHeader,u32 ulAckNumber);
inline void		tcp_setseqnumber(TCPHEADER* spHeader,u32 ulSeqNumber);
inline void		tcp_setwindowsize(TCPHEADER* spHeader,u16 uiWindowSize);
inline void		tcp_seturgentpointer(TCPHEADER* spHeader,u16 uiUrgentPointer);
inline u16		tcp_windowsize(TCPHEADER* spHeader);
inline u32		tcp_acknumber(TCPHEADER* spHeader);
inline u32		tcp_seqnumber(TCPHEADER* spHeader);
u16				tcp_getlocalport(void);
inline u16		tcp_rxbytefree(TCPTCB* spTCB);
inline u16		tcp_txbytefree(TCPTCB* spTCB);
inline u16		tcp_rxbytefull(TCPTCB* spTCB);
inline u16		tcp_txbytefull(TCPTCB* spTCB);
inline u8*		tcp_txbufferstart(TCPTCB* spTCB);
inline u8*		tcp_txbufferend(TCPTCB* spTCB);
inline u8*		tcp_rxbufferstart(TCPTCB* spTCB);
inline u8*		tcp_rxbufferend(TCPTCB* spTCB);
inline u8*		tcp_txbufferreadpos(TCPTCB* spTCB);
inline u8*		tcp_txbufferwritepos(TCPTCB* spTCB);
inline u8*		tcp_rxbufferreadpos(TCPTCB* spTCB);
inline u8*		tcp_rxbufferwritepos(TCPTCB* spTCB);
/*****************************************************************************/
/*	Global Variables */
TCPTCB*			gspTcpQueue;
TCPTCB*			gspTcpSendQueue;
TCPLISTEN*		gspTcpListenList;
TIMER			gstcp_timer;
u16				gu16TcpDefaultLocalPort;
/*	ARP-Request for TCP-OPEN */
MACADDRESS		gsArpMac;
IPADDRESS		gsArpIp;
char			gsArpRequest;
/*	The Beginning of the arp-query-queue:
	This is a list of arp-queries */
ARPQUERYNODE 	gsTCPArpQuery;
TCPTCB*			gspTcpArpTCB;
EVENT			gsTCPArpEvent;
#define	TCPARPSTATUS_SUCCESS	1
#define	TCPARPSTATUS_ERROR		0
char			gsTCPArpStatus;
/*****************************************************************************/
#endif
