/*****************************************************************************/
/*	Copyright Sebastian Meuren 2003, 2008 
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*****************************************************************************/
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#include "inttypes.h"
#include "vs10xx.h"
#include "spi.h"

/*	Enters a critical region: Interrupts will be globaly disabled and
	the privious status is saved */
#define	EnterCritical 					\
	asm	volatile("in r0,__SREG__"::);	\
	asm	volatile("push r0"::);			\
	asm	volatile("cli"::);

/*	Leaves a critical region: The Interrupt-Status will be restored */
#define LeaveCritical					\
	asm	volatile("pop r0"::);			\
	asm	volatile("out __SREG__,r0"::);


/*****************************************************************************/
/*	VS10xx: Read a SCI-Register */
/*****************************************************************************/
u16	vs10xx_cntr_rd(u8 u8Address){
u16		u16Ret	= 0;
		/*	Wait until the SPI-Bus is ready */
		WaitEvent(&gsSpiLockEvent);
		/*	Switch to SPI-Lowspeed mode:
			1/128 System-Clock, SPR0=1,SPR1=1,SPI2X=0 */
			SPCR0	|=  (_BV(SPR00) | _BV(SPR01));
		/*	Set SPI2X = 0!! (diffrent register!)  */
			SPSR0	= 0;
		/*	Make XCS low */
		VS10xx_DESELECT_XCS
		/*	Write the control-word: READ */
		spi_rw(3);
		/*	Write the Number of the register out */
		spi_rw(u8Address);
		/*	Read the Register in */
		u16Ret =		((u16)spi_rw(0)) << 8;
		u16Ret |=	((u16)spi_rw(0));
		/* XCS high again */
		VS10xx_SELECT_XCS
		/*	Release the SPI-Bus */
		SignalEvent(&gsSpiLockEvent);
		/*	Wait for DREQ to go high */
		while(VS10xx_DREQ_STATUS == 0);
		return u16Ret;
}
/*****************************************************************************/
/*	VS10xx: Write a SCI-Register */
/*****************************************************************************/
void	vs10xx_cntr_wr(u8 u8Address,u16 u16Value){
		/*	Wait until the SPI-Bus is ready */
		WaitEvent(&gsSpiLockEvent);
		/*	Switch to SPI-Lowspeed mode:
			1/128 System-Clock, SPR0=1,SPR1=1,SPI2X=0 */
			SPCR0	|=  (_BV(SPR00) | _BV(SPR01));
		/*	Set SPI2X = 0!! (diffrent register!)  */
			SPSR0	= 0;
		/*	Make XCS low */
		VS10xx_DESELECT_XCS
		/*	Write the control-word: READ */
		spi_rw(2);
		/*	Write the Number of the register out */
		spi_rw(u8Address);
		/*	Read the Register in */
		spi_rw(u16Value>>8);
		spi_rw(u16Value);
		/* XCS high again */
		VS10xx_SELECT_XCS
		/*	Release the SPI-Bus */
		SignalEvent(&gsSpiLockEvent);
		/*	Wait for DREQ to go high */
		while(VS10xx_DREQ_STATUS == 0);
}
/*****************************************************************************/
/*	VS10xx: Reset a SCI-Register */
/*****************************************************************************/
void	vs10xx_reset(void){
	/*	The vs1001 needs the following configuration:
		We use the clock-doubler of the chip. This means we have
		to start with a SPI-Frequency below 2 Mhz (see Appnote)
		We set	SPR0 = SPR1 = 1, SPI2X=0 which
		equals 1/128 of the system frequency 
		SPIE = 0: No Interrupts used so far
		SPE	= 1: SPI enabled
		DORD = 0: MSB first 
		MSTR = 1: Master Operation
		CPOL = 0: SCK low when idle
		CPHA = 0: Sample on the leading edge
		(SPI2X = 0): Choose fosc/4 as the SCK-frequency
		SPR1 = 1
		SPR0 = 1 */
		SPCR0	= 	_BV(SPE0) | _BV(MSTR0) | _BV(SPR01);
	/*	Set SPI2X = 0!! (diffrent register!)  */
		SPSR0	= 0;
		//-----------------------------------------------------------------------------
		/*	Setting up the VS1002 */
		//-----------------------------------------------------------------------------
		/*	Enable AVR-INT0-Pin (DREQ-Interrupt):
			A rising Flag should trigger (ATmega64) */
		//	EICRA	|=	(_BV(ISC31) | _BV(ISC30));
		//	EIMSK	|=	_BV(INT3);	//Enable the Interrupt!
		//-----------------------------------------------------------------------------
		/*	VS10xx Hardware Reset */
		//-----------------------------------------------------------------------------
		/*	Hardware-reset (SDI and SCI-Interfaces are deselected !!!) */
			VS10xx_DESELECT_XRESET
			_delay_ms(10);
			VS10xx_SELECT_XRESET
			_delay_ms(10);
		/*	Software-Reset: The vs10xx is in High-Speed SPI-Mode
			afterwards (e.g. the clock doubler is active */
		//-----------------------------------------------------------------------------
		/*	VS10xx Software Reset */
		//-----------------------------------------------------------------------------
		//	vs10xx_reset();
		/*	Setting Up the initial Audio-Volume */
		//	vs10xx_cntr_wr(11,VS10xx_INITVOLUME | (VS10xx_INITVOLUME << 8));
		//	while(VS10xx_DREQ_STATUS == 0);		//Wait for DREQ to go high

		//	VS10xx_DESELECT_XDCS
		//	asm volatile("nop"::);
		//	asm volatile("nop"::);
		//	VS10xx_SELECT_XDCS
		//	vs10xx_sinetest(100);
	u16	u16reg;
		//Adjust the clock frequency to times 4
	vs10xx_cntr_wr(3,0xD800);
	u16reg	= vs10xx_cntr_rd(3);

	vs10xx_cntr_wr(13,0x5555);
	u16reg	= vs10xx_cntr_rd(13);
	
	
//		vs10xx_cntr_wr(2,0xf3fa);
//		u16reg	= vs10xx_cntr_rd(2);
		//Now the clock-frequency is adjusted - send data faster
	/*	The vs1001 needs the following configuration:
		We use the clock-doubler of the chip. This means we have
		to start with a SPI-Frequency below 2 Mhz (see Appnote)
		We set	SPR0 = SPR1 = 1, SPI2X=0 which
		equals 1/128 of the system frequency 
		SPIE = 0: No Interrupts used so far
		SPE	= 1: SPI enabled
		DORD = 0: MSB first 
		MSTR = 1: Master Operation
		CPOL = 0: SCK low when idle
		CPHA = 0: Sample on the leading edge
		(SPI2X = 0): Choose fosc/4 as the SCK-frequency
		SPR1 = 0
		SPR0 = 0 */
		SPCR0	= 	_BV(SPE0) | _BV(MSTR0) ;
	/*	Set SPI2X = 0!! (diffrent register!)  */
		SPSR0	= 0;


	
	
	/*	Reset the vs10xx */
	/*	Set the Software-Reset Flag in the Mode-register */
//	vs10xx_cntr_wr(0,0x0004);
	/*	Wait at least 2 us */
//	_delay_us(40);
	/*	Now wait for DREQ to go high */
//	while(VS10xx_DREQ_STATUS == 0);
	/*	Setupt the Interface:
		We dont share the XCS-Pin and we use the VS1002-Native Mode for
		Datatransfer 
		Bit.2:	SoftwareReset 0
		Bit.5:	Allow SDI-Tests 0
		Bit.10: SDI-Share 0
		Bit.11:	Native SPI-Mode 1
		*/
//	vs10xx_cntr_wr(0,0x0800);
/*	Activate the Clock-Doubler: we use a 12.288 Mhz Clock. We have to write
	6144 and to set the MSB to one: */
//	vs10xx_cntr_wr(3,0x9800);
/*	Write s.th. to the AUDATA-Register */
//	vs10xx_cntr_wr(5,0x1f40);		//0x1f40 for 8000 Hz (from Appnote)
/*	wait for at least 11000 clock cycles (Appnote):
	@12.288 Mhz we have to wait ~ 1ms*/
//	_delay_ms(2);  //!!!!!!!!!!!!!!!!!!!!
//	while(VS10xx_DREQ_STATUS == 0);		//Wait for DREQ to go high
/*	Now we can use an SPI-speed up to 4 Mhz */
}
/*****************************************************************************/
/*	vs1001 - Interrupt Routine */
/*****************************************************************************/
ISR(INT0_vect){
/*	The vs1001 has risen the DREQ-Line. We should inform the task */
	SignalEvent(&gVS10xxDREQEvent);
}
/*****************************************************************************/
/*	INIT */
/*****************************************************************************/
u8		vs10xx_init(void){
//-----------------------------------------------------------------------------
/*	Setting up the VS1002 */
//-----------------------------------------------------------------------------
	VS10xx_CONFIGURE_XRESET
	VS10xx_CONFIGURE_XDCS
	VS10xx_CONFIGURE_XCS
	VS10xx_CONFIGURE_DREQ
	EVENTINIT(gVS10xxDREQEvent);
	
	/*	Enable AVR-INT0-Pin (DREQ-Interrupt):
		A rising Flag should trigger (ATmega644) */
		EICRA	|=	(_BV(ISC01) | _BV(ISC00));
		EIMSK	|=	_BV(INT0);	//Enable the Interrupt!

	
/*	Enable AVR-INT0-Pin (DREQ-Interrupt):
	A rising Flag should trigger (ATmega64) */
//	EICRA	|=	(_BV(ISC31) | _BV(ISC30));
//	EIMSK	|=	_BV(INT3);	//Enable the Interrupt!
//-----------------------------------------------------------------------------
/*	VS10xx Hardware Reset */
//-----------------------------------------------------------------------------
/*	Hardware-reset (SDI and SCI-Interfaces are deselected !!!) */
//	_delay_ms(100);
//	VS10xx_DESELECT_XRESET
//	_delay_ms(100);
//	VS10xx_SELECT_XRESET
/*	Software-Reset: The vs10xx is in High-Speed SPI-Mode
	afterwards (e.g. the clock doubler is active */
//-----------------------------------------------------------------------------
/*	VS10xx Software Reset */
//-----------------------------------------------------------------------------
	vs10xx_reset();
/*	Setting Up the initial Audio-Volume */
//	vs10xx_cntr_wr(11,VS10xx_INITVOLUME | (VS10xx_INITVOLUME << 8));
//	while(VS10xx_DREQ_STATUS == 0);		//Wait for DREQ to go high

//	VS10xx_DESELECT_XDCS
//	asm volatile("nop"::);
//	asm volatile("nop"::);
//	VS10xx_SELECT_XDCS
//	vs10xx_sinetest(100);
	return	0;
}
/*****************************************************************************/

/*****************************************************************************/
/*	VS10xx: Sine-Test */
/*****************************************************************************/
void	vs10xx_sinetest(u32 u32time){
/*	Allow the sine-test */
u16	u16reg;

/*	Read the current value of the SCI_MODE-Register */
	u16reg	= vs10xx_cntr_rd(0);
/*	Set Bit.5: Allow SDI-tests */
	u16reg	|=	_BV(5);
	vs10xx_cntr_wr(0,u16reg);
	while(VS10xx_DREQ_STATUS == 0);		//Wait for DREQ to go high
/*	Perform the Test */
	VS10xx_DESELECT_XDCS
/*	Start SINE-Test */
	spi_rw(0x53);
	spi_rw(0xEF);
	spi_rw(0x6E);
	spi_rw(252);
	spi_rw(0x00);
	spi_rw(0x00);
	spi_rw(0x00);
	spi_rw(0x00);
/*	Wait the given time */
//	_delay_ms(100);
/*	Stop it again */
	spi_rw(0x45);
	spi_rw(0x78);
	spi_rw(0x69);
	spi_rw(0x74);
	spi_rw(0x00);
	spi_rw(0x00);
	spi_rw(0x00);
	spi_rw(0x00);
	VS10xx_SELECT_XDCS
/*	Disable SDI-Tests again */
	u16reg	&=	~_BV(5);
	vs10xx_cntr_wr(0,u16reg);
	while(VS10xx_DREQ_STATUS == 0);		//Wait for DREQ to go high
}
